--------------------------------------------------------------------------
-- Assirc_Database.Alignment                Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 10/01/1999 --
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;
with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use Strlib;
with Strlib.Io; use Strlib.Io;
with Options_Util; use Options_Util;
  
with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

with Sequence_Alignment; use Sequence_Alignment;
with Sequence_Alignment.Print; use Sequence_Alignment.Print;
with Sequence_Alignment.Global; use Sequence_Alignment.Global;
with Sequence_Alignment.Local; use Sequence_Alignment.Local;

with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Database; use Assirc_Specifications.Database;
with Assirc_Global_Variables; use Assirc_Global_Variables;
with Assirc_Database_Functions; use Assirc_Database_Functions;
with Assirc_Statistics; use Assirc_Statistics;

package body Assirc_Database.Alignment is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
  use Type_Specifications.IO.Real_Io;
  use Type_Specifications.Real_Lib;

  package SRRBT renames Assirc_Specifications.Database.Similar_Regions_RB_Tree;
  use SRRBT;

  procedure Perform_Alignments (Name : in String;
				Display : in Output_Display;
				Minimal_Length : in Integer;
				Minimal_Similarity : in Percentage;
				Method : in Alignment_Method;
				Mismatch_Weight, Match_Weight : in Integer;
				Gap_Weight,Gap_Length_Weight : in Integer;
				With_End_Weight : in Boolean;
				Max_Length : in Integer;
			        Adjust : in Integer;
				Similarity_Threshold : in Integer;
				Randomization : in Integer;
				Preserve : in Integer) is

    Suffix : constant String := ".sim";
    
    Disp : Output_Display := Display;
    -- Pour dire ou on ecrit les resultats
    File : aliased Ada.Text_Io.File_Type;
    -- Le pointeur vers le fichier
    Output : Ada.Text_Io.File_Access;
    -- et un accs  ce pointeur

    procedure Put_Not_Align(File : in Ada.Text_Io.File_Type;
			    Seq1, Seq2 : in Sequence_Data;
			    Pair   : in out Segment_Pair) is
      Id_Length : constant Integer := Summary_Id_Length;
    begin
      Put_The_N_Last_Characters(File,To_String(Seq1.Identifier),Id_Length);
      Put(File," " & Image(Pair.Seg1.Strand));
      case Pair.Seg1.Strand is
	when Watson =>
	  Put(File,Pair.Seg1.Start,12);
	  Put(File,Pair.Seg1.Stop,12);
	when Crick => 
	  Put(File,Pair.Seg1.Stop,12);
	  Put(File,Pair.Seg1.Start,12);
	when Both =>
	  raise Invalid_Strand;
      end case;
      Put(File,' ');
      Put_The_N_Last_Characters(File,To_String(Seq2.Identifier),Id_Length);
      Put(File," " & Image(Pair.Seg2.Strand));
      case Pair.Seg2.Strand is
	when Watson =>
	  Put(File,Pair.Seg2.Start,12);
	  Put(File,Pair.Seg2.Stop,12);
	when Crick =>
	  Put(File,Pair.Seg2.Stop,12);
	  Put(File,Pair.Seg2.Start,12);
      	when Both =>
	  raise Invalid_Strand;
      end case;
      Put_Line(File," NOT ALIGNED");
    end Put_Not_Align;

    procedure Perform_One_Alignment(Pair   : in out Segment_Pair;
				    Output : in out Ada.Text_Io.File_Access) is
      T_Seq1,T_Seq2  : Sequence_Data;
      Alignments     : Alignments_Access;
      L_Start,L_Stop : Integer;
    begin
      if Pair.Length < Minimal_Length then return; end if;
      -- Dtermination de la position corrige de dbut et de fin
      -- Calcul sur le brin watson. On ajuste aprs.
      L_Start := Integer'Max(Pair.Seg1.Start - Adjust,
			     Sequences(Pair.Seg1.Id).Nucleotides'First);
      L_Stop  := Integer'Min(Pair.Seg1.Stop + Adjust,
			     Sequences(Pair.Seg1.Id).Nucleotides'Last);
      -- Initialisation de la premire squence
      case Pair.Seg1.Strand is
	when Watson =>
	  T_Seq1.Name := Sequences(Pair.Seg1.Id).Name;
	  T_Seq1.Identifier := 
	    Sequences(Pair.Seg1.Id).Identifier;-- & Image(Pair.Seg1.Strand);
	  T_Seq1.Model := Sequences(Pair.Seg1.Id).Model;
	  T_Seq1.Format := Sequences(Pair.Seg1.Id).Format;
	  T_Seq1.Length := Sequences(Pair.Seg1.Id).Length;
	  T_Seq1.Nucleotides := Sequences(Pair.Seg1.Id).Nucleotides;
	  T_Seq1.Start := L_Start;
	  T_Seq1.Stop := L_Stop;
      	when Crick =>
	  T_Seq1.Name := Sequences(Pair.Seg1.Id).Name;
	  T_Seq1.Identifier := 
	    Sequences(Pair.Seg1.Id).Identifier;-- & Image(Pair.Seg1.Strand);
	  T_Seq1.Model := Sequences(Pair.Seg1.Id).Model;
	  T_Seq1.Format := Sequences(Pair.Seg1.Id).Format;
	  T_Seq1.Length := Sequences(Pair.Seg1.Id).Length;
	  T_Seq1.Start := Sequences(Pair.Seg1.Id).Length-L_Stop+1;
	  T_Seq1.Stop := Sequences(Pair.Seg1.Id).Length-L_Start+1;
	  T_Seq1.Nucleotides :=  
	    new Nucleotidic_Array(L_Start..L_Stop);
	  T_Seq1.Nucleotides.all := 
	    Sequences(Pair.Seg1.Id).Nucleotides(L_Start..L_Stop);
	  Reverse_Sequence(T_Seq1);
	  Cmp_Sequence(T_Seq1);
	when Both =>
	  raise INVALID_STRAND;
      end case;
      -- Dtermination de la position corrige de dbut et de fin
      -- Calcul sur le brin watson. On ajuste aprs.
      L_Start := Integer'Max(Pair.Seg2.Start - Adjust,
			     Sequences(Pair.Seg2.Id).Nucleotides'First);
      L_Stop  := Integer'Min(Pair.Seg2.Stop + Adjust,
			     Sequences(Pair.Seg2.Id).Nucleotides'Last);
      -- Initialisation de la premire squence
      case Pair.Seg2.Strand is
	when Watson =>
	  T_Seq2.Name := Sequences(Pair.Seg2.Id).Name;
	  T_Seq2.Identifier := 
	    Sequences(Pair.Seg2.Id).Identifier;-- & Image(Pair.Seg2.Strand);
	  T_Seq2.Model := Sequences(Pair.Seg2.Id).Model;
	  T_Seq2.Format := Sequences(Pair.Seg2.Id).Format;
	  T_Seq2.Length := Sequences(Pair.Seg2.Id).Length;
	  T_Seq2.Nucleotides := Sequences(Pair.Seg2.Id).Nucleotides;
	  T_Seq2.Start := L_Start;
	  T_Seq2.Stop := L_Stop;
      	when Crick =>
	  T_Seq2.Name := Sequences(Pair.Seg2.Id).Name;
	  T_Seq2.Identifier := 
	    Sequences(Pair.Seg2.Id).Identifier;-- & Image(Pair.Seg2.Strand);
	  T_Seq2.Model := Sequences(Pair.Seg2.Id).Model;
	  T_Seq2.Format := Sequences(Pair.Seg2.Id).Format;
	  T_Seq2.Length := Sequences(Pair.Seg2.Id).Length;
	  T_Seq2.Start := Sequences(Pair.Seg2.Id).Length-L_Stop+1;
	  T_Seq2.Stop := Sequences(Pair.Seg2.Id).Length-L_Start+1;
	  T_Seq2.Nucleotides :=  
	    new Nucleotidic_Array(L_Start..L_Stop);
	  T_Seq2.Nucleotides.all := 
	    Sequences(Pair.Seg2.Id).Nucleotides(L_Start..L_Stop);
	  Reverse_Sequence(T_Seq2);
	  Cmp_Sequence(T_Seq2);
	when Both =>
	  raise INVALID_STRAND;
      end case;
      if ((T_Seq1.Stop - T_Seq1.Start + 1 > Max_Length) or
	  (T_Seq2.Stop - T_Seq2.Start + 1 > Max_Length)) then
	if Disp.File then
	  Put_Not_Align(Output.all,T_Seq1,T_Seq2,Pair);
	end if;
      	if Disp.Screen then
	  Put_Not_Align(Ada.Text_Io.Standard_Output,T_Seq1,T_Seq2,Pair);
	end if;
      else
	case Method is
          when Global_Alignment =>
            Perform_Global_Alignment(T_Seq1,T_Seq2,Alignments, 
				     T_Seq1.Start,T_Seq1.Stop,
				     T_Seq2.Start,T_Seq2.Stop,
				     "",
				     Match_Weight,Mismatch_Weight,
				     Gap_Weight,Gap_Length_Weight,
				     DVH,One,No,With_End_Weight,
				     Similarity_Threshold,
				     Randomization,Preserve);
	    
          when Local_Alignment =>
	    Perform_Local_Alignment(T_Seq1,T_Seq2,Alignments,
				    T_Seq1.Start,T_Seq1.Stop,
				    T_Seq2.Start,T_Seq2.Stop,
				    "",
				    Match_Weight,Mismatch_Weight,
				    Gap_Weight,Gap_Length_Weight,
				    DVH,One,No,Similarity_Threshold,
				    Randomization,Preserve);
	  when others =>
            raise NOT_IMPLEMENTED;
	end case;
	-- Modifier les positions de dbut et de fin de l'alignement pour
	-- tenir compte du brin sur lequel s'est fait le calcul et
	-- imprimer le rsultat
	for I in Alignments'Range loop
	  Alignments(I).Strand1 := Pair.Seg1.Strand;
	  case Pair.Seg1.Strand is
            when Watson =>
              null;
            when Crick =>
              Alignments(I).Start1 := 
                Sequences(Pair.Seg1.Id).Length - Alignments(I).Start1 + 1;
              Alignments(I).Stop1  := 
                Sequences(Pair.Seg1.Id).Length - Alignments(I).Stop1 + 1;
            when others =>
              raise INVALID_STRAND;
	  end case;
	  Alignments(I).Strand2 := Pair.Seg2.Strand;
	  case Pair.Seg2.Strand is
	    when Watson =>
	      null;
	    when Crick =>
	      Alignments(I).Start2 := 
		Sequences(Pair.Seg2.Id).Length - Alignments(I).Start2 + 1;
	      Alignments(I).Stop2  := 
		Sequences(Pair.Seg2.Id).Length - Alignments(I).Stop2 + 1;
	    when others =>
	      raise INVALID_STRAND;
	  end case;
	  -- Afficher le rsultat
	  if not(Alignments(I).Start1 = Alignments(I).Start2 and
		 Alignments(I).Stop1 = Alignments(I).Stop2 and
		 Alignments(I).Name1 = Alignments(I).Name2 and
		 Alignments(I).Strand1 = Alignments(I).Strand2) 
	    and then
	    (Alignments(I).Length >= Minimal_Length and
	     --Alignments(I).Similarity >= Minimal_Similarity) 
	     Alignments(I).Identity >= Minimal_Similarity) 
	  then
	    if Disp.File then
	      Print_Summary(Output.all,Alignments(I).all);
	      New_Line(Output.all);
	    end if;
	    if Disp.Screen then
	      Print_Summary(Ada.Text_Io.Standard_Output,Alignments(I).all);
	      New_Line(Ada.Text_Io.Standard_Output);
	    end if;
	    Statistics.Increment_Validated;
	  end if;
	end loop;
      end if;
      -- Librer la place rserve pour dterminer la squence rverse
      -- dans le cas d'un brin Crick
      if Pair.Seg1.Strand = Crick then
	Free(T_Seq1.Nucleotides);
      end if;
      if Pair.Seg2.Strand = Crick then
	Free(T_Seq2.Nucleotides);
      end if;
    end Perform_One_Alignment;

    procedure Perform_All_Alignments is new 
     SRRBT.Apply_With_Data(Ada.Text_Io.File_Access,Perform_One_Alignment);    
  begin
    if Disp.File then
      -- ouvre un fichier name avec terminaison ".sim"
      if Name'Length > 5 and then Name(Name'Last-4..Name'Last) = Suffix then
	Create(File,Out_File,Name);
      else
	Create(File,Out_File,Name & Suffix);
      end if;
      Output := File'Unchecked_Access; 
      -- Le Unchecked est requis car un File_Type est limit et faire
      -- cette opration donne la possibilit de realiser une affectation
      -- dessus. La dfinition de Rb_Tree.Apply_With_Data suppose que le
      -- second parametre est pass en in_out. Il doit donc etre donne
      -- comme une variable, d'ou la ncessit de passer par le File_Access
      -- Voir autres remarques dans assirc_database.print_pairs
    end if;
    Perform_All_Alignments(Similar_Regions.Tree,Output);
    if Disp.File then
      Close(File);
    end if;
  exception
    when Name_Error =>
      -- si cela n'est pas possible, on ouvre le fichier dans /tmp
      declare
        New_Name : constant String := "/tmp/" & 
          Root_Name(To_Varstring(Name)) & Suffix;
      begin
        Create(File,Out_File,New_Name);
        Put_Line("WARNING: Output data are write in " & New_Name);
        Output := File'Unchecked_Access;
        -- Meme remarque que ci-dessus
        Perform_All_Alignments(Similar_Regions.Tree,Output);
        Close(File); 
      exception
        when Name_Error =>
          -- et sinon on ecrit sur la sortie standard
          Output := Standard_Output;
          Put_Line("WARNING: Output data are write to standard output");
	  Disp.Screen := False; -- Pour eviter d'afficher deux fois 
          Perform_All_Alignments(Similar_Regions.Tree,Output);
      end;
  end Perform_Alignments;

end Assirc_Database.Alignment;
  
