--------------------------------------------------------------------------
-- Assirc_Database_Functions                Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 23/03/1998 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications.IO; use Type_Specifications.IO;

with Sequence_Specifications; use Sequence_Specifications;

package body Assirc_Database_Functions is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
  use Type_Specifications.IO.Real_Io;
  use Type_Specifications.Real_Lib;

  package SRRBT renames Assirc_Specifications.Database.Similar_Regions_RB_Tree;
  use SRRBT;
  
--------------------------------------------------------------------------
-- Dtermination des bornes hautes et basses pour deux segments         --
--------------------------------------------------------------------------
  procedure Define_Low_High(Seg1,Seg2   : in Segment_Type;
                            Max_Seg_Len : in Integer;
                            Low,High    : in out Segment_Pair);
  pragma Inline(Define_Low_High);
  procedure Define_Low_High(Seg1,Seg2   : in Segment_Type;
                            Max_Seg_Len : in Integer;
                            Low,High    : in out Segment_Pair) is
    -- Compte-tenu de la taille de deux segments, ce programme calcule
    -- la position minimale et la position maximale o ils peuvent
    -- apparaitre dans la base. Seul le champ start des deux paires doit
    -- etre determine, compte-tenu du principe d'enregistrement dans la base de 
    -- donnees.
  begin
    Low.Seg1.Id := Seg1.Id;
    Low.Seg1.Strand := Seg1.Strand;
    Low.Seg2.Id := Seg2.Id;
    Low.Seg2.Strand := Seg2.Strand;
    High.Seg1.Id := Seg1.Id;
    High.Seg1.Strand := Seg1.Strand;
    High.Seg2.Id := Seg2.Id;
    High.Seg2.Strand := Seg2.Strand;
    Low.Seg1.Stop  := Integer'Last;
    Low.Seg2.Stop  := Integer'Last;
    High.Seg1.Stop  := Integer'Last;
    High.Seg2.Stop  := Integer'Last;
    if Seg1.Id = Seg2.Id and Seg1.Strand = Seg2.Strand then
      -- ATTENTION, on suppose que l'on n'a jamais les deux brins en Crick
      -- Cette configuration n'etant jamais analysee. On est donc ici
      -- obligatoirement avec les deux brins en Watson
      -- Cela est cependant sans consequence car dans la base de donnee, on
      -- memorise toujours les paires avec Start < Stop
      if Seg1.Start < Seg2.Start then
        Low.Seg1.Start := Seg1.Start - Max_Seg_len;
      else
        Low.Seg1.Start := Seg2.Start - Max_Seg_len;
      end if;
      Low.Seg2.Start := Low.Seg1.Start;
      if Seg1.Stop > Seg2.Stop then
        High.Seg1.Start := Seg1.Stop;
      else
        High.Seg1.Start := Seg2.Stop;
      end if;
      High.Seg2.Start := High.Seg1.Start;
    else
      -- Que l'on soit en Watson ou en Crick, le raisonnement est le
      -- meme car dans tous les cas Start < Stop
      Low.Seg1.Start := Seg1.Start - Max_Seg_len;
      Low.Seg2.Start := Seg2.Start - Max_Seg_len;
      High.Seg1.Start := Seg1.Stop;
      High.Seg2.Start := Seg2.Stop;
    end if;
  end Define_Low_High;

--------------------------------------------------------------------------
-- Dtermination de l'inclusion d'une paire dans la base                --
--------------------------------------------------------------------------
  function Included(Seg1,Seg2       : in Segment_Type;
                    Similar_Regions : in Similar_Regions_Tree;
                    Excluded_Marge  : in Integer)
    return Boolean is

    Present   : Boolean := False;
    Low,High : Segment_Pair;

    function Check_Inclusion_For_An_Element(Element : in Segment_Pair;
                                            Data    : in Boolean)
                                            return Boolean is
      Mrg1 : constant Integer :=
        ((Element.Seg1.Stop-Element.Seg1.Start+1)*Excluded_Marge) / 100;
      Mrg2 : constant Integer :=
        ((Element.Seg2.Stop-Element.Seg2.Start+1)*Excluded_Marge) / 100;
    begin
      if Seg1.Id = Element.Seg1.Id and Seg2.Id = Element.Seg2.Id and
         Seg1.Strand = Element.Seg1.Strand and Seg2.Strand = Element.Seg2.Strand then
        -- les segments referent aux bonnes sequences
        if (Seg1.Start >= Element.Seg1.Start+Mrg1 and
            Seg1.Stop  <= Element.Seg1.Stop-Mrg1  and
            Seg2.Start >= Element.Seg2.Start+Mrg2 and
            Seg2.Stop  <= Element.Seg2.Stop-Mrg2) then
          -- le segment est bien inclus
          return True;
        end if;
        if ((Seg1.Id = Seg2.Id and Seg1.Strand = Seg2.Strand) and then
            (Seg1.Start >= Element.Seg2.Start+Mrg1 and
             Seg1.Stop  <= Element.Seg2.Stop-Mrg1  and
             Seg2.Start >= Element.Seg1.Start+Mrg2 and
             Seg2.Stop  <= Element.Seg1.Stop-Mrg2)) then
          -- les deux sequences sont identiques, on a verifie
          -- que le segment 1 est inclus dans 2 et vice-versa
          return True;
        end if;
      end if;
      return False;
    end Check_Inclusion_For_An_Element;

    function Check_Inclusion is new
      SRRBT.Find_Between(Boolean,Check_Inclusion_For_An_Element);
  begin
    Define_Low_High(Seg1,Seg2,Similar_Regions.Max_Seg_Length,Low,High);
    return Check_Inclusion(Similar_Regions.Tree,Present,Low,High) /= Null_Ptr;
  end Included;

--------------------------------------------------------------------------
-- Dtermination si une paire a une intersection avec une paire de la   --
-- base                                                                 --
--------------------------------------------------------------------------
  function Find_Intersection(Seg1,Seg2          : in Segment_Type;
                             Similar_Regions    : in Similar_Regions_Tree;
                             Intersection_Marge : in Integer)
                             return Similar_Regions_Ptr;
  pragma Inline(Find_Intersection);
  function Find_Intersection(Seg1,Seg2          : in Segment_Type;
                             Similar_Regions    : in Similar_Regions_Tree;
                             Intersection_Marge : in Integer)
                             return Similar_Regions_Ptr is

    Present   : Boolean := False;
    Low,High  : Segment_Pair;
    Extension : Integer renames Intersection_Marge;
    Nseg1     : Segment_Type := (Id     => Seg1.Id,
                                 Strand => Seg1.Strand,
                                 Start  => Seg1.Start - Extension,
                                 Stop   => Seg1.Stop  + Extension);
    Nseg2     : Segment_Type := (Id     => Seg2.Id,
                                 Strand => Seg2.Strand,
                                 Start  => Seg2.Start - Extension,
                                 Stop   => Seg2.Stop  + Extension);

    function Check_Intersection_For_An_Element(Element : in Segment_Pair;
                                               Data    : in Boolean)
                                               return Boolean is
    begin
      if Seg1.Id = Element.Seg1.Id and Seg2.Id = Element.Seg2.Id and
         Seg1.Strand = Element.Seg1.Strand and Seg2.Strand = Element.Seg2.Strand then
        -- le debut d'un segment doit etre inclus dans l'autre
        if (((Nseg1.Start >= Element.Seg1.Start and
              Nseg1.Start <= Element.Seg1.Stop)  or
             (Element.Seg1.Start >= Nseg1.Start and
              Element.Seg1.Start <= Nseg1.Stop))  and
            ((Nseg2.Start >= Element.Seg2.Start and
              Nseg2.Start <= Element.Seg2.Stop)  or
             (Element.Seg2.Start >= Nseg2.Start and
              Element.Seg2.Start <= Nseg2.Stop))) then
          -- les deux paires de segments sont chevauchantes
          return True;
        end if;
        if ((Seg1.Id = Seg2.Id and Seg1.Strand = Seg2.Strand) and then
            (((Nseg1.Start >= Element.Seg2.Start and
               Nseg1.Start <= Element.Seg2.Stop)  or
              (Element.Seg1.Start >= Nseg2.Start and
               Element.Seg1.Start <= Nseg2.Stop))  and
             ((Nseg2.Start >= Element.Seg1.Start and
               Nseg2.Start <= Element.Seg1.Stop)  or
              (Element.Seg2.Start >= Nseg1.Start and
               Element.Seg2.Start <= Nseg1.Stop)))) then
          -- les deux sequences sont identiques, on a verifie
          -- que le segment 1 chevauche 2 et vice-versa
          return True;
        end if;
      end if;
      return False;
    end Check_Intersection_For_An_Element;


    function Check_Intersection is new
      SRRBT.Find_Between(Boolean,Check_Intersection_For_An_Element);

  begin
    Define_Low_High(Nseg1,Nseg2,Similar_Regions.Max_Seg_Length,Low,High);
    return Check_Intersection(Similar_Regions.Tree,Present,Low,High);
  end Find_Intersection;

--------------------------------------------------------------------------
-- Dtermination de l'autorecouvrement de segments                      --
--------------------------------------------------------------------------

  function Is_Auto_Recover(A : in Segment_Pair) return Boolean is
  begin
    return ((A.Seg1.Id = A.Seg2.Id and A.Seg1.Strand = A.Seg2.Strand) 
            and then
            (A.Seg1.Start = A.Seg2.Start and  A.Seg1.Stop  = A.Seg2.Stop));
  end Is_Auto_Recover;

--------------------------------------------------------------------------
-- Dtermination de l'origine commune ventuelle de deux segments       --
--------------------------------------------------------------------------
  function Common_Origin(Seg1,Seg2 : in Segment_Type;
                         Max_Dist  : in Integer) return boolean is
  begin
    if Seg1.Id = Seg2.Id and Seg1.Strand = Seg2.Strand then
      return abs(Seg1.Start - Seg2.Start) < Max_Dist;
    end if;
    return False;
  end Common_Origin;

--------------------------------------------------------------------------
-- Concatnation de deux segments                                       --
--------------------------------------------------------------------------
  procedure Concatene(A : in out Segment_Pair; B : in Segment_Pair);
  pragma Inline(Concatene);
  procedure Concatene(A : in out Segment_Pair; B : in Segment_Pair) is
    Len1,Len2 : Integer;
  begin
    if not((A.Seg1.Id = B.Seg1.Id and A.Seg2.Id = B.Seg2.Id) and 
           (A.Seg1.Strand = B.Seg1.Strand and A.Seg2.Strand = B.Seg2.Strand)) then
      raise ERROR_CONCATENE;
    end if;
    -- Seg1.start
    if A.Seg1.Start > B.Seg1.Start then
      A.Seg1.Start := B.Seg1.Start;
    end if;
    -- Seg2.start
    if A.Seg2.Start > B.Seg2.Start then
      A.Seg2.Start := B.Seg2.Start;
    end if;
    -- Seg1.stop
    if A.Seg1.Stop < B.Seg1.Stop then
      A.Seg1.Stop := B.Seg1.Stop;
    end if;
    -- Seg2.stop
    if A.Seg2.Stop < B.Seg2.Stop then
      A.Seg2.Stop := B.Seg2.Stop;
    end if;
    -- La longueur
    Len1 := A.Seg1.Stop - A.Seg1.Start + 1;
    Len2 := A.Seg2.Stop - A.Seg2.Start + 1;
    if Len1 > Len2 then
      A.Length := Len1;
    else
      A.Length := Len2;
    end if;
    if A.Distance < B.Distance then
      A.Distance := B.Distance;
    end if;
  end Concatene;

--------------------------------------------------------------------------
-- Ajout d'une paire dans l'arbre                                       --
--------------------------------------------------------------------------
  procedure Append_Pair_In_Tree(Similar_Regions  : in out Similar_Regions_Tree;
                                Pair             : in Segment_Pair;
                                Intersection_Marge : in Integer) is
    Ptr : Similar_Regions_Ptr;
    Trace : constant Boolean := False;
    New_Pair : Segment_Pair := Pair;
  begin
    loop
      -- Verifier si le segment a une intersection valide avec un segment
      -- deja existant (l'intersection doit etre sur les deux brins)
      Ptr := Find_Intersection(New_Pair.Seg1,New_Pair.Seg2,Similar_Regions,
                               Intersection_Marge);
      if Ptr = Null_Ptr then
        -- Pas de chevauchement, enregistrer le couple
        Insert(Similar_Regions.Tree,New_Pair);
        -- Mettre a jour la taille du plus grand segment
        if New_Pair.Length > Similar_Regions.Max_Seg_Length then
          Similar_Regions.Max_Seg_Length := New_Pair.Length;
        end if;
        -- Afficher le resultat a l'ecran
        if Trace then
          Put(">> ");
          Put(SRRBT.Count(Similar_Regions.Tree),8);
          Put(Integer(New_Pair.Seg1.Id),3);
          if New_Pair.Seg1.Strand = Watson then 
            put(" W"); 
          else 
            put(" C");
          end if; 
          Put(New_Pair.Seg1.Start,10);
          Put(New_Pair.Seg1.Stop,10);
          Put(Integer(New_Pair.Seg2.Id),3);
          if New_Pair.Seg2.Strand = Watson then 
            put(" W"); 
          else 
            put(" C");
          end if; 
          Put(New_Pair.Seg2.Start,10);
          Put(New_Pair.Seg2.Stop,10);
          Put(New_Pair.Length,8);
          Put(Sqrt(Real(New_Pair.Distance)),6,3,0);
          New_Line;
        end if;
        exit;
      end if;
      -- associer les deux elements
      Concatene(New_Pair,Consult(Ptr));
      -- retirer de l'arbre l'element pointe
      Delete(Similar_Regions.Tree,Ptr);
      -- et reessayer de le rajouter
    end loop;
  end Append_Pair_In_Tree;

--------------------------------------------------------------------------
-- Comptabilisation du nombre de paires enregistres dans l'arbre       --
--------------------------------------------------------------------------
  function Count_Pairs_In_Tree(Similar_Regions : in Similar_Regions_Tree)
    return Natural is
  begin
    return SRRBT.Count(Similar_Regions.Tree);
  end Count_Pairs_In_Tree;

--------------------------------------------------------------------------
-- Gestion d'un tableau regroupant toutes les paires de la base         --
--------------------------------------------------------------------------
  procedure Get_Entries(Similar_Regions  : in Similar_Regions_Tree;
                        Entries          : in out Similar_Regions_Entries) is
  begin
    SRRBT.Get_Entries(Similar_Regions.Tree,Entries);
  end Get_Entries;

  procedure Free_Entries(Entries : in out Similar_Regions_Entries) is
  begin
    Free(Entries);
  end Free_Entries;

  function Get_Pair(Entries : in Similar_Regions_Entries;
                    N       : in Integer) return Segment_Pair is
  begin
    if N < Entries'First or N > Entries'Last  then
      raise OUT_OF_ENTRIES;
    end if;
    return Consult(Entries(N));
  end Get_Pair;

  procedure Get_Pairs(Entries        : in Similar_Regions_Entries;
                      From           : in out Integer;
                      Minimal_Length : in Integer;
                      Pairs          : out Segment_Pairs;
                      Last           : out Integer) is
    I : Integer := Pairs'First;
  begin
    if Entries'Length = 0 or From < Entries'First or From > Entries'Last  then
      raise OUT_OF_ENTRIES;
    end if;
    loop
      if From > Entries'Last then
        -- On a atteint la fin du tableau des entres
        Last := I - 1;
        return;
      elsif Consult(Entries(From)).Length >= Minimal_Length then
        -- La longueur vrifie la condition
        exit when I > Pairs'Last; -- Le tableau est plein
        Pairs(I) := Consult(Entries(From));
        I := I + 1;
      end if;
      From := From + 1;
    end loop;
    Last := Pairs'Last;
  end Get_Pairs;
  
end Assirc_Database_Functions;  
