--------------------------------------------------------------------------
-- Assirc_Database_Functions                Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 23/03/1998 --
--------------------------------------------------------------------------
-- Gestion de la base de donnes des fragments similaires
-- Note: ce paquetage a t cr lors de la ralisation de la version
-- distribue dans le but de sparer la gnration des fragments et
-- leur archivage.
--------------------------------------------------------------------------
with Type_Specifications; use Type_Specifications;

with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Database; use Assirc_Specifications.Database;

package Assirc_Database_Functions is

  OUT_OF_ENTRIES : exception;
  -- Exception lev si l'on cherche  accder  un lment hors borne
  -- du tableau

  function Included(Seg1,Seg2       : in Segment_Type;
                    Similar_Regions : in Similar_Regions_Tree;
                    Excluded_Marge  : in Integer) return Boolean;
  pragma Inline(Included);
  -- Test de la prsence d'une graine dans l'arbre.
  -- Le parametre Excluded_Marge indique le pourcentage de la zone commune
  -- que l'on ne considere pas pour dcider que la nouvelle paire
  -- est incluse dans la base.

  procedure Append_Pair_In_Tree(Similar_Regions  : in out Similar_Regions_Tree;
                                Pair             : in Segment_Pair;
                                Intersection_Marge : in Integer);
  pragma Inline(Append_Pair_In_Tree);
  -- Ajout d'une paire dans l'arbre
  -- Le parametre Intersection_Marge indique le nombre de bases maximal
  -- tolr entre la nouvelle paire et une paire de la base pour que
  -- l'on considre qu'il s'agit d'une mme similarit

  function Count_Pairs_In_Tree(Similar_Regions : in Similar_Regions_Tree)
    return Natural;
  pragma Inline(Count_Pairs_In_Tree);
  -- Obtention du nombre de paires enregistres dans l'arbre

  procedure Get_Entries(Similar_Regions : in Similar_Regions_Tree;
                        Entries         : in out Similar_Regions_Entries);
  pragma Inline(Get_Entries);
  -- Cration d'un tableau contenant un accs  toutes les paires de
  -- la base. ce tableau n'est valable que tant qu'il n'y a aucun lment
  -- ajout dans la base

  procedure Free_Entries(Entries : in out Similar_Regions_Entries);
  pragma Inline(Free_Entries);
  -- Destruction du tableau prcdent

  function Get_Pair(Entries : in Similar_Regions_Entries;
                    N       : in Integer) return Segment_Pair;
  pragma Inline(Get_Pair);
  -- Obtention de la n ime paire du tableau. Si la valeur de N
  -- est en dehors des bornes autorises, alors OUT_OF_ENTRIES est
  -- lev. Il n'y a pas de contrle sur la taille de la paire.


  procedure Get_Pairs(Entries        : in Similar_Regions_Entries;
                      From           : in out Integer;
                      Minimal_Length : in Integer;
                      Pairs          : out Segment_Pairs;
                      Last           : out Integer);
  pragma Inline(Get_Pairs);
  -- Obtention de Pairs'Length paires  partir de la from ime paire du tableau
  -- vrifiant la condition d'tre de taille suprieure  Minimal_Length.
  -- Last donne l'indice dans le tableau Pairs de la dernire paire
  -- mmorise. Une valeur infrieure  Pairs'First indique qu'il n'y avait
  -- aucun lment disponible. Une valeur gale  Pairs'Last indique que le
  -- tableau est plein. Si la valeur du From initial est en dehors des bornes 
  -- autorises, alors l'exception OUT_OF_ENTRIES est leve. En entre, From
  -- prend la valeur du premier indice  tester. Il renvoie en sortie la
  -- valeur suivante de l'indice du tableau Entries  tester.

end Assirc_Database_Functions;
