--------------------------------------------------------------------------
-- Assirc_Error                             Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 21/05/1996 --
-- Pierre VINCENS                   V 01.000                 29/04/1998 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;

with Type_specifications; use Type_Specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use Strlib;
with Strlib.IO; use Strlib.IO;

with Sequence_Specifications; use Sequence_Specifications;
--with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Io; use Sequence_Io;

with Assirc_Options; use Assirc_Options;
--with Assirc_Specifications; use Assirc_Specifications;
--with Assirc_Specifications.Basic; use Assirc_Specifications.Basic;
--with Assirc_Specifications.Database; use Assirc_Specifications.Database;

with Assirc_Options; use Assirc_Options;

package body Assirc_Exceptions is

  procedure Check_Exception(The_Error : in Exception_Occurrence) is
    Word1,Word2 : Varstring;
    Current,Last : Natural;

    procedure Read_First_Word is
    begin
      Current := Exception_Message(The_Error)'First;
      Next_Word(Exception_Message(The_Error),Word1,Current," ");
    end Read_First_Word;

    procedure Read_Second_Word is
    begin
      Last := Exception_Message(The_Error)'Last;
      Next_Word(Exception_Message(The_Error)(Current..Last),
		Word2,Current," ");
    end Read_Second_Word;

    procedure Read_Two_Words is
    begin
      Read_First_Word;
      Read_Second_Word;
    end Read_Two_Words;

  begin
    -- Dans le cas o l'exception n'a pas de message associ, on affiche
    -- simplement le nom de l'erreur
    if Exception_Message(The_Error)'Length = 0 then
      Put_Line("Raised " & Exception_Name(The_Error));
      return;
    end if;
    -- Sinon on traite en focntion de l'erreur
    Put_Line("Program interrupted!"); 
    if Exception_Identity(The_Error) = Name_Error'Identity then
      -- Erreur d'ouverture d'un fichier
      Read_Two_Words;
      Put_Line("The " & Word1 & " file called " &
	       Word2 & " has not been found. Please check.");
    elsif Exception_Identity(The_Error) = Sequence_Io_Error'Identity then
      -- Erreur dans le contenu d'un fichier
      Read_First_Word;
      Put_Line("The file called " & Word1 & 
		" can not be read correctly. Please check.");
    elsif Exception_Identity(The_Error) = Invalid_Data'Identity then
      -- Erreur dans le nom du format donnee
      Read_First_Word;
      if To_String(Word1) = "F_Start" then
	Read_Second_Word;
	Put_Line("The start position in line " & Word2 & 
		 " is invalid. Please check.");
      elsif To_String(Word1) = "F_Stop" then
	Read_Second_Word;
	Put_Line("The stop position in line " & Word2 & 
		 " is invalid. Please check.");
      elsif To_String(Word1) = "F_Direction" then
	Read_Second_Word;
	Put_Line("The specifications of direction in line " & Word2 & 
		 " are invalid. Please check.");
      elsif To_String(Word1) = "O_Start" then
	Read_Second_Word;
	Put_Line("The start position for sequence " & Word2 &
		 " is invalid. Please check.");
      elsif To_String(Word1) = "O_Stop" then
	Read_Second_Word;
	Put_Line("The stop position for sequence " & Word2 &
		 " is invalid. Please check.");
      elsif To_String(Word1) = "O_Seed" then
	Put_Line("The seed option is invalid. Please check.");
      elsif To_String(Word1) = "O_Mask" then
	Put_Line("The mask option is invalid. Please check.");
      elsif To_String(Word1) = "O_FSimilarity" then
	Put("The option specifying minimal value of similarity is ");
	Put_Line("invalid. Please check.");
      elsif To_String(Word1) = "O_FLength" then
	Put_Line("The final length option is invalid. Please check.");
      elsif To_String(Word1) = "O_RLength" then
	Put_Line("The recorded length option is invalid. Please check.");
      elsif To_String(Word1) = "O_Distance" then
	Put_Line("The distance option is invalid. Please check.");
      elsif To_String(Word1) = "O_Origin" then
	Put_Line("The origin option is invalid. Please check.");
      elsif To_String(Word1) = "O_Excluded" then
	Put_Line("The excluded marge option is invalid. Please check.");
      elsif To_String(Word1) = "O_Intersection" then
	Put_Line("The intersection marge option is invalid. Please check.");
      elsif To_String(Word1) = "O_Tolerance" then
	Put_Line("The tolerance option is invalid. Please check.");
      elsif To_String(Word1) = "O_Skip" then
	Put_Line("The skip option is invalid. Please check.");
      elsif To_String(Word1) = "O_Shift" then
	Put_Line("The shift option is invalid. Please check.");
      elsif To_String(Word1) = "O_Align_Method" then
	Put_Line("The method for alignment is invalid. Please check.");
      elsif To_String(Word1) = "O_Match_Weight" then
	Put_Line("The match weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Mismatch_Weight" then
	Put_Line("The mismatch weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Gap_Weight" then
	Put_Line("The gap weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Gap_Length_Weight" then
	Put_Line("The gap length weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Similarity_Threshold" then
	Put_Line("The similarity threshold option is invalid. Please check.");
      elsif To_String(Word1) = "O_Randomization" then
	Put_Line("The randomization option is invalid. Please check.");
      elsif To_String(Word1) = "O_Preserve" then
	Put_Line("The preserve option is invalid. Please check.");
      elsif To_String(Word1) = "O_Length_Max" then
	Put_Line("The length max option is invalid. Please check.");
      elsif To_String(Word1) = "O_Adjust" then
	Put_Line("The adjust option is invalid. Please check.");
      elsif To_String(Word1) = "O_Min_Dist_Seeds" then
	Put_Line("The minimal distance between seeds (SDMI option) is invalid.");
	Put_Line("Please check.");
      elsif To_String(Word1) = "O_Max_Dist_Seeds" then
	Put_Line("The maximal distance between seeds (SDMA option) is invalid.");
	Put_Line("Please check.");
      end if;
    elsif Exception_Identity(The_Error) = Invalid_Sequence_Format'Identity then
      -- Erreur dans le nom du format donnee
      Read_First_Word;
      if To_String(Word1) = "L" then
	Read_Second_Word;
	Put_Line("The sequence format specified in line " & Word2 & 
		 " is invalid. Please check.");
      else
	Put_Line("The sequence format specified for option " & Word1 &
		 " is invalid. Please check.");
      end if;
    elsif Exception_Identity(The_Error) = Invalid_Strand'Identity then
      -- Erreur dans le nom du brin donnee
      Read_First_Word;
      if To_String(Word1) = "L" then
	Read_Second_Word;
	Put_Line("The strand specified in line " & Word2 & 
		 " is invalid. Please check.");
      else
	Put_Line("The strand specified for sequence " & Word1 &
		 " is invalid. Please check.");
      end if;
    else
      -- Erreur indeterminee
      Put_Line("Raised " & Exception_Information(The_Error));
    end if;
  end Check_Exception;

end Assirc_Exceptions;
