--------------------------------------------------------------------------
-- Assirc_Options                           Copyrigth ENS ULM 1996-1997 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 24/07/1997 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Exceptions; use Ada.Exceptions;

with Options_Util; use Options_Util;

with Sequence_Alignment; use Sequence_Alignment;

with Assirc_Version; use Assirc_Version;

package body Assirc_Options is

  procedure Get_Options(Options : in out Options_Type) is
    Count : Positive := 1;
    type Options_List is (Mismatch_Weight);
    type Options_Controlled_Array is array (Options_List) of Boolean;
    Options_Controlled : Options_Controlled_Array := (others => False);
  begin
    if Argument_Count = 0 then
      raise ARGUMENT_ERROR;
    end if;
    while Count <= Argument_Count loop
      if Is_An_Alias(Argument(Count),"-SEED",3) then
        -- lire la taille des graines a generer
        Count := Count + 1;
	begin
	  Options.Seed_Size := Opt_Int(Argument(Count));
	  if Options.Seed_Size < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Seed");
	end;
      elsif Is_An_Alias(Argument(Count),"-MASK",3) then
        -- lire la taille des Seeds de selection
        Count := Count + 1;
	begin
	  Options.Mask_Size := Opt_Int(Argument(Count));
	  if Options.Mask_Size < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Mask");
	end;  
      elsif Is_An_Alias(Argument(Count),"-SIMILARITY",3) then
        -- lire le seuil minimal de similitude de similitude
        Count := Count + 1;
	begin
	  Options.Min_Similarity := Opt_Int(Argument(Count));
	  if Options.Min_Similarity < 0 then raise Data_Error; end if;
	  if Options.Min_Similarity > 100 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_FSimilarity");
	end; 
      elsif Is_An_Alias(Argument(Count),"-FLENGTH",3) or
            Is_An_Alias(Argument(Count),"-FINAL_LENGTH",3)then
        -- lire la longueur minimale de similitude
        Count := Count + 1;
	begin
	  Options.Final_Length := Opt_Int(Argument(Count));
	  if Options.Final_Length < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_FLength");
	end; 
      elsif Is_An_Alias(Argument(Count),"-RLENGTH",3) or
            Is_An_Alias(Argument(Count),"-RECORDED_LENGTH",3) then
        -- lire la longueur de similitude minimale utilisee pour
        -- enregistrer une sequence dans la base
        Count := Count + 1;
	begin
	  Options.Recorded_Length := Opt_Int(Argument(Count));
	  if Options.Recorded_Length < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_RLength");
	end;  
      elsif Is_An_Alias(Argument(Count),"-DISTANCE",2) then
        -- lire la distance maximale pour un segment
        Count := Count + 1;
	begin
	  Options.Distance := Opt_Int(Argument(Count));
	  if Options.Distance < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Distance");
	end;
      elsif Is_An_Alias(Argument(Count),"-ORIGIN",3) then
        -- lire la distance maximale pour definir que deux segments ont
        -- une origine commune
        Count := Count + 1;
	begin
	  Options.Common_Max_Dist := Opt_Int(Argument(Count));
	  if Options.Common_Max_Dist < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Origin");
	end;
      elsif Is_An_Alias(Argument(Count),"-EXCLUDED",3) then
        -- Zone a exclure pour definir si un segment est
        -- inclus dans un autre
        Count := Count + 1;
	begin
	  Options.Excluded_Marge := Opt_Int(Argument(Count));
	  if Options.Excluded_Marge < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Excluded");
	end;  
      elsif Is_An_Alias(Argument(Count),"-INTERSECTION",3) then
        -- Nombre de nucleotides ajoutes virtuellement a droite et
        -- a gauche du segment pour considerer si le segment teste
        -- chevauche un segment de la base
        Count := Count + 1;
	begin
	  Options.Intersection_Marge := Opt_Int(Argument(Count));
	  if Options.Intersection_Marge < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Intersection");
	end;
      elsif Is_An_Alias(Argument(Count),"-TOLERANCE",2) then
        -- lire la tolerance de franchissement du seuil
        Count := Count + 1;
	begin
	  Options.Tolerance := Opt_Int(Argument(Count));
	  if Options.Tolerance < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Tolerance");
	end;
      elsif Is_An_Alias(Argument(Count),"-SKIP",3) then
        -- lire le saut fait avant de realiser une nouvelle valuation
        -- de la distance entre les 2 marches alatoires.
        Count := Count + 1;
	begin
	  Options.Skip := Opt_Int(Argument(Count));
	  if Options.Skip < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Skip");
	end;
      elsif Is_An_Alias(Argument(Count),"-SDMI",5) then
        -- lire la distance minimale qui doit exister entre les deux segments
	-- pour la recherche des ltr (Ltr_Assirc)
        Count := Count + 1;
	begin
	  Options.Min_Dist_Seeds := Opt_Int(Argument(Count));
	  if Options.Min_Dist_Seeds < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Min_Dist_Seeds");
	end;
      elsif Is_An_Alias(Argument(Count),"-SDMA",5) then
        -- lire la distance maximale qui doit exister entre les deux segments
	-- pour la recherche des ltr (Ltr_Assirc)
        Count := Count + 1;
	begin
	  Options.Max_Dist_Seeds := Opt_Int(Argument(Count));
	  if Options.Max_Dist_Seeds < 0 then raise Data_Error; end if;
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Max_Dist_Seeds");
	end;
      elsif Is_An_Alias(Argument(Count),"-OUTPUT",3) then
        -- lire le nom du fichier ou seront memorise les resultats
        Count := Count + 1;
        Copy(Argument(count),Options.Output);
      elsif Is_An_Alias(Argument(Count),"-LIST",3) then
        -- lire le nom du fichier contenant les noms des sequences
        Count := Count + 1;
        Copy(Argument(count),Options.Name);
      elsif Is_An_Alias(Argument(Count),"-S1",3) then
        -- lire le nom de la sequence 1
        Count := Count + 1;
        Copy(Argument(count),Options.Name1);
      elsif Is_An_Alias(Argument(Count),"-B1",3) then
        -- lire la position de debut dans la sequence 1
        Count := Count + 1;
	begin
	  Options.Start1 := Opt_Int(Argument(Count));
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Start 1");
	end;
      elsif Is_An_Alias(Argument(Count),"-E1",3) then
        -- lire la position de fin dans la sequence 1
        Count := Count + 1;
	begin
	  Options.Stop1 := Opt_Int(Argument(Count));
	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Stop 1");
	end;
      elsif Is_An_Alias(Argument(Count),"-ST1",4) then
        -- indique sur quel brin doit porter l'analyse
        Count := Count + 1;
	begin
	  Options.Strand1 := Opt_Sequence_Strand(Argument(Count));
	exception
	  when Invalid_Strand => 
	    Raise_Exception(Invalid_Strand'Identity,"1");
	end;
      elsif Is_An_Alias(Argument(Count),"-R1",3) then
        -- indique que l'on inverse la sequence 1
        Options.Opposite1 := True;
      elsif Is_An_Alias(Argument(Count),"-C1",3) then
        -- indique que l'on prend la sequence complementaire
        Options.Complementary1 := True;
      elsif Is_An_Alias(Argument(Count),"-I1",3) then
        -- specifier l'identificateur de la sequence 1
        Count := Count + 1;
        Copy(Argument(count),Options.Identifier1);
      elsif Is_An_Alias(Argument(Count),"-F1",3) then
        -- definir le format de lecture de la sequence 1
        Count := Count + 1;
	begin
	  Options.Format1 := Opt_Sequence_Format(Argument(Count));
	exception
	   when Invalid_Sequence_Format =>
	     Raise_Exception(Invalid_Sequence_Format'Identity,"F1");
	end;
      elsif Is_An_Alias(Argument(Count),"-S2",3) then
        -- lire le nom de la sequence 2
        Options.Nb_Sequences := 2;
        Count := Count + 1;
        Copy(Argument(count),Options.Name2);
      elsif Is_An_Alias(Argument(Count),"-B2",3) then
        -- lire la position de debut dans la sequence 2
        Count := Count + 1;
	begin
	  Options.Start2 := Opt_Int(Argument(Count));
      	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Start 2");
	end;
      elsif Is_An_Alias(Argument(Count),"-E2",3) then
        -- lire la position de fin dans la sequence 2
        Count := Count + 1;
	begin
	  Options.Stop2 := Opt_Int(Argument(Count));
       	exception
	  when Data_Error => 
	    Raise_Exception(Invalid_Data'Identity,"O_Stop 2");
	end;
      elsif Is_An_Alias(Argument(Count),"-ST2",4) then
        -- indique sur quel brin doit porter l'analyse
        Count := Count + 1;
	begin
	  Options.Strand2 := Opt_Sequence_Strand(Argument(Count));
	exception
	  when Invalid_Strand =>
	    Raise_Exception(Invalid_Strand'Identity,"2");
	end;
      elsif Is_An_Alias(Argument(Count),"-R2",3) then
        -- indique que l'on inverse la sequence 2
         Options.Opposite2 := True;
      elsif Is_An_Alias(Argument(Count),"-C2",3) then
        -- indique que l'on prend la sequence complementaire 2
         Options.Complementary2 := True;
      elsif Is_An_Alias(Argument(Count),"-F2",3) then
        -- definir le format de lecture de la sequence 2
        Count := Count + 1;
	begin
	  Options.Format2 := Opt_Sequence_Format(Argument(Count));
	exception
	   when Invalid_Sequence_Format =>
	     Raise_Exception(Invalid_Sequence_Format'Identity,"F2");
	end;
      elsif Is_An_Alias(Argument(Count),"-I2",3) then
        -- specifier l'identificateur de la sequence 2
        Count := Count + 1;
        Copy(Argument(count),Options.Identifier2);
      elsif Is_An_Alias(Argument(Count),"-RPM",4) then
        -- specifier une distribution par dcoupage
	-- automatique si longueur ou overlap spcifis
        Count := Count + 1;
        Options.Regions_Mode := True;
      elsif Is_An_Alias(Argument(Count),"-RPL",4) then
        -- specifier la longueur d'une rgion en distribution par dcoupage
        Count := Count + 1;
        Options.Regions_Length := Opt_Int(Argument(Count));
	    Options.Regions_Mode := True;
      elsif Is_An_Alias(Argument(Count),"-RPO",4) then
        -- specifier la longueur de recouvrement en distribution par dcoupage
        Count := Count + 1;
        Options.Regions_Overlap := Opt_Int(Argument(Count));
      	Options.Regions_Mode := True;
      elsif Is_An_Alias(Argument(Count),"-XONLY",2) then
        -- precise que l'on souhaite seulement un test entre seq 1 et 2
        Options.Xonly := True;
      elsif Is_An_Alias(Argument(Count),"-REPRESENTATION",3) then
        -- precise la representation souhaitee
        Count := Count + 1;
	begin
	   Options.Shift := Shift_Model(Opt_Int(Argument(Count)));
	exception
	  when Data_Error =>
	    -- l'argument donn n'est pas un nombre. On considere donc
	    -- qu'il s'agit d'un nom de fichier
	    Copy(Argument(count),Options.Shift_Name);
	  when Constraint_Error =>
	     Raise_Exception(Invalid_Data'Identity,"O_Shift");
	end;
      elsif Is_An_Alias(Argument(Count),"-ALIGNMENT",3) then
        -- lire la mthode d'alignement
        Count := Count + 1;
        if Is_An_Alias(Argument(Count),"No",1) then
	  Options.Alignment := False;
	elsif Is_An_Alias(Argument(Count),"Global",1) then
	  Options.Alignment_Method := Global_Alignment;
	elsif Is_An_Alias(Argument(Count),"Local",1) then
	  Options.Alignment_Method := Local_Alignment;
	else
	  Raise_Exception(Invalid_Data'Identity,"O_Align_Method");
	end if;
      elsif Is_An_Alias(Argument(Count),"-MATCH_WEIGHT",4)  or
	Is_An_Alias(Argument(Count),"-MW",3) then
        -- Poids affect  un match
        Count := Count + 1;
	begin
	  Options.Match_Weight := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Match_Weight");
        end;
      elsif Is_An_Alias(Argument(Count),"-MISMATCH_WEIGHT",3) then
        -- Poids affect  un mismatch
        Count := Count + 1;
	begin
	  Options_Controlled(Mismatch_Weight) := True;
	  Options.Mismatch_Weight := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Mismatch_Weight");
        end;
      elsif Is_An_Alias(Argument(Count),"-GAP_WEIGHT",6) or
	Is_An_Alias(Argument(Count),"-GW",3) then
        -- Poids affect  la cration dn nouveau gap
        Count := Count + 1;
	begin
	  Options.Gap_Weight := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Gap_Weight");
        end;
      elsif Is_An_Alias(Argument(Count),"-GAP_LENGTH_WEIGHT",6) or
	Is_An_Alias(Argument(Count),"-LW",3) then
        -- Poids affect  l'ajout d'un lment dans un gap existant
        Count := Count + 1;
	begin
	  Options.Gap_Length_Weight := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Gap_Length_Weight");
        end;
      elsif Is_An_Alias(Argument(Count),"-SIMILARITY_THRESHOLD",13) or
	Is_An_Alias(Argument(Count),"-SIT",4) then
        -- Seuil de similitude entre deux rsidus
        Count := Count + 1;
	begin
	  Options.Similarity_Threshold := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Similarity_Threshold");
        end;
      elsif Is_An_Alias(Argument(Count),"-RANDomization",5) or
	Is_An_Alias(Argument(Count),"-RND",4) then
	-- Pour dterminer le seuil de signification d'un alignement
        Count := Count + 1;
	begin
	  Options.Randomization := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Randomization");
        end;
      elsif Is_An_Alias(Argument(Count),"-PREserve",4) then
	-- Dans le cas de squences nucliques pour dfinir si l'on
	-- prserve la frquence nuclotidique, di, tri-nuclotidique,... 
        Count := Count + 1;
	begin
	  Options.Preserve := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Preserve");
        end;
      elsif Is_An_Alias(Argument(Count),"-END_WEIGHT",7) or
	Is_An_Alias(Argument(Count),"-EW",3)then
        -- indique que l'on doit tenir compte des gaps extrmits
	-- dans l'alignement global
        Options.With_End_Weight := True;
      elsif Is_An_Alias(Argument(Count),"-LENGTHMAX",3) then
        -- indique que l'on passe une liste de position a traiter
        Count := Count + 1;
	begin
	  Options.Max_Length := Opt_Int(Argument(Count));
	  if Options.Max_Length < 0 then raise Data_Error; end if; 
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Length_Max");
        end;
      elsif Is_An_Alias(Argument(Count),"-ADJUST",2) then
        -- indique que l'on veut tendre la zone  aligner au del
        -- des valeurs passes
        Count := Count + 1;
	begin
	  Options.Adjust := Opt_Int(Argument(Count));
	exception
          when Data_Error => 
            Raise_Exception(Invalid_Data'Identity,"O_Adjust");
        end;
      elsif Is_An_Alias(Argument(Count),"-VERBOSE",2) then
        -- pour obtenir le numero de version
        Put_Line("Assirc version " & Version_Id & " - Copyright ENS Ulm");
	Options.Verbose := True;
      elsif Is_An_Alias(Argument(Count),"-NOSUMMARY",2) then
        -- pour ne pas afficher les paires alignees au fur et a mesure
        Options.Summary := False;
      elsif Is_An_Alias(Argument(Count),"--BOOT_SERVER",13) then
        -- Cette option est utilise par Glade en version distribue
        -- pour prciser le numro du port de la socket
        -- On l'ignore ainsi que la valeur de la socket donne en argument 
	-- suivant
        Count := Count + 1;
      elsif Argument(Count)'Length >= 2 and then
	 Argument(Count)(Argument(Count)'First..Argument(Count)'First+1) = "-@"
      then
        -- On ignore les options qui dbutent par 2 tirets
        -- elles sont utilises par Glade en version distribue
        null;
      else
        Put_Line("Invalid options value: " & Argument(Count));
	raise ARGUMENT_ERROR;
      end if;
      Count := Count + 1;
    end loop;

    -- On vrifie si au moins un nom a t donne
    if Length(Options.Name) = 0 and 
      Length(Options.Name1) = 0 and Length(Options.Name1) = 0 then
      Put_Line("You haven't specified the names of the files, where are");
      Put_Line("registered the sequences.");
      New_Line;
      raise ARGUMENT_ERROR;
    end if;
    -- Si un seul nom de squences, on le considre pour nom 1 et nom2
    if Length(Options.Name2) = 0 then
      Copy(Root_Name(Options.Name1),Options.Name2);
      Options.Format2 := Options.Format1;
    elsif Length(Options.Name1) = 0 then
      Copy(Root_Name(Options.Name2),Options.Name1);
      Options.Format1 := Options.Format2;
    end if;
    -- Si pas de nom d'identificateur, on le dduit du nom de la squence
    if Length(Options.Identifier1) = 0 then
      Copy(Root_Name(Options.Name1),Options.Identifier1);
    end if;
    if Length(Options.Identifier2) = 0 then
      Copy(Root_Name(Options.Name2),Options.Identifier2);
    end if;
    
    -- Test de validit de quelques options
    if Options.Mask_Size >= Options.Seed_Size then
      Options.Mask_Size := 0;
    end if;
    if Options.Recorded_Length <= 0 or
      Options.Recorded_Length > Options.Final_Length then
      Options.Recorded_Length := Options.Final_Length;
    end if;
    if Options.Regions_Overlap >= Options.Regions_Length then
      raise ARGUMENT_ERROR;
    end if;

    if not Options_Controlled(Mismatch_Weight) then
      -- La valeur par dfaut dpend de la mthode
      case Options.Alignment_Method is
        when Global_Alignment => Options.Mismatch_Weight :=  0;
        when Local_Alignment  => Options.Mismatch_Weight := -9;
      end case;
    end if;

  end Get_Options;

  procedure Print_Options is
  begin
    Put_Line("ASSIRC: argument error");
    Put_Line("    syntax is: assirc  ");
    Put_Line("                      [-SEed <size>]");
    Put_Line("                      [-MAsk <size>]");
    Put_Line("                      [-REPResentation <0|1|2|name>]");
    Put_Line("                      [-Distance <max_value>]");
    Put_Line("                      [-SImilarity <value>]");
    Put_Line("                      [-FLength <value>]");
    Put_Line("                      [-RLength <value>]");
    Put_Line("                      [-Tolerance <value>]");
    Put_Line("                      [-SKip <value>]");
    Put_Line("                      [-EXcluded <value>]");
    Put_Line("                      [-Intersection <value>]");
    Put_Line("                      [-ORigin <value>]");
    Put_Line("                      [-ALign <Global|Local|No>]");
    Put_Line("                      [-MATch_weight value | [-MW value]]");
    Put_Line("                      [-MIsmatch_weight value]");
    Put_Line("                      [-GAP_Weight value] | [-GW value]");
    Put_Line("                      [-GAP_Length_weight value] | [-LW value]");
    Put_Line("                      [-SIMILARITY_Threshold value] | [-SIT value]");
    Put_Line("                      [-RANDomization value] | [-RND value]");
    Put_Line("                      [-PREserve value]");
    Put_Line("                      [-END_WEight] | [-EW]");
    Put_Line("                      [-ADjust value]");
    Put_Line("                      [-LEngthmax value]");
    Put_Line("                      [-B1 <start_position>]");
    Put_Line("                      [-E1 <stop_position>]");
    Put_Line("                      [-B2 <start_position>]");
    Put_Line("                      [-E2 <stop_position>]");
    Put_Line("                      [-I1 <string> ]");
    Put_Line("                      [-I2 <string> ]");
    Put_Line("                      [-ST1 <Watson|Crick|Both>]");
    Put_Line("                      [-ST2 <Watson|Crick|Both>]");
    Put_Line("                      [-F1 <Ascii|GCg|GEnbank|Strider>]");
    Put_Line("                      [-F2 <Ascii|GCg|GEnbank|Strider>]");
    Put_Line("                      [-R1]");
    Put_Line("                      [-R2]");
    Put_Line("                      [-C1]");
    Put_Line("                      [-C2]");
    Put_Line("                      [-S1 sequence_file]");
    Put_Line("                      [-S2 sequence_file]");
    Put_Line("                      [-LIst sequence_list]");
    Put_Line("                      [-Nosummary]");
    Put_Line("                      [-Verbose]");
    -- Ltr_assirc options
    Put_Line("                      [-SDMI value]");
    Put_Line("                      [-SDMA value]");
    -- Dassirc options
    Put_Line("                      [-RPM]");
    Put_Line("                      [-RPL <length>]");
    Put_Line("                      [-RPO <overlap>]");
    New_Line;
  end Print_Options;

end Assirc_Options;
