--------------------------------------------------------------------------
-- Assirc_Read                              Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 16/04/1998 --
--------------------------------------------------------------------------
with Ada.Text_Io; use  Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;

with Type_Specifications.IO; use Type_Specifications.IO;
with Options_Util; use Options_Util;
with Strlib.IO; use Strlib.IO;

with Sequence_Io; use Sequence_Io;
with Sequence_Alignment; use Sequence_Alignment;

with Assirc_Options; use Assirc_Options;

package body Assirc_Read is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
   
--------------------------------------------------------------------------
-- Lecture d'une squence                                               --
--------------------------------------------------------------------------
  procedure Read_One_Sequence(Name          : in String;
                              Identifier    : in Varstring;
                              Id            : in Accession_Type;
                              Format        : in Sequence_Format;
                              Strand        : in Strand_Type;
                              Start, Stop   : in out Integer;
                              Sequence      : in out Assirc_Sequence_Data;
                              Opposite      : in Boolean := False;
                              Complementary : in Boolean := False;
			      Verbose       : in Boolean := False) is
  begin
    if Verbose then
      Put_Line("Reading: " & Name & "...");
    end if;
    Read_Sequence(Name,Sequence,DNA,Format,Internal);
    if Verbose then
      Put_Line("Sequence length is" & Integer'Image(Sequence.Length));
    end if;
    -- Prendre en compte si necessaire les positions de debut et de fin
    -- de sequence pour le trace et en verifier la validite
    if Start >= 0 and Start >= Sequence.Start then
      Sequence.Start := Start;
    else
      Start := Sequence.Start;
    end if;
    if Stop >= 0 and Stop <= Sequence.Stop then
      Sequence.Stop := Stop;
    else
      Stop := Sequence.Stop;
    end if;
    if Opposite then
      Reverse_Sequence(Sequence);
    end if;
    if Complementary then
      Cmp_Sequence(Sequence);
    end if;
    if Length(Identifier) > 8 then
      Sequence.Identifier := Substring(Identifier,1,8);
    else
      Copy(Identifier,Sequence.Identifier);
    end if;
    -- Declarer quel brin doit etre traiter
    Sequence.Strand := Strand;
    -- et identifier la squence
    Sequence.Id := Id;
  exception
    when Name_Error =>
      Raise_Exception(Name_Error'Identity,"sequence " & Name);
    when Sequence_Io_Error =>
      Raise_Exception(Sequence_Io_Error'Identity,Name);
  end Read_One_Sequence;

--------------------------------------------------------------------------
-- Lecture du fichier contenant les noms des squences                  --
--------------------------------------------------------------------------
  function Read_List_Of_Sequences(Name : in String;
				  Verbose : in Boolean := False) 
				  return Assirc_Sequence_Array_Access is
    -- Les lignes commencant par un # sont ignorees. Si une option n'est pas
    -- precisee, toutes les options suivantes sont prises par defaut.
    -- Le format de la ligne du fichier est:
    -- Nom_Sequence Format Brin Identificateur Debut Fin

    File            : Ada.Text_Io.File_Type;
    Count           : Accession_Type := 0;
    Line            : String(1..1024);
    Line_Number     : Natural := 0;
    Last, S_Last    : Natural;
    Current         : Natural;
    Word            : Varstring;
    Sequences       : Assirc_Sequence_Array_Access;
    S_Id            : Accession_Type := 0;
    S_Name          : Varstring;
    S_Identifier    : Varstring;
    S_Start,S_Stop  : Integer;
    S_Strand        : Strand_type;
    S_Model         : Sequence_Model;
    S_Format        : Sequence_Format;
    S_Opposite      : Boolean;
    S_Complementary : Boolean;

    procedure Ignore_Spaces is
      -- Incremente current jusqu'au premier caractre non blanc
      -- ou jusqu' la fin de la ligne
    begin
      while Line(Current) = ' ' loop
        Current := Current + 1;
        exit when Current > Last;
      end loop;
    end Ignore_Spaces;

  begin
    -- Compter le nombre de sequences a traiter
    begin
      Open(File,In_File,Name);
    exception
      when Name_Error =>
	Raise_Exception(Name_Error'Identity,"list " & Name);
    end;
    while not End_Of_File(File) loop
      -- Lire la ligne
      Get_Line(File,Line,Last);
      -- Verifier la validite de la ligne
      for i in 1..Last loop
        if Line(i) /= ' ' and line(i) /= '#' then
          -- Ce n'est pas une ligne de commentaire
          Count := Count + 1;
          exit;
        end if;
        exit when Line(i) = '#'; -- C'est une ligne de commentaire
      end loop;      
    end loop;
    Close(File);
    -- Reserver l'espace pour memoriser les sequences
    Sequences := new Assirc_Sequence_Array(Accession_Type(1)..Count);
    -- Lire effectivement les sequences
    Open(File,In_File,Name);
    while not End_Of_File(File) loop
      Line_Number := Line_Number + 1;
      -- Lire la ligne
      Get_Line(File,Line,Last);
      -- Ignorer les blancs initiaux
      Current := Line'First;
      Ignore_Spaces;
      -- Regarder s'il reste quelque chose sur la ligne
      if Current <= Last and Line(Current) /= '#' then
        -- On lit le nom du fichier contenant la squence (Obligatoire)
        Next_Word(Line(Current..Last),S_Name,Current," ");
        -- Definir le type de la sequence
	Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Model := NO;
	  else
	    S_Model := Opt_Sequence_Model(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Model := NO;
        end if;
        -- Definir le format de la sequence
	Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Format := Unknown;
	  else
	    S_Format := Opt_Sequence_Format(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Format := Unknown;
        end if;
        -- Definir le brin a analyser (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
          if Equal(Word,"-") then
	    S_Strand := Both;
	  else
	    S_Strand := Opt_Sequence_Strand(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Strand := Both;
        end if;
        -- Definir l'identificateur de la sequence (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),S_Identifier,Current," ");
	  if Equal(S_Identifier,"-") then
	    S_Identifier := To_Varstring(Root_Name(S_Name));
	  end if;
        else
          -- Ou en le deduisant du nom de la sequence
          S_Identifier := To_Varstring(Root_Name(S_Name));
        end if;
        -- Definir la position de debut (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- E la lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Start := Complete;
	  else
	    begin
	      Get(To_String(Word),S_Start,S_Last);
	    exception
	      when Data_Error => 
		Raise_Exception(Invalid_Data'Identity,
				"F_Start " & Natural'Image(Line_Number));
	    end;
	  end if;
        else
          -- Valeur par defaut
          S_Start := Complete;
        end if;
        -- Et la position de fin (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- E la lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Stop := Complete;
	  else
	    begin
	      Get(To_String(Word),S_Stop,S_Last);
	    exception
	      when Data_Error => 
		Raise_Exception(Invalid_Data'Identity,
				"F_Stop " & Natural'Image(Line_Number));
	    end;
	  end if;
        else
          -- Valeur par defaut
          S_Stop := Complete;
        end if;
	-- Indiquer si l'on doit prendre le reverse ou le complementaire
        Ignore_Spaces;
	S_Opposite := False;
	S_Complementary := False;
	while Current <= Last loop
	  -- on lit le mot suivant
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Is_An_Alias(To_String(Word),"REVERSE",1) then
	    S_Opposite := True;
	  elsif Is_An_Alias(To_String(Word),"COMPLEMENTARY",1) then
	    S_Complementary := True;
	  elsif Is_An_Alias(To_String(Word),"RC",2) then
	    S_Opposite := True;
	    S_Complementary := True;
	  elsif Equal(Word,"-") then
	    S_Opposite := False;
	    S_Complementary := False;
	  else
	    Raise_Exception(Invalid_Data'Identity,
			    "F_Direction " & Natural'Image(Line_Number));
	  end if;
	  Ignore_Spaces;
        end loop;
	-- Affichage des infos
	if Verbose then
	  Put_Line("Sequence to read:");
	  Put_Line("   Name       : " & To_String(S_Name));
	  Put_Line("   Type       : " & Sequence_Model'Image(S_Model));
	  Put_Line("   Format     : " & Sequence_Format'Image(S_Format));
	  Put_Line("   Strand     : " & Strand_Type'Image(S_Strand));
	  Put_Line("   Identifier : " & To_String(S_Identifier));
	  Put     ("   Range      : "); put(S_Start,0); Put(".."); Put(S_Stop,0);
	  New_Line;
	  Put     ("   Direction  : ");
	  if S_Opposite then Put("R "); else Put("N "); end if;
	  if S_Complementary then Put("C "); else Put("N"); end if;
	  New_Line;
	end if;
        -- Il ne reste plus qu'a lire la sequence
        S_Id := S_Id + 1;
        Read_One_Sequence(To_String(S_Name),S_Identifier,S_Id,
                          S_Format,S_Strand,S_Start,S_Stop,
                          Sequences(S_Id),
			  S_Opposite,S_Complementary,
			  Verbose);
      end if;
    end loop;
    Close(File);
    return Sequences;
  exception
    when Invalid_Sequence_Format =>
      Raise_Exception(Invalid_Sequence_Format'Identity,
		      "L " & Natural'Image(Line_Number));
    when Invalid_Strand =>
      Raise_Exception(Invalid_Strand'Identity,"L " & 
		      Natural'Image(Line_Number));
  end Read_List_Of_Sequences;

  function Define_Shift(Model : in Shift_Model;
			Name  : in String) return Shift_Access is
    Shift : Shift_Access;
  begin
    case Model is
      when Shift_In_File =>
	-- la lecture se fait dans un fichier
	declare
	  File : Ada.Text_Io.File_Type;
	  Line : String(1..1024);
	  Count : Natural := 0;
	  Last : Natural;
	begin
	  -- le fichier doit contenir quatre valeurs entieres par ligne
	  -- chacune donnant le coefficent pour un nucleotide.
	  -- Il y a autant de coordonnes que de lignes dans le fichier
	  -- les lignes commenant par un # ne sont pas compatibilises
	  -- comptage du nombre de lignes
	  Open(File,In_File,Name);
	  while not End_Of_File(File) loop
	    -- Lire la ligne
	    Get_Line(File,Line,Last);
	    -- Verifier la validite de la ligne
	    if Last >= 1 and then Line(1) /= '#' then
	      Count := Count + 1;
	    end if;
	  end loop;
	  Close(File);
	  -- Reserver l'espace pour memoriser les parametres
	  Shift := new Shift_Type(Nucleotide_Type'('A')..Nucleotide_Type'('C'),
				  1..Count);
	  -- Lire effectivement les donnes
	  Count := 1;
	  Open(File,In_File,Name);
	  while not End_Of_File(File) loop
	    if Last >= 1 and then Line(1) /= '#' then
	      -- lire les quatres valeurs de la ligne
	      for I in Shift'Range(1) loop
		Get(File,Shift(I,Count));
	      end loop;
	      Count := Count + 1;
	    end if;
	  end loop;
	  Close(File);
	end;
      when Shift_Default =>
	Shift := new Shift_Type'('A' => ( 2, 3), 'T' => ( 2,-3),
				 'G' => (-3, 2), 'C' => (-3,-2));
      when 1 =>
	Shift := new Shift_Type'('A' => ( 3, 0), 'T' => ( 0, 3) ,
				 'G' => (-3, 0), 'C' => ( 0,-3));
      when 2 =>
	Shift := new Shift_Type'('A' => ( 3, 0), 'T' => (-3,-0) ,
				 'G' => ( 0, 3), 'C' => ( 0,-3));
      when 3 =>
	Shift := new Shift_Type'('A' => ( 2, 3), 'T' => (-3,-2) ,
				 'G' => (-2, 3), 'C' => ( 3,-2));
      when 4 =>
	Shift := new Shift_Type'('A' => ( 6, 0, 0, 0), 'T' => ( 0, 6, 0, 0) ,
				 'G' => ( 0, 0, 6, 0), 'C' => ( 0, 0, 0, 6));
      when 5 =>
	Shift := new Shift_Type'('A' => ( 4, 0, 0, 0), 'T' => ( 0, 4, 0, 0) ,
				 'G' => ( 0, 0, 6, 0), 'C' => ( 0, 0, 0, 6));
      when 6 =>
	Shift := new Shift_Type'('A' => ( 3, 3, 3), 'T' => ( 3,-3,-3) ,
				 'G' => (-3, 3,-3), 'C' => (-3,-3, 3));
      when 7 =>
	Shift := new Shift_Type'('A' => (-4,-6,-3), 'T' => ( 0, 0, 8) ,
				 'G' => (12, 0, 4), 'C' => (-6,10, 4));
      when 8 =>
	Shift := new Shift_Type'('A' => ( 2, 2, 2), 'T' => ( 2,-2,-2) ,
				 'G' => (-3, 3,-3), 'C' => (-3,-3, 3));
    end case;
    return Shift;
  exception
    when Name_Error =>
      Raise_Exception(Name_Error'Identity,"shift " & Name);
    when Sequence_Io_Error =>
      Raise_Exception(Sequence_Io_Error'Identity,Name);
  end Define_Shift;

end Assirc_Read;
