--------------------------------------------------------------------------
-- Assirc_Seeds.D_Eval_M2                   Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 23/03/1998 --
-- Pierre VINCENS                   V 01.001                 17/04/1998 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;

with Cpu_Times.Counters; use Cpu_Times.Counters;

with Assirc_Statistics; use Assirc_Statistics;

with Assirc_Database;
with Assirc_Seeds.Database; use Assirc_Seeds.Database;

with Dassirc_Database;
with Dassirc_Semaphore; use Dassirc_Semaphore;

package body Assirc_Seeds.D_Eval_M2 is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  --use Type_Specifications.IO.Int_Io;
  package SL renames Assirc_Specifications.Database.Segment_List;
  use SL;

--------------------------------------------------------------------------
-- Quelques constantes                                                  --
--------------------------------------------------------------------------
  Max_Pairs_For_Checking : constant := 200;
  -- Le nombre de paires transmises  la base pour vrifier si elles
  -- sont dj prise en compte
  Max_Pairs_For_Registering : constant := 
    Dassirc_Database.Max_Pairs_For_Registering;
  -- Le nombre de paires transmises pour etre enregistr dans la base

--------------------------------------------------------------------------
-- Gestion des paires de brins a analyser                               --
--------------------------------------------------------------------------
  Pairs_To_Check : Access_To_Array_Pairs_To_Check;
  procedure Reset_Pairs_To_Check is
  begin
    Free(Pairs_To_Check);
  end Reset_Pairs_To_Check;

--------------------------------------------------------------------------
-- Evaluation des graines                                               --
--------------------------------------------------------------------------
  procedure Eval_Seeds(Sequences       : in Assirc_Sequence_Array_Access;
                       Seeds           : in out Seed_Segments;
                       Min_Length      : in Integer;
                       Max_Dist        : in Integer;
                       Tolerance       : in Integer;
                       Common_Max_Dist : in Integer;
                       Skip            : in Integer;
                       Xonly           : in Boolean) is
    Max_Dist2 : constant Integer := Max_Dist * Max_Dist;
    Max_Check_Task : constant := 4;
    Pairs : Segment_Pairs(1..Max_Pairs_For_Checking);
    Current : Integer := Pairs'First;
    
    protected Output_Pairs is
      procedure Put(Pair : in Segment_Pair);
      -- Ajouter une paire  enregistrer dans la base
      procedure Purge;
      -- Forcer l'envoi vers la base de toutes les paires en attente
      -- Cette entre doit tre imprativement adresse avant de quitter
      -- ce sous-programme pour viter de perdre les donnes dans le buffer    
    private
      Pairs   : Segment_Pairs(1..Max_Pairs_For_Registering);
      Current : Integer := 1; --Pairs'First;
    end Output_Pairs;
      
    protected body Output_Pairs is
      procedure Put(Pair : in Segment_Pair) is
      begin
        Pairs(Current) := Pair;
        if Current = Pairs'Last then
          -- Le buffer est plein, on le poste
          Statistics.Increment_Registered(Pairs'Length);
          Put_Line("Eval_M2.Output_Pairs: Append_Pairs_In_Tree" & 
		   Integer'Image(Current));
	  pragma Warnings(Off);
	  -- L'accs est effectivement potentiellement bloquant
	  -- mais c'est le but recherch pour viter un embouteillage
	  -- dans les communications avec la partition database
          Database_Access.Lock;
	  Dassirc_Database.Append_Pairs_In_Tree(Pairs);
          Database_Access.Unlock;
	  pragma Warnings(On);
          Put_Line("Eval_M2.Output_Pairs: After Append_Pairs_In_Tree");
          Current := Pairs'First;
        else
          Current := Current + 1;
        end if;
      end Put;
      procedure Purge is
      begin
        if Current > Pairs'First then
          Statistics.Increment_Registered(Current-Pairs'First);
	  Put_Line("Eval_M2.Purge: Append_Pairs_In_Tree" & Integer'Image(Current));
          pragma Warnings(Off);
	  -- L'accs est effectivement potentiellement bloquant
	  -- mais c'est le but recherch pour viter un embouteillage
	  -- dans les communications avec la partition database
          Database_Access.Lock;
	  Dassirc_Database.Append_Pairs_In_Tree(Pairs(Pairs'First..Current-1));
	  Database_Access.Unlock;
	  pragma Warnings(On);
          Put_Line("Eval_M2.Purge: After Append_Pairs_In_Tree");
	end if;
      end Purge;
    end Output_Pairs;

    task type Check_Task is
      entry New_Pairs(Pairs : in Segment_Pairs);
      -- Traite un tableau de segment
      entry Stop;
      -- Pour arrter proprement la tache
    end Check_Task;

    task body Check_Task is
      Input_Pairs  : Segment_Pairs(1..Max_Pairs_For_Checking);
      Answers      : Answers_Array(1..Max_Pairs_For_Checking);
      Pair         : Segment_Pair;
      First        : constant Integer := Input_Pairs'First;
      Last         : Integer;
    begin
      Cpu_Counters.Add("Assirc_Seeds.D_Eval_M2");
      Main: loop
        select 
          accept New_Pairs(Pairs : in Segment_Pairs) do
            Last  := Input_Pairs'First+Pairs'Length-1;
            Input_Pairs(First..Last) := Pairs;
          end New_Pairs;
	  --Put_Line("Eval_M2.D_Eval_M2.Check_Task: Before Are_Included " &
          --        Integer'Image(First) & " .." & Integer'Image(Last));
	  Database_Access.Lock;
	  --Put_Line("Assirc_Seeds.D_Eval_M2.Check_Task : 1");
	  Answers(First..Last) := 
	    Dassirc_Database.Are_Included(Input_Pairs(First..Last));
	  --Put_Line("Assirc_Seeds.D_Eval_M2.Check_Task : 2");
	  Database_Access.Unlock;        
	   for I in First..Last loop
            if not Answers(I) then
              -- Cette graine est dans une region a tester
              Statistics.Increment_Tested;
              -- pour chaque couple non inclus, verifier
              -- s'il y a une zone de similitude
              Check_A_Pair(Input_Pairs(I).Seg1,Input_Pairs(I).Seg2,Sequences,
                           Max_Dist2,Tolerance,Skip,Pair);
              -- verifier si le couple satisfait aux exigences de longueur
              -- et de distance
              if (Pair.Distance <= Max_Dist2 and 
                  Pair.Length >= Min_Length and
                  not Common_Origin(Pair.Seg1,Pair.Seg2,
                                    Common_Max_Dist)) then
                -- enregistrer le couple dans la base locale
		Local_Database.Append_Pair_In_Tree(Pair);
                -- enregistrer le couple pour post  la base de donnes 
		-- distante
                Output_Pairs.Put(Pair);
	end if;
            end if;
          end loop;
        or
          accept Stop;
          exit Main;
        end select;
      end loop Main;
      Cpu_Counters.Remove;
    exception
      when The_Error : others =>
	Put_Line("Exception raised in Assirc_Seeds.D_Eval_M2.Check_Task");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
	Cpu_Counters.Remove;
    end Check_Task;

    --type Check_Task_Range is mod Max_Check_Task;
    type Check_Task_Range is new Integer range 0..Max_Check_Task-1;
    type Check_Task_Array is array(Check_Task_Range) of Check_Task;
    Checking : Check_Task_Array;
    Current_Task : Check_Task_Range := Check_Task_Range'Last;
    
    procedure Repartition(Pairs : in Segment_Pairs);
    pragma Inline(Repartition);
    procedure Repartition(Pairs : in Segment_Pairs) is
      -- Le but ici est de repartir le travail de manire a
      -- profiter des attentes non bloquantes sur les appels
      -- a la base de donnes pour avancer la recherche.
      Previous_Task : Check_Task_Range := Check_Task_Range'Last;
    begin
      loop
        Current_Task := (Current_Task + 1) mod Max_Check_Task;
        select
          Checking(Current_Task).New_Pairs(Pairs);
          Previous_Task := Current_Task;
          return;
        else
          null;
        end select;
        if Current_Task = Previous_Task then
          -- on attend un peu qu'au moins une tache en cours 
          -- d'execution se libere
          delay 0.01;
        end if;
      end loop;
    end Repartition;

    procedure Eval_A_Seed(Seed   : in out Segment_List.List_Handler) is
      -- cette procedure genere tous les couples associes a une graine
      -- et les evalue
      Entries : SL.List_Entries;
    begin
      -- transformer la liste des segments en tableau
      Get_Entries(Seed,Entries);
      -- generer la liste des couples
      for I in Entries'Range loop
        for J in I+1..Entries'Last loop
          -- Constitution d'un nouveau couple
	  Statistics.Increment_Constituted;
	  -- On ne traite que les paires prevues
          if Pairs_To_Check(Consult(Entries(I)).Id,Consult(Entries(I)).Strand,
                            Consult(Entries(J)).Id,Consult(Entries(J)).Strand)
	  then
            declare
	      Seg1 : Segment_Type := Consult(Entries(I));
	      Seg2 : Segment_Type := Consult(Entries(J));
	      Segb : Segment_Type;
	    begin
	      if Seg1.Id = Seg2.Id and Seg1.Strand /= Seg2.Strand then
		if Seg1.Strand > Seg2.Strand then
		  -- on inverse les deux segments pour que le brin 
		  -- Crick soit en premier
		  Segb := Seg1;
		  Seg1 := Seg2;
		  Seg2 := Segb;
		end if;
		if Seg1.Start > Seg2.Start then
		  -- on inverse les deux positions car  W:a-b et C:c-d
		  -- est equivalent a C:a-b avec W:c-d
		  Segb := Seg1;
		  Seg1.Start := Seg2.Start;
		  Seg1.Stop  := Seg2.Stop;
		  Seg2.Start := Segb.Start;  
		  Seg2.Stop  := Segb.Stop;  
		end if;
	      end if;
	      -- Pour chaque couple, le segment d'indice plus bas est mis
	      -- en premier
	      if Seg1.Id < Seg2.Id 
		or else
		(Seg1.Id = Seg2.Id and
		 (Seg1.Strand < Seg2.Strand
		  or else
		(Seg1.Strand = Seg2.Strand and
		 Seg1.Start  < Seg2.Start)))
	      then
		null;
	      else
		Segb := Seg1;
		Seg1 := Seg2;
		Seg2 := Segb;
	      end if;
	      -- pour chaque couple, verifier s'il est deja inclus dans un 
	      -- segment et qu'il n'est pas superpose. Si Xonly est vrai on 
	      -- ne teste que les fragments entre deux sequences differentes
	      -- On regarde aussi si localement, la paire n'est pas dj 
	      -- enregistre
	      if (not (Xonly and Seg1.Id = Seg2.Id) and then
                not Common_Origin(Seg1,Seg2,Common_Max_Dist) and then
		not Local_Database.Included(Seg1,Seg2)) 
	      then
		Pairs(Current).Seg1 := Seg1;
		Pairs(Current).Seg2 := seg2;
		if Current = Pairs'Last then
		  -- le tableau est plein
		  Repartition(Pairs);
		  Current := Pairs'First;
		else
		  Current := Current + 1;
		end if;
	      end if;
	    end;
	  end if;
        end loop;
      end loop;
      -- liberer la place associee au tableau
      Free(Entries);
    end Eval_A_Seed;

  begin -- Eval_Seeds
    if Pairs_To_Check = null then
      -- On ne fait ce calcul qu'au premier appel ou apres un reset
      Pairs_To_Check := Build_Array_Of_Pairs_To_Check(Sequences);
    end if;
    for I in Seeds'Range loop
      Eval_A_Seed(Seeds(I));
    end loop;
    -- Purger le dernier buffer de paires  analyser
    if Current > Pairs'First then
      Repartition(Pairs(Pairs'First..Current-1));
    end if;
    -- Arreter les taches
    for I in Checking'Range loop
      Checking(I).Stop;
    end loop;
    -- Purger le dernier buffer de paires  enregistrer dans la base
    Output_Pairs.Purge;
  end Eval_Seeds;

end Assirc_Seeds.D_Eval_M2;
