--------------------------------------------------------------------------
-- Assirc_Seeds.D_Eval_M3                   Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 31/03/1998 --
-- Pierre VINCENS                   V 01.001                 17/04/1998 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;

with Assirc_Statistics; use Assirc_Statistics;

with Assirc_Database;
with Assirc_Seeds.Database; use Assirc_Seeds.Database;

with Dassirc_Database; 
with Dassirc_Semaphore; use Dassirc_Semaphore;

package body Assirc_Seeds.D_Eval_M3 is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  --use Type_Specifications.IO.Int_Io;
  package SL renames Assirc_Specifications.Database.Segment_List;
  use SL;

--------------------------------------------------------------------------
-- Quelques constantes                                                  --
--------------------------------------------------------------------------
  Max_Seeds_For_Registering : constant := 4000;
  Max_Pairs_For_Checking    : constant := 1000;
  -- Le nombre de graines transmises ou rcupres  chaque interaction
  -- avec la base
  Max_Pairs_For_Registering : constant := 
    Dassirc_Database.Max_Pairs_For_Registering;
  -- Le nombre de paires transmises pour enregistrement dans la base

--------------------------------------------------------------------------
-- Gestion des paires de brins a analyser                               --
--------------------------------------------------------------------------
  Pairs_To_Check : Access_To_Array_Pairs_To_Check;
  
  procedure Reset_Pairs_To_Check is
  begin
    Free(Pairs_To_Check);
  end Reset_Pairs_To_Check;

  procedure Set_Pairs_To_Check(Sequences: in Assirc_Sequence_Array_Access) is
  begin
    Free(Pairs_To_Check);
    Pairs_To_Check := Build_Array_Of_Pairs_To_Check(Sequences);
  end Set_Pairs_To_Check;

--------------------------------------------------------------------------
-- Production des graines et post vers la base de donnes               --
--------------------------------------------------------------------------
  procedure Product_Seeds(Seeds           : in out Seed_Segments;
                          Common_Max_Dist : in Integer;
                          Xonly           : in Boolean) is

    Pairs : Segment_Pairs(1..Max_Seeds_For_Registering);
    Current : Integer := Pairs'First;
    Wait_For_Next : Boolean;

    procedure Eval_A_Seed(Seed   : in out Segment_List.List_Handler) is
      -- cette procedure genere tous les couples associes a une graine
      -- et les evalue
      Entries : SL.List_Entries;
    begin
      -- transformer la liste des segments en tableau
      Get_Entries(Seed,Entries);
      -- generer la liste des couples
      for I in Entries'Range loop
        for J in I+1..Entries'Last loop
          -- Constitution d'un nouveau couple
	  Statistics.Increment_Constituted;
	  -- On ne traite que les paires prevues
          if Pairs_To_Check(Consult(Entries(I)).Id,Consult(Entries(I)).Strand,
                            Consult(Entries(J)).Id,Consult(Entries(J)).Strand)
	  then
            declare
	      Seg1 : Segment_Type := Consult(Entries(I));
	      Seg2 : Segment_Type := Consult(Entries(J));
	      Segb : Segment_Type;
	    begin
	      -- Dans le cas de segments situes sur un meme chromosome
	      -- et sur des brins differents, on determine le couple 
	      -- d'indice le plus bas
	      if Seg1.Id = Seg2.Id and Seg1.Strand /= Seg2.Strand then
		if Seg1.Strand > Seg2.Strand then
		  -- on inverse les deux segments pour que le brin 
		  -- Crick soit en premier
		  Segb := Seg1;
		  Seg1 := Seg2;
		  Seg2 := Segb;
		end if;
		if Seg1.Start > Seg2.Start then
		  -- on inverse les deux positions car  W:a-b et C:c-d
		  -- est equivalent a C:a-b avec W:c-d
		  Segb := Seg1;
		  Seg1.Start := Seg2.Start;
		  Seg1.Stop  := Seg2.Stop;
		  Seg2.Start := Segb.Start;  
		  Seg2.Stop  := Segb.Stop;  
		end if;
	      end if;
	      -- Pour chaque couple, le segment d'indice plus bas est mis
	      -- en premier
	      if Seg1.Id < Seg2.Id 
		or else
		(Seg1.Id = Seg2.Id and
		 (Seg1.Strand < Seg2.Strand
		  or else
		(Seg1.Strand = Seg2.Strand and
		 Seg1.Start  < Seg2.Start)))
	      then
		null;
	      else
		Segb := Seg1;
		Seg1 := Seg2;
		Seg2 := Segb;
	      end if;
	      -- pour chaque couple, verifier s'il est deja inclus dans un 
	      -- segment et qu'il n'est pas superpose. Si Xonly est vrai on 
	      -- ne teste que les fragments entre deux sequences differentes
	      if (not (Xonly and Seg1.Id = Seg2.Id) and then
                not Common_Origin(Seg1,Seg2,Common_Max_Dist) and then
		not Local_Database.Included(Seg1,Seg2))
	      then
		-- On regarde si localement, la paire est dj enregistr
		Pairs(Current).Seg1 := Seg1;
		Pairs(Current).Seg2 := Seg2;
		if Current = Pairs'Last then
		  -- le tableau est plein
		  Statistics.Increment_Tested(Pairs'Length);
		  Database_Access.Lock;
		  Dassirc_Database.To_Check_Inclusion(Pairs,Wait_For_Next);
		  Database_Access.Unlock;
		  Current := Pairs'First;
		  if Wait_For_Next then
		    -- La pile a dpasse son seuil de saturation. On bloque
		    -- le processus temporairement. Cela permet aux autres
		    -- taches de progresser 
		    delay 3.0;
		  end if;
		else
		  Current := Current + 1;
		end if;
	      end if;
	    end;
          end if;
        end loop;
      end loop;
      -- liberer la place associee au tableau
      Free(Entries);
    end Eval_A_Seed;

  begin -- Product_Seeds
    for I in Seeds'Range loop
      Eval_A_Seed(Seeds(I));
    end loop;
    -- Poster le dernier paquet
    if Current > Pairs'First then
      Statistics.Increment_Tested(Current-Pairs'First);
      Database_Access.Lock;
      Dassirc_Database.To_Check_Inclusion(Pairs(Pairs'First..Current-1),
					  Wait_For_Next);
      Database_Access.Unlock;
    end if;
  end Product_Seeds;

--------------------------------------------------------------------------
-- Expansion des graines rcupres de la base de donnes               --
--------------------------------------------------------------------------
  procedure Expand_Seeds(Sequences       : in Assirc_Sequence_Array_Access;
                         Min_Length      : in Integer;
                         Max_Dist        : in Integer;
                         Tolerance       : in Integer;
                         Common_Max_Dist : in Integer;
                         Skip            : in Integer) is
    Max_Dist2 : constant Integer := Max_Dist * Max_Dist;
    Input_Pairs  : Segment_Pairs(1..Max_Pairs_For_Checking);
    Output_Pairs : Segment_Pairs(1..Max_Pairs_For_Registering);
    Last  : Integer;
    Current : Integer := Output_Pairs'First;
  begin
    Put_Line("Assirc_Seeds.D_Eval_M3: entre dans Expand_Seeds");
    loop
      Database_Access.Lock;
      Dassirc_Database.Next_Pairs_To_Expand(Input_Pairs,Last);
      Database_Access.Unlock;
      for I in Input_Pairs'First..Last loop
        -- raliser l'extension pour chaque paire
        -- pour chaque couple non inclus, verifier
        -- s'il y a une zone d'homologie
        Check_A_Pair(Input_Pairs(I).Seg1,Input_Pairs(I).Seg2,Sequences,
                     Max_Dist2,Tolerance,Skip,Output_Pairs(Current));
        -- verifier si le couple satisfait aux exigences de longueur
        -- et de distance
        if (Output_Pairs(Current).Distance <= Max_Dist2 and 
            Output_Pairs(Current).Length >= Min_Length and
            not Common_Origin(Output_Pairs(Current).Seg1,
                              Output_Pairs(Current).Seg2,
                              Common_Max_Dist)) then
	  -- Ajouter le couple  la base locale
	  Local_Database.Append_Pair_In_Tree(Output_Pairs(Current));
	  -- enregistrer le couple dans la base remote
	  if Current = Output_Pairs'Last then
	    -- le tableau est plein
	    Statistics.Increment_Registered(Output_Pairs'Length);
	    Database_Access.Lock;
	    Dassirc_Database.Append_Pairs_In_Tree(Output_Pairs);
	    Database_Access.Unlock;
	    Current := Output_Pairs'First;
	  else
	    Current := Current + 1;
	  end if;
        end if;
      end loop;
      -- on quitte la boucle si le nombre de paires transmis par 
      -- la base est plus petit que la taille du buffer et si la
      -- base valide le fait qu'il n'y a plus de paires  valuer.
      -- Cette dernire condition n'est teste que si la premire
      -- est satisfaite dans le but de limiter les transactions
      -- avec la base.  
      --exit when Last < Input_Pairs'Last and then No_Pair_To_Expand;
      if Last < Input_Pairs'Last then
	declare
	  Finished : Boolean;
	begin
	  Database_Access.Lock;
	  Finished := Dassirc_Database.No_Pair_To_Expand;
	  Database_Access.Unlock;
	  if Finished then
	    -- C'est fini, la base n'enverra plus de paires  traiter
	    exit;
	  else
	    -- On attend un peu avant de redemander des paires 
	    delay 0.1;
	  end if;
	end;
      end if;
    end loop;
    -- Il reste  poster le dernier tableau
    if Current > Output_Pairs'First then
      Statistics.Increment_Registered(Current-Output_Pairs'First);
      Database_Access.Lock;
      Dassirc_Database.Append_Pairs_In_Tree(Output_Pairs(Output_Pairs'First..Current-1));
      Database_Access.Unlock;
    end if;
    Put_Line("Assirc_Seeds.D_Eval_M3: sort de Expand_Seeds");    
  end Expand_Seeds;
  
end Assirc_Seeds.D_Eval_M3;
