--------------------------------------------------------------------------
-- Assirc_Seeds.D_Eval_M4                   Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 24/06/1999 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;
with Ada.Calendar; use Ada.Calendar;  
      
with Cpu_Times.Counters; use Cpu_Times.Counters;

with Assirc_Statistics; use Assirc_Statistics;

with Assirc_Database;
with Assirc_Seeds.Database; use Assirc_Seeds.Database;

with Dassirc_Database;
with Dassirc_Semaphore; use Dassirc_Semaphore;

pragma Elaborate_All(Cpu_Times.Counters);
pragma Elaborate_All(Assirc_Seeds.Database);
pragma Elaborate_All(Dassirc_Semaphore);
pragma Elaborate_All(Dassirc_Database);

package body Assirc_Seeds.D_Eval_M4 is

  
--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  --use Type_Specifications.IO.Int_Io;
  package SL renames Assirc_Specifications.Database.Segment_List;
  use SL;

--------------------------------------------------------------------------
-- Quelques constantes                                                  --
--------------------------------------------------------------------------
  Max_Pairs_For_Registering : constant := 
    Dassirc_Database.Max_Pairs_For_Registering;
  -- Le nombre de paires transmises pour etre enregistr dans la base

--------------------------------------------------------------------------
-- Gestion des paires de brins a analyser                               --
--------------------------------------------------------------------------
  Pairs_To_Check : Access_To_Array_Pairs_To_Check;
  procedure Reset_Pairs_To_Check is
  begin
    Free(Pairs_To_Check);
  end Reset_Pairs_To_Check;

--------------------------------------------------------------------------
-- L'identificateur numero de client de la partition                    --
--------------------------------------------------------------------------
  Local_Client : View_Id := Unassigned_Id;
  -- Cette variable est mise  jour par la tache Update_Local_Database
  -- et n'est que consult dans les autres cas. Il ne devrait pas y
  -- avoir de problme  ce qu'elle soit partage dans la mesure ou
  -- sa mise  jour est faite durant un rendez-vous entre les deux
  -- appelants potentiels, une seule fois bien avant son utilisation
  -- effective

--------------------------------------------------------------------------
-- Un type protg pour poster les donnes vers la base distante        --
--------------------------------------------------------------------------
  protected Output_Pairs is
    procedure Put(Pair : in Segment_Pair);
    -- Ajouter une paire  enregistrer dans la base
    procedure Purge;
    -- Forcer l'envoi vers la base de toutes les paires en attente
    -- Cette entre doit tre imprativement adresse avant de quitter
    -- ce sous-programme pour viter de perdre les donnes dans le buffer    
  private
    Pairs   : Segment_Pairs(1..Max_Pairs_For_Registering);
    Current : Integer := 1; --Pairs'First;
  end Output_Pairs;
  
  protected body Output_Pairs is
    procedure Put(Pair : in Segment_Pair) is
    begin
      Pairs(Current) := Pair;
      if Current = Pairs'Last then
	-- Le buffer est plein, on le poste
	Put_Line("Eval_M4.Output_Pairs: Append_Pairs_In_Tree" & 
		 Integer'Image(Current));
	pragma Warnings(Off);
	-- L'accs est effectivement potentiellement bloquant
	-- mais c'est le but recherch pour viter un embouteillage
	-- dans les communications avec la partition database
	Database_Access.Lock;
	Dassirc_Database.Append_Pairs_In_Tree(Local_Client,Pairs);
	Database_Access.Unlock;
	pragma Warnings(On);
	Current := Pairs'First;
      else
	Current := Current + 1;
      end if;
    exception
      when The_Error : others =>
	Put_Line("Exception raised in Assirc_Seeds.D_Eval_M4.Output_Pairs.Put");
	Put_Line("Assirc_Seeds.D_Eval_M4.Output_Pairs.Put: Local_Client" &
		 View_Id'Image(Local_Client));
	Put_Line("Assirc_Seeds.D_Eval_M4.Output_Pairs.Put: Current" &
		 Integer'Image(Current));
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Put;
    procedure Purge is
    begin
      if Current > Pairs'First then
	Put_Line("Eval_M4.Purge: Append_Pairs_In_Tree" & Integer'Image(Current));
	pragma Warnings(Off);
	-- L'accs est effectivement potentiellement bloquant
	-- mais c'est le but recherch pour viter un embouteillage
	-- dans les communications avec la partition database
	Database_Access.Lock;
	Dassirc_Database.Append_Pairs_In_Tree(Local_Client,
					      Pairs(Pairs'First..Current-1));
	Database_Access.Unlock;
	pragma Warnings(On);
	Put_Line("Eval_M4.Purge: After Append_Pairs_In_Tree");
      end if;
    exception
      when The_Error : others =>
	Put_Line("Exception raised in Assirc_Seeds.D_Eval_M4.Output_Pairs.Purge");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Purge;
  end Output_Pairs;

--------------------------------------------------------------------------
-- La tache rcuprant les donnes dans la base distante                --
--------------------------------------------------------------------------
  task Update_Local_Database is
    entry Start(Client : in Client_Id);
    entry Shutdown;
    entry Stop;
  end Update_Local_Database;

  task body Update_Local_Database is
    Period : constant Duration := 2.0; 
    -- Temps entre deux mise a jour de la base locale depuis la base distante 
    Next_Time : Time := Clock + Period;
    Pairs : Segment_Pairs(1..Dassirc_Database.Max_Pairs_For_Diffusion);
    Last : Integer;
    Next : Boolean;
    NO_TREATMENT : exception;
  begin
    Cpu_Counters.Add("Assirc_Seeds.D_Eval_M4");
    Put_Line("Assirc_Seeds.D_Eval_M4.Update_Local_Database: initialise");
    select
      accept Start(Client : in Client_Id) do
	Local_Client := Client; -- mise  jour de la variable globase
      end Start;
    or 
      accept Shutdown;
      raise NO_TREATMENT;
    or 
      terminate;
    end select;
    Put_Line("Assirc_Seeds.D_Eval_M4.Update_Local_Database: active");
    loop
      -- Attendre un ordre de fin de traitement
      select
	accept Stop;
	exit;
      or
	delay until Next_Time;
      end select;
      Next_Time := Next_Time + Period;
      -- Rcuprer les donnes distantes et mettre  jour la base locale
      loop
	Database_Access.Lock;
	Dassirc_Database.Last_Pairs_From_Client(Local_Client,Pairs,Last,Next);
	Database_Access.Unlock;
	Put_Line("Assirc_Seeds.D_Eval_M4:" & Integer'Image(Last) & " paires lues");
	exit when Last = 0;
	for I in 1..Last loop
	  Local_Database.Append_Pair_In_Tree(Pairs(I));
	end loop;
	exit when not Next; -- pas de paquets en attente pour ce client
      end loop;
    end loop;
    -- Terminer la tache de mise  jour
    Cpu_Counters.Remove;
  exception
    when NO_TREATMENT =>
      Cpu_Counters.Remove;
      Put_Line("Assirc_Seeds.D_Eval_M4.Update_Local_Database: shutdown");
    when The_Error : others =>
      Put_Line("Exception raised in Assirc_Seeds.D_Eval_M4.Update_Local_Database");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      Cpu_Counters.Remove;
  end Update_Local_Database;

  procedure Start_Local_Database(Client : in Client_Id) is
  begin
    -- S'enregistrer auprs du serveur central
    Database_Access.Lock;
    Dassirc_Database.Add_Client(Client);
    Database_Access.Unlock;
    Put_Line("Assirc_Seeds.D_Eval_M4.Start_Local_Database: client" & Client_Id'Image(Client) &
	     " enregistre dans la base");
    -- Initier la tache correspondante
    Update_Local_Database.Start(Client);
  end Start_Local_Database;

  procedure Shutdown_Local_Database is
  begin
    Update_Local_Database.Shutdown;
  end Shutdown_Local_Database;

  procedure Stop_Local_Database is
  begin
    -- Informer le serveur que l'on ne veut plus etre recevoir les
    -- mises  jour
    Put_Line("Assirc_Seeds.D_Eval_M4.Stop_Local_Database: Arret des demandes de diffusion");
    Database_Access.Lock;
    Put_Line("Assirc_Seeds.D_Eval_M4.Stop_Local_Database: Lock active");
    Dassirc_Database.Remove_Client(Local_Client);
    Put_Line("Assirc_Seeds.D_Eval_M4.Stop_Local_Database: Deconnexion");
    Database_Access.Unlock;
    Put_Line("Assirc_Seeds.D_Eval_M4.Update_Local_Database: Client" &
	     Client_Id'Image(Local_Client) & " deconnecte");
    -- Arrter la tache locale
    Update_Local_Database.Stop;
  end Stop_Local_Database;

--------------------------------------------------------------------------
-- Evaluation des graines                                               --
--------------------------------------------------------------------------
  procedure Eval_Seeds(Sequences       : in Assirc_Sequence_Array_Access;
                       Seeds           : in out Seed_Segments;
                       Min_Length      : in Integer;
                       Max_Dist        : in Integer;
                       Tolerance       : in Integer;
                       Common_Max_Dist : in Integer;
                       Skip            : in Integer;
                       Xonly           : in Boolean) is
    Max_Dist2 : constant Integer := Max_Dist * Max_Dist;
    Max_Check_Task : constant := 4;
    --Pairs : Segment_Pairs(1..Max_Pairs_For_Checking);
    --Current : Integer := Pairs'First;
    


    procedure Eval_A_Seed(Seed   : in out Segment_List.List_Handler) is
      -- cette procedure genere tous les couples associes a une graine
      -- et les evalue
      Entries : SL.List_Entries;
      Pair    : Segment_Pair;
    begin
      -- transformer la liste des segments en tableau
      Get_Entries(Seed,Entries);
      -- generer la liste des couples
      for I in Entries'Range loop
        for J in I+1..Entries'Last loop
          -- Constitution d'un nouveau couple
	  Statistics.Increment_Constituted;
	  -- On ne traite que les paires prevues
          if Pairs_To_Check(Consult(Entries(I)).Id,Consult(Entries(I)).Strand,
                            Consult(Entries(J)).Id,Consult(Entries(J)).Strand)
	  then
	    declare
	      Seg1 : Segment_Type := Consult(Entries(I));
	      Seg2 : Segment_Type := Consult(Entries(J));
	      Segb : Segment_Type;
	    begin
	      -- Dans le cas de segments situes sur un meme chromosome
	      -- et sur des brins differents, on determine le couple 
	      -- d'indice le plus bas
	      if Seg1.Id = Seg2.Id and Seg1.Strand /= Seg2.Strand then
		if Seg1.Strand > Seg2.Strand then
		  -- on inverse les deux segments pour que le brin 
		  -- Crick soit en premier
		  Segb := Seg1;
		  Seg1 := Seg2;
		  Seg2 := Segb;
		end if;
		if Seg1.Start > Seg2.Start then
		  -- on inverse les deux positions car  W:a-b et C:c-d
		  -- est equivalent a C:a-b avec W:c-d
		  Segb := Seg1;
		  Seg1.Start := Seg2.Start;
		  Seg1.Stop  := Seg2.Stop;
		  Seg2.Start := Segb.Start;  
		  Seg2.Stop  := Segb.Stop;  
		end if;
	      end if;
	      -- Pour chaque couple, le segment d'indice plus bas est mis
	      -- en premier
	      if Seg1.Id < Seg2.Id 
		or else
		(Seg1.Id = Seg2.Id and
		 (Seg1.Strand < Seg2.Strand
		  or else
		(Seg1.Strand = Seg2.Strand and
		 Seg1.Start  < Seg2.Start)))
	      then
		null;
	      else
		Segb := Seg1;
		Seg1 := Seg2;
		Seg2 := Segb;
	      end if;
	      -- pour chaque couple, verifier s'il est deja inclus dans un 
	      -- segment et qu'il n'est pas superpose. Si Xonly est vrai on 
	      -- ne teste que les fragments entre deux sequences differentes
	      if (not (Xonly and Seg1.Id = Seg2.Id) and then
                not Common_Origin(Seg1,Seg2,Common_Max_Dist) and then
		not Local_Database.Included(Seg1,Seg2)) 
	      then
		-- Il est dans une region a tester
		Statistics.Increment_Tested;
		-- pour chaque couple non inclus, verifier
		-- s'il y a une zone d'homologie
		Check_A_Pair(Seg1,Seg2,Sequences,
			     Max_Dist2,Tolerance,Skip,Pair);
		-- verifier si le couple satisfait aux exigences de longueur
		-- et de distance
		if Pair.Distance <= Max_Dist2 and 
		  Pair.Length >= Min_Length and
		  not Common_Origin(Pair.Seg1,Pair.Seg2,Common_Max_Dist) then
		  -- enregistrer le couple
		  Statistics.Increment_Registered;
		  -- dans la base locale
		  Local_Database.Append_Pair_In_Tree(Pair);
		  -- et pour post dans la base distante
		  Output_Pairs.Put(Pair);
		end if;
	      end if;
	    end;
          end if;
        end loop;
      end loop;
      -- liberer la place associee au tableau
      Free(Entries);
    end Eval_A_Seed;

  begin -- Eval_Seeds
    if Pairs_To_Check = null then
      -- On ne fait ce calcul qu'au premier appel ou apres un reset
      Pairs_To_Check := Build_Array_Of_Pairs_To_Check(Sequences);
    end if;
    Put_Line("Assirc_Seeds.D_Eval_M4.Eval_Seeds: Evaluation des graines activee");
    for I in Seeds'Range loop
      Eval_A_Seed(Seeds(I));
    end loop;
    -- Purger le dernier buffer de paires  enregistrer dans la base
    Output_Pairs.Purge;
  exception
    when The_Error : others =>
      Put_Line("Exception raised in Assirc_Seeds.D_Eval_M4.Eval_Seeds");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      raise;
  end Eval_Seeds;

end Assirc_Seeds.D_Eval_M4;
