--------------------------------------------------------------------------
-- Assirc_Seeds                             Copyrigth ENS ULM 1996-1997 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 24/07/1997 --
-- Pierre VINCENS                   V 01.001                 23/03/1998 --
-- Pierre VINCENS                   V 01.002                 23/03/1998 --
-- Pierre VINCENS                   V 01.003                 17/04/1998 --
--------------------------------------------------------------------------
-- Recherche des graines et test
--------------------------------------------------------------------------
-- V 01.001
-- Amelioration de la prise en compte des gaps. L'enregistrement dans la
-- base de donnes d'un fragment tendu se fait sur la base d'une
-- taille de squence plus courte (parametre rlength). D'autre par,
-- on considere comme un seul bloc de similarite des fragments spar
-- par une distance au plus egale a un seuil fixe.
--------------------------------------------------------------------------
-- V 01.002
-- Adaptation du paquetage pour le rendre compatible avec la version
-- distribue. Eval_Seeds est reporte dans un paquetage fils dpendant
-- du type de distribution choisi
--------------------------------------------------------------------------
-- V 01.003
-- Transformation du paquetage pour pouvoir traiter les deux brins sans
-- avoir besoin de calculer la sequence complementaire. Cela introduit
-- l'ajout d'un parametre Strand aux procedures Get_Seeds et 
-- Get_Seeds_By_Mask
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;
with Type_specifications; use Type_Specifications;
with Sequence_Specifications; use Sequence_Specifications;
with Assirc_Specifications; use Assirc_Specifications;
with Assirc_Specifications.Basic; use Assirc_Specifications.Basic;
with Assirc_Specifications.Database; use Assirc_Specifications.Database;

package Assirc_Seeds is

--------------------------------------------------------------------------
-- Gestion des parametres                                               --
--------------------------------------------------------------------------
  procedure Set_Seed_Parameters(Shift : Shift_Type);
  pragma Inline(Set_Seed_Parameters);
  -- Enregistrement des parametres utilis pour la recherche et
  -- l'valuation des graines. Seul le parametre Shift a ete
  -- maintenu en raison de sa taille et du nombre de fois o il
  -- serait ncessaire de le transmettre

--------------------------------------------------------------------------
-- Gestion de la structure de donnes associes au stockage des graines --
--------------------------------------------------------------------------
  procedure Init_Seeds(Seed_Size   : in Integer;
                       Seeds       : in out Seed_Segments);
  procedure Reset_Seeds(Seeds : in out Seed_Segments);
  procedure Free_Seeds(Seeds : in out Seed_Segments);
  pragma Inline(Init_Seeds,Reset_Seeds,Free_Seeds);

--------------------------------------------------------------------------
-- Gnration des graines                                               --
--------------------------------------------------------------------------
  procedure Get_Seeds(Sequence    : in Assirc_Sequence_Data;
                      Strand      : in Strand_Type;
                      Seed_Size   : in Integer;
                      Seeds       : in out Seed_Segments);
  -- Dtermination des graines d'une squence et enregistrement de
  -- leur position dans la structure de donnes Seeds

  procedure Get_Seeds_By_Mask(Sequence    : in Assirc_Sequence_Data;
                              Strand      : in Strand_Type;
                              Seed_Size   : in Integer;
                              Mask_Size   : in Integer;
                              Mask        : in Integer;
                              Seeds       : in out Seed_Segments);
  -- Dtermination des graines d'une squence dont la partie masque
  -- est gale  la valeur de masque spcifie et enregistrement de
  -- leur position dans la structure de donnes Seeds. Seed_Size est
  -- la taille totale de la graine et Mask_Size la taille reservee
  -- pour le calcul du masque.

private

  -- Les procdures suivantes sont privees. Leur dclaration  ce niveau
  -- est destine  les rendre accessible au fils du paquetage.

  function Common_Origin(Seg1,Seg2 : in Segment_Type;
                         Max_Dist  : in Integer) return Boolean;
  pragma Inline(Common_Origin);
  -- Dtermination de l'origine commune ventuelle de deux segments
 
  procedure Check_A_Pair(Seg1,Seg2 : in Segment_Type;
                         Sequence  : in Assirc_Sequence_Array_Access;
                         Max_Dist2 : in Integer;
                         Tolerance : in Integer;
                         Skip      : in Integer;
                         Pair      : in out Segment_Pair);
  pragma Inline(Check_A_Pair);
  -- Dtermination de la zone de similitude par application d'une
  -- extension a base de random_walk. 
  
  type Array_Of_Pairs_To_Check is array(Accession_Type range <>,
                                        Strand_Type range <>,
                                        Accession_Type range <>,
                                        Strand_Type range <>) of boolean;
  type Access_To_Array_Pairs_To_Check is access Array_Of_Pairs_To_Check; 
  procedure Free is new 
    Ada.Unchecked_Deallocation(Array_Of_Pairs_To_Check,
                               Access_To_Array_Pairs_To_Check);
  function Build_Array_Of_Pairs_To_Check
    (Sequences : in Assirc_Sequence_Array_Access)
    return Access_To_Array_Pairs_To_Check;
  -- Construction d'un tableau permettant de determiner quels sont les
  -- couples qui doivent etre analyses compte-tenu des brins consideres
  -- En effet, si l'analyse demandee porte sur deux sequences et pour chacune
  -- sur les deux brins, il suffit de considerer les paires Watson-Watson et
  -- Watson-Crick. Les autres paires son deductibles par symetrie.
  
end Assirc_Seeds;
