--------------------------------------------------------------------------
-- Assirc_Specifications.Database           Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 12/03/1998 --
--------------------------------------------------------------------------
-- Dfinition des type de donnes de base du programme Assirc utilises
-- pour la gestion de la base de donnes
--------------------------------------------------------------------------

with Ada.Unchecked_Deallocation;
with Strlib; use Strlib;
with Single_List,Rb_Tree;
with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

pragma Elaborate_All(Rb_Tree);

package Assirc_Specifications.Database is

  ------------------------------------------------------------------------
  -- Mmorisation des segments                                          --
  ------------------------------------------------------------------------
  package Segment_List is new Single_List(Segment_Type);
    -- definissent la maniere dont sont grs les segments

  type Seed_Segment_Array is array(Integer range <>) of
    Segment_List.List_Handler;
  type Seed_Segments is access Seed_Segment_Array;
  procedure Free is new Ada.Unchecked_Deallocation(Seed_Segment_Array,
                                                   Seed_Segments);
    -- dfinissent le tableau utilis pour mmoriser les segments
    -- en fonction du rsultat de la fonction de hash-coding

  ------------------------------------------------------------------------
  -- Mmorisation des paires de segments                                --
  ------------------------------------------------------------------------
  function Eq(A,B : in Segment_Pair) return Boolean;
  function inf(A,B : in Segment_Pair) return Boolean;
  pragma Inline(Eq,Inf);
  package Similar_Regions_RB_Tree is new Rb_Tree(Segment_Pair,Eq,Inf);
  type Similar_Regions_Tree is record
    Tree           : Similar_Regions_RB_Tree.Tree_Handler;
    -- Pointeur sur l'arbre
    Max_Seg_Length : Integer := 0;
    -- Longueur du segment de taille maximale
  end record;		    
  subtype Similar_Regions_Ptr is Similar_Regions_RB_Tree.Tree_Ptr;
  subtype Similar_Regions_Entries is Similar_Regions_RB_Tree.Tree_Entries;
    -- definissent la manire dont sont mmorises les paires
    -- similaires potentielles. L'enregistrement se fait dans
    -- un red-black tree dont l'ordonnancement est fix par
    -- les fonctions eq et inf. La gestion de cette structure
    -- est ddie au paquetage rb_tree (optimisation des recherches,
    -- maintient de l'quilibre de l'arbre,...)
  
end Assirc_Specifications.Database;
