--------------------------------------------------------------------------
-- Assirc_Specifications                    Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 24/07/1997 --
-- Pierre VINCENS                   V 01.001                 12/03/1998 --
--------------------------------------------------------------------------
-- Dfinition des type de donnes de base du programme Assirc
--------------------------------------------------------------------------
-- V 01.001
-- Report des dclarations interdisant de faire un paquetage PURE dans
-- Assirc_Specification.Basic
-- Report des dfinitions lies  la gestion de la base de donnes
-- interne dans le paquetage Assirc_Specifications.Database
--------------------------------------------------------------------------
with Type_Specifications; use Type_Specifications;
with Sequence_Specifications; use Sequence_Specifications;

package Assirc_Specifications is

  pragma Pure;

  ------------------------------------------------------------------------
  -- Gestion des erreurs                                                --
  ------------------------------------------------------------------------
  ERROR_CONCATENE      : exception;

  ------------------------------------------------------------------------
  -- Constantes                                                         --
  ------------------------------------------------------------------------
  Nb_Symbol : constant Integer := 4;
    -- definit le nombre de nucleotides differents pour la fonction de
    -- hash-coding. Les zones comportant des nuclotides autres que
    -- A, T, G, C et U sont ignores dans la localisation des paires
    -- initales

  ------------------------------------------------------------------------
  -- Modle des dplacements                                            --
  ------------------------------------------------------------------------
  type Shift_Model is new Integer range -1..8;
  Shift_In_file : constant Shift_Model := -1;
  Shift_Default : constant Shift_Model :=  0;
  -- prcide la rfrence aux modles par dfaut

  type Shift_Type is array(Nucleotide_Type range <>, 
			   Integer range <>) of Integer;
  -- dcrit les paramtre du shift
  
  ------------------------------------------------------------------------
  -- Identification des squences                                       --
  ------------------------------------------------------------------------
  type Accession_Type is new Integer_16; -- Entier sur deux octets
  
  ------------------------------------------------------------------------
  -- Mmorisation des segments                                          --
  ------------------------------------------------------------------------
  type Segment_Type is record
    Id : Accession_Type;  -- identificateur de la sequence
    Strand : Strand_Type; -- brin considere
    Start,Stop : Integer; -- position dans la sequence
  end record;
  -- On pourrait eviter de memoriser le brin en considerant que l'on
  -- est sur le brin Watson si Stop > Start et sur le brin Crick si
  -- Start > Stop. Toutefois, conserve ce champ permet de gagner du temps 
  -- lors des tests et facilite la gestion de la base interne. Quel que soit
  -- le brin considere, on aura donc toujours start < stop.
  -- On notera que l'on ne perd pas de place dans la mesure ou Id
  -- est cod sur deux octets et Strand sur un
  -- A VERIFIER et FORCER L'ALIGNEMENT SI NECESSAIRE
  
  ------------------------------------------------------------------------
  -- Mmorisation des paires de segments                                --
  ------------------------------------------------------------------------
  type Segment_Pair is record
    Seg1,Seg2 : Segment_Type;
    Distance  : Integer;
    Length    : Integer;
  end record;
  type Segment_Pairs is array(Integer range <>) of Segment_Pair;

  ------------------------------------------------------------------------
  -- Mmorisation d'une liste de rponse                                --
  ------------------------------------------------------------------------
  type Answers_Array is array(Integer range <>) of Boolean;
  pragma Pack(Answers_Array);
  
end Assirc_Specifications;
