--------------------------------------------------------------------------
-- Cpu_Times.Utilities                      Copyrigth ENS ULM 1998-1999 --
-- Pierre VINCENS                   V 01.000                 14/09/1998 --
-- Pierre VINCENS                   V 01.001                 27/05/1999 --
--------------------------------------------------------------------------
-- VERSION Unix pour Linux
--------------------------------------------------------------------------
-- Dtermination du temps utilis par un process et ses fils.
--
-- ATTENTION, ce package est compilable sur n'importe quel systme, mais
-- n'a un sens que sous Linux. Il n'y a pas de verification de l'OS
-- utilis.
--------------------------------------------------------------------------
-- v1.001
-- Ajout de la fonctionnalite de gestion de liste de process et de cumul
-- des temps associes. Pour cela, il faut constituer la liste des process
-- actif en appelant Add_Pid_In_List en dbut de tache et Remove_Pid_In_List
-- en fin de tache.
--------------------------------------------------------------------------
with System;
with Interfaces.C; use type Interfaces.C.Long;

with Ada.Text_Io; use Ada.Text_Io;
with Ada.Integer_Text_Io; use Ada.Integer_Text_Io;
with Ada.Calendar; use Ada.Calendar;

with System_Constants; use System_Constants;

with Strlib; use Strlib; 
with Linked_List;

with Cpu_Times.Basic; 

package Cpu_Times.Utilities is 

  -- type Process_Id is private;
  type Process_Id is new Interfaces.C.Int;
  Null_Pid : constant Process_Id := 0;

  type Process_Unit is record
    Up   : Boolean := True;
    Pid  : Process_Id := Null_Pid;
    Key  : Varstring;
    User_Time   : Duration := 0.0;
    System_Time : Duration := 0.0;
  end record;
  package Process_Unit_List is new Linked_List(Process_Unit);
  subtype Process_List is Process_Unit_List.List_Handler;
 
  function Getpid return Process_Id;
  -- Retourne le numro du process courant
  function Getppid return Process_Id;
  -- Retourne le numro du process pre

  procedure Add_Pid_In_List(List : in out Process_List;
			    Key  : in String);
  procedure Remove_Pid_From_List(List : in out Process_List);
  -- Ajoute ou retire de la liste le process courant
  -- Key permet simplement d'associer un identificateur au process
  procedure Add_Pid_In_List(List : in out Process_List;
			    Pid  : in Process_Id;
			    Key  : in String);
  procedure Remove_Pid_From_List(List : in out Process_List;
			       Pid  : in Process_Id);
  -- Ajoute ou retire de la liste le process pid
  -- Key permet simplement d'associer un identificateur au process

  -- Note: pour Add_Pid_In_List, le process n'est pas ajout s'il est dj 
  -- dans la liste

  procedure Add_Childs(List : in out Process_List);
  -- Cette procedure ajoute  la liste des process de la liste
  -- leur fils. Le but est de pouvoir rcuprer l'identificateur des
  -- process "systemes" utiliss pour la gestion des communications

  procedure Print_Process_List(List : in Process_List);
  -- Affiche la liste des process actifs

  procedure Print_Elapsed_Times(List : in Process_List);
  -- Cummule le temps occupe par tous les process

  function Elapsed_Times(Pid : in Process_Id) return Cpu_Duration;
  pragma Inline(Elapsed_Times);
  -- Donne les temps CPU (real+user+system) coul depuis le dbut du process
  -- donn pid.
  -- L'obtention de la donne se fait en consultatnt le fichier /proc/<pid>/stat
  -- Cette fonction est specifique  Linux

  procedure Elapsed_Times(List : in Process_List; 
			  Elapsed : in out Cpu_Duration);
  -- Donne les temps CPU (real+user+system) coul depuis le dbut des process
  -- spcifis dans la liste.
  -- L'obtention de la donne se fait en consultant le fichier: 
  --        /proc/<pid>/stat
  -- Cette fonction est specifique  Linux

private
  --type Process_Id is new Interfaces.C.Int;
  pragma Import(C,Getpid,"getpid");
  pragma Import(C,Getppid,"getppid");
end Cpu_Times.Utilities;
