--------------------------------------------------------------------------
-- Dassirc_Common                           Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 01.000                 06/03/1998 --
--------------------------------------------------------------------------
-- Paquetage de dclarations des types necessaires a la gestion 
-- rparties des taches
--------------------------------------------------------------------------
with Type_Specifications; use Type_Specifications;
with Cpu_Times; use Cpu_Times;
with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Alignment; use Sequence_Alignment;
with Assirc_Specifications; use Assirc_Specifications;

package Dassirc_Common is

  pragma Pure;


--------------------------------------------------------------------------
-- Les exceptions impliques dans les relations clients-serveurs        --
--------------------------------------------------------------------------
  DISPATCHER_COMPLETED : exception;
  -- Le dispatcher a termin et rejete les demandes de progression

  TOO_MANY_CLIENTS : exception;
  -- Saturation de la table d'affectation des clients sur le serveur

  NOT_CONNECTED : exception;
  -- Un client non dclare soumet une requte

  INVALID_VIEW : exception;
  -- Une view non attendue a essay de se connecter

  INVALID_STATUS : exception;
  -- Un client transmet une requte incompatible avec son tt
  -- dclar sur le serveur

  INVALID_MASK : exception;
  -- Un client prtend termin une requte qui ne lui a pas t affecte
  
  NO_REGION_AVAILABLE : exception;
  -- Il n'y a plus de rgions  affecter

  NO_MASK_AVAILABLE : exception;
  -- Il n'y a plus de masques  affecter

  NO_SEED_AVAILABLE : exception;
  -- Il n'y a plus de graines  tendre   


--------------------------------------------------------------------------
-- Les types d'actions qui peuvent tre pris en charge par le client    --
--------------------------------------------------------------------------
  type Action_Type is (Region_Computing, Seed_Computing, 
		       Expand_Computing, Alignment_Computing);
  type Action_List is array(Action_Type) of Boolean;
  No_Action   : constant Action_List := (others => False);
  All_Actions : constant Action_List := (others => True);
  By_Regions_Action : constant Action_List := 
    (Region_Computing     => True,
     Seed_Computing       => False,
     Expand_Computing     => False,
     Alignment_Computing  => False);
  
--------------------------------------------------------------------------
-- Les mthodes de prise en charge du traitement des graines            --
--------------------------------------------------------------------------
  type Method_Of_Seed_Computing is (Simple,Combined,Local_Combined);
  -- Dfinit comment doit tre effectu l'expansion des graines
  -- aprs leur gnration. Si la mthode est simple, on calcule
  -- les graines et on les poste  la base de donnes. Si la mthode
  -- est combined, la tache prend aussi en charge l'expansion. Si la
  -- mthode est Local_Combined, la tache prend en charge la gnration
  -- des graines, la vrification de leur prsence dans la base de donnes 
  -- locale mise  jour  partir de la base distante et leur expansion.

--------------------------------------------------------------------------
-- L'identification des clients                                         --
--------------------------------------------------------------------------
  Max_View_Id : constant Natural := 32;
  type View_Id is new Natural range 0..Max_View_Id;
  Unassigned_Id   : constant View_Id := 0;
  Manager_Id      : constant View_Id := 1;
  Dispatcher_Id   : constant View_Id := 2;
  Database_Id     : constant View_Id := 3;
  First_Client_Id : constant View_Id := 4;
  Last_Client_Id  : constant View_Id := View_Id(Max_View_Id);
  subtype Client_Id is View_Id range First_Client_Id..Last_Client_Id;

--------------------------------------------------------------------------
-- L'tat des partitions                                                --
--------------------------------------------------------------------------
  type Status_Type is (Unknown,Waiting,Running,Terminated,Aborted);
  type Partition_State is record
    Status            : Status_Type := Unknown;
    -- Etat
    Times             : Cpu_Duration := Zero;  
    -- Temps d'execution depuis l'laboration de la partition
    Database_Count    : Natural      := 0;     
    -- Nombre de paires effectivement dans la base de donnes
    Constituted_Count : Natural      := 0;
    -- Nombre de couples qui ont t constitus
    Tested_Count      : Natural      := 0;
    -- Nombre de couples qui ont t tests (validit + inclusion)
    Registered_Count  : Natural      := 0;
    -- Nombre de couples dont l'enregistrement dans la base de donnes  
    -- est demand (avant concatenation)    
    Validated_Count   : Natural      := 0;
    -- Nombre de couples dont la validation a t ralise    
  end record;
  type Partition_State_Array is array(Dispatcher_Id..Last_Client_Id) 
    of Partition_State;


--------------------------------------------------------------------------
-- Les paramtres  transmettre au client                               --
--------------------------------------------------------------------------
  type Searching_Parameters is record
    -- Searching parameters
    Xonly                : Boolean;
    Seed_Size            : Integer;
    Mask_Size            : Integer;
    Final_Length         : Integer;
    Recorded_Length      : Integer;
    Distance             : Integer;
    Tolerance            : Integer;
    Skip                 : Integer;
    Common_Max_Dist      : Integer;
    Excluded_Marge       : Integer;
    Intersection_Marge   : Integer;
    -- Alignment parameters
    Alignment            : Boolean;
    Method               : Alignment_Method;
    Minimal_Similarity   : Percentage;
    Match_Weight         : Integer;
    Mismatch_Weight      : Integer;
    Gap_Weight           : Integer;
    Gap_Length_Weight    : Integer;
    Similarity_Threshold : Integer;
    Randomization        : Integer;
    Preserve             : Integer;
    With_End_Weight      : Boolean;
    Max_Length           : Integer;
    Adjust               : Integer;
  end record;

end Dassirc_Common;
