--------------------------------------------------------------------------
-- Dassirc_Database                         Copyrigth ENS ULM 1996-1998 --
-- Pierre VINCENS                   V 00.000                 10/04/1996 --
-- Pierre VINCENS                   V 01.000                 26/03/1998 --
--------------------------------------------------------------------------
with Ada.Calendar;
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;
with Ada.Unchecked_Deallocation;

with System_Constants; use System_Constants;

with Static_Fifo;
with Linked_List;
with Cpu_Times; use Cpu_Times;
with Cpu_Times.Basic; use Cpu_Times.Basic;
with Cpu_Times.Counters; use Cpu_Times.Counters;

with Assirc_Specifications.Database; use Assirc_Specifications.Database;
with Assirc_Database_Functions; use Assirc_Database_Functions;

with Dassirc_Common;
--with Dassirc_Dispatcher;
with Dassirc_Statistics;

pragma Elaborate_All(Static_Fifo);
pragma Elaborate_All(Assirc_Specifications,Cpu_Times.Counters);
pragma Elaborate_All(Dassirc_Statistics);

package body Dassirc_Database is
    
--------------------------------------------------------------------------
-- Mise  jour de la liste des process                                  --
--------------------------------------------------------------------------
  Update_Childs_Required : Boolean := False;

--------------------------------------------------------------------------
-- Instanciation du paquetage Fifo pour grer les piles de paires       --
--------------------------------------------------------------------------
  Max_Pairs_To_Check        : constant := 400000;
  -- Nombre maximal de paires dans la pile d'attente des paires pour lesquelles
  -- on doit vrifier qu'elles n'ont pas dj t test.
  Max_Pairs_To_Expand       : constant := 200000;
  -- Nombre maximal de paires dans la pile d'attente des paires sur laquelle
  -- la marche alatoire doit etre faite
  Threshold_Pairs_To_Check  : constant := Max_Pairs_To_Check / 2;
  -- Seuil au dela duquel on demande au producteurs de graines d'attendre
  -- avant d'en gnrer de nouvelles
  Max_Packets_To_Register : constant := 10000;
  -- Nombre maximal de paquets de paires en attente d'enregistrement
  
  type Segment_Pairs_Access is access Segment_Pairs;
  procedure Free is new Ada.Unchecked_Deallocation(Segment_Pairs,
						   Segment_Pairs_Access);
  type Packet_Of_Pairs is record
    Client : Client_Id;
    Pairs  : Segment_Pairs_Access;
  end record;
  
  package Fifo_Stack_Of_Packets_To_Register is 
    new Static_Fifo(Packet_Of_Pairs,Max_Packets_To_Register);
  package FSOPTR renames Fifo_Stack_Of_Packets_To_Register;
  use FSOPTR;
  package Fifo_Stack_Of_Segment_Pair_To_Check is 
    new Static_Fifo(Segment_Pair,Max_Pairs_To_Check);
  package FSOSPTC renames Fifo_Stack_Of_Segment_Pair_To_Check;
  use FSOSPTC;
  package Fifo_Stack_Of_Segment_Pair_To_Expand is 
    new Static_Fifo(Segment_Pair,Max_Pairs_To_Expand);
  package FSOSPTE renames Fifo_Stack_Of_Segment_Pair_To_Expand;
  use FSOSPTE;

--------------------------------------------------------------------------
-- Diffusion des paires dans les base de donnes locales                --
--------------------------------------------------------------------------
  type Clients_For_Diffusion_Array is array(Client_Id) of Boolean;
  pragma Pack(Clients_For_Diffusion_Array);
  type Account_By_Clients_For_Diffusion_Array is array(Client_Id) of Integer;
  
  type Register_Pairs is record
    Count   : Integer; -- Nombre de paires dans le segment
    Pairs   : Segment_Pairs_Access;
    Clients : Clients_For_Diffusion_Array;
  end record;
  package Register_Pairs_List is new Linked_List(Register_Pairs);
  package PPL renames Register_Pairs_List;

--------------------------------------------------------------------------
-- Un type protg destin  grer les accs simultans  la base       --
--------------------------------------------------------------------------
  protected Similar_Regions is
    --
    -- Primitives de gestion et d'information sur la base centralise
    procedure Set_Parameters(Excluded_Marge     : in Integer;
                             Intersection_Marge : in Integer);
    -- Mmorisation des paramtres
    function Count return Natural;
    -- Obtention du nombre de paires mmorises dans la base de donnes
    --
    -- Primitives d'accs  la base centralise
    function Included(Seg1,Seg2 : in Segment_Type) return Boolean;
    function Are_Included(Pairs : in Segment_Pairs) return Answers_Array;
    -- Test de la presence d'un couple ou d'un tableau de couples
    -- de segments dans la base.
    -- Accs concurrent permis en consultation
    entry Append_Pair_In_Tree(Pair : in Segment_Pair);
    entry Append_Pairs_In_Tree(Pairs : in Segment_Pairs);
    -- Ajout d'une paire ou d'un tableau de paires dans l'arbre
    -- Accs concurrent interdit en mise  jour
    entry Build_Entries;
    -- Contruction d'un tableau regroupant toutes les paires de la base
    entry Release_Entries;
    -- Destruction du tableau prcdent
    function Is_Locked return Boolean;
    -- Pour connaitre l'tat de la variable
    entry Next_Pair(Pair : out Segment_Pair);
    -- Obtention de la paire suivante
    entry Next_Pairs(Minimal_Length : in Integer;
                     Pairs : out Segment_Pairs; 
                     Last : out Integer);
    -- Obtention de Pairs'Length paires suivantes de taille minimal_length
    --
    -- Primitives associes  la diffusion vers les bases locales
    function Count_Calls_For_Diffusion return Integer;
    -- Connaitre le nombre de sollicitation de la base de donnes
    function Diffusion_Is_Active return Boolean;
    -- Indique s'il y a des clients abonns ou non
    entry Add_Client_For_Diffusion(Client : in Client_Id);
    -- Ajout d'un client au service
    entry Remove_Client_For_Diffusion(Client : in Client_Id);
    -- Suppression d'un client du service
    procedure Put_Pairs_For_Diffusion(View  : in View_Id;
				      Pairs : in Segment_Pairs);
    -- Ajout de paquets de paire  la liste en attente de diffusion.
    -- Si view est dans l'intervalle client_id, on le note pour ne pas 
    -- rediffuser vers ce client
    procedure Get_Pairs_For_Diffusion(Client : in View_Id;
				      Pairs  : out Segment_Pairs; 
				      Last   : out Integer;
				      Next   : out Boolean);
    -- Obtenir les paires pour un client donn
  private
    -- Variables associes  la base centralise
    Locked : Boolean := False;
    Data : Similar_Regions_Tree;
    Excluded : Integer := 10;
    Intersection : Integer := 50;
    Entries : Similar_Regions_Entries;
    Current_Entries : Integer;
    -- Variables associes  la diffusion vers les bases locales
    Active     : Boolean := False;
    Nb_Clients : Integer := 0;
    Clients    : Clients_For_Diffusion_Array := (others => False);
    List       : PPL.List_Handler;
    Count_Gets : Integer := 0;
    Account_By_Clients : Account_By_Clients_For_Diffusion_Array 
      := (others => 0);
  end Similar_Regions;

  protected body Similar_Regions is
    procedure Set_Parameters(Excluded_Marge     : in Integer;
                             Intersection_Marge : in Integer) is
    begin
      Excluded := Excluded_Marge;
      Intersection := Intersection_Marge;
    end Set_Parameters;
    function Included(Seg1,Seg2 : in Segment_Type) return Boolean is
    begin
      return Included(Seg1,Seg2,Data,Excluded);
    end Included;
    function Are_Included(Pairs : in Segment_Pairs) return Answers_Array is
      Answers : Answers_Array(Pairs'Range);
    begin
      for I in Pairs'Range loop
        Answers(I) := Included(Pairs(I).Seg1,Pairs(I).Seg2,Data,Excluded);
      end loop;
      return Answers;
    end Are_Included;
    entry Append_Pair_In_Tree(Pair : in Segment_Pair) when not Locked is
    begin
      Append_Pair_In_Tree(Data,Pair,Intersection);
    end Append_Pair_In_Tree;
    entry Append_Pairs_In_Tree(Pairs : in Segment_Pairs) when not Locked is
    begin
      for I in Pairs'Range loop
	Append_Pair_In_Tree(Data,Pairs(I),Intersection);
      end loop;
    end Append_Pairs_In_Tree;
    entry Build_Entries when not Locked and not Active is
    begin
      Locked := True;
      Get_Entries(Data,Entries);
      Current_Entries := Entries'First;
    end Build_Entries;
    entry Release_Entries when Locked is
    begin
      Put_Line("Similar_Regions.Release_Entries: avant free_entries");
      begin
	Free_Entries(Entries);
      exception
	when The_Error : others =>
	  Put_Line("Exception raised in Release_Entries");
	  Put_Line(Ada.Exceptions.Exception_Information(The_Error));
      end;
      Put_Line("Similar_Regions.Release_Entries: apres free_entries");
      Locked := False;
    end Release_Entries;
    function Is_Locked return Boolean is
    begin
      return Locked;
    end Is_Locked;
    entry Next_Pair(Pair : out Segment_Pair) when Locked is
    begin
      Pair := Get_Pair(Entries,Current_Entries);
      Current_Entries := Current_Entries + 1;
    end Next_Pair;  
    entry Next_Pairs(Minimal_Length : in Integer;
                     Pairs : out Segment_Pairs; 
                     Last : out Integer) 
    when Locked is
    begin
      Put_Line("Dassirc_Database.Similar_Regions.Next_Pairs: avant get_pairs");
      Get_Pairs(Entries,Current_Entries,Minimal_Length,Pairs,Last);
    exception
      when OUT_OF_ENTRIES =>
	-- Le transmission d'exception ne se faisant pas correctement
	-- sous Linux, on affecte  Last une valeur nulle
	-- pour indiquer que c'est fini
	Last := 0;
    end Next_Pairs;
    function Count return Natural is
    begin
      return Count_Pairs_In_Tree(Data);
    end Count;
    
    -- Primitives pour la diffusion des paires de regions enregistre
    -- dans la base centralise vers les bases dlocalises
    function Count_Calls_For_Diffusion return Integer is
    begin
      return Count_Gets;
    end Count_Calls_For_Diffusion;
    function Diffusion_Is_Active return Boolean is
    begin
      return Active;
    end Diffusion_Is_Active;

    entry Add_Client_For_Diffusion(Client : in Client_Id) when not Locked is
      Register : Register_Pairs;
      Pairs    : Segment_Pairs(1..Max_Pairs_For_Diffusion);
      Entries : Similar_Regions_Entries;
      Current_Entries : Integer;
    begin
      Put_Line("For_diffusion: ajout du client" & Client_Id'Image(Client));
      Nb_Clients := Nb_Clients + 1;
      Clients(Client) := True;
      if not Active then Active := True; end if;
      -- regarder s'il y a des deja des paires dans la base
      if Count_Pairs_In_Tree(Data) = 0 then
	-- La base est encore vide
	Put_Line("For_diffusion: base vide");
	return;
      end if;
      -- ajouter  la liste les paires de la base actuelle
      -- Attention, cela provoque un blocage partiel des acces  la base
      Register.Clients := (others => False);
      Register.Clients(Client) := True;
      Put_Line("For_diffusion: Gnration des blocks a poster");
      Get_Entries(Data,Entries);
      Current_Entries := Entries'First;
      -- Les accs  la base sont sous la forme d'un tableau linaire
      begin
	loop
	  Put_Line("For_diffusion: recuperation des paires");
	  Get_Pairs(Entries,Current_Entries,0,Pairs,Register.Count);
	  exit when Register.Count = 0;
	  Register.Pairs := new Segment_Pairs(1..Register.Count);
	  Register.Pairs.all := Pairs(1..Register.Count);
	  PPL.Append(List,Register);
	  Put_Line("For_diffusion: Ajout d'un bloc de" & 
		   Integer'Image(Register.Count) & "  la pile");
	  Account_By_Clients(Client) := Account_By_Clients(Client) + 1;
	end loop;
	Put_Line("For_diffusion: sortie de boucle normale");
      exception
	when OUT_OF_ENTRIES => 
	  Put_Line("For_diffusion: sortie de boucle sur OUT_OF_ENTRIES"); 
	  null; -- sortie normale de la boucle
      end;
      Free_Entries(Entries);
      Put_Line("For_diffusion: quitte add_client");
    exception
      when The_Error : others =>
	Put_Line("Exception lev dans Add_Client_For_Diffusion");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Add_Client_For_Diffusion;
    
    entry Remove_Client_For_Diffusion(Client : in Client_Id) when not locked is
      Subtype Data_Null is Boolean; 
      Data : Data_Null; -- idem
      -- requis par le profil du gnrique, inutilis
      procedure Remove_One(Register : in out Register_Pairs;
			   Data     : in out Data_Null;
			   Remove   : out Boolean) is
	Other_Active : Boolean := False;
      begin
	for I in Clients'Range loop
	  if I /= Client then
	    Other_Active := Other_Active or Register.Clients(I);
	  end if;
	end loop;
	if Other_Active then
	  -- Ce paquet doit etre distribu  d'autres clients
	  Register.Clients(Client) := False;
	  Remove := False;
	else
	  -- Ce paquet doit etre supprim
	  Free(Register.Pairs);
	  Remove := True;
	end if;
      end Remove_One;
      procedure Remove_Register_Pairs is new 
	PPL.Apply_And_Remove_When(Data_Null,Remove_One);
    begin
      Put_Line("Remove_Client_For_Diffusion: start client: " &
	       Client_Id'Image(Client));
      if not Clients(Client) then return; end if; -- client non enregistre
      Clients(Client) := False;
      Nb_Clients := Nb_Clients - 1;
      if Nb_Clients = 0 then
	Active := False;
      end if;
      -- supprimer de la liste toutes les paires non encore diffuses 
      -- vers ce client
      Remove_Register_Pairs(List,Data);
      -- et noter qu'il n'y a plus rien pour ce client
      Account_By_Clients(Client) := 0;
      Put_Line("Remove_Client_For_Diffusion: fin client: " &
	       Client_Id'Image(Client));
    end Remove_Client_For_Diffusion;

    procedure Put_Pairs_For_Diffusion(View  : in View_Id;
				      Pairs : in Segment_Pairs) is
      Register : Register_Pairs;
      Account  : Integer := 0; -- pour comptabiliser le nombre de paquets
    begin
      Register.Clients := Clients;
      if View in Client_Id then
	-- Il est inutile de rediffuser vers le client emmetteur
	Register.Clients(View) := False;
      end if;
      if Pairs'Length > Max_Pairs_For_Diffusion then
	-- on dcoupe en plusieurs paquets
	declare 
	  First : Integer := Pairs'First;
	  Last  : Integer := First + Max_Pairs_For_Diffusion - 1;
	begin
	  while First <= Max_Pairs_For_Diffusion loop
	    Register.Pairs := new Segment_Pairs(1..Last-First+1);
	    Register.Pairs.all := Pairs(First..Last);
	    Register.Count := Last - First + 1;
	    PPL.Append(List,Register);
	    Account := Account + 1;
	    First := Last + 1;
	    Last := First + Max_Pairs_For_Diffusion - 1;
	    if Last > Pairs'Last then
	      Last := Pairs'Last;
	    end if;
	  end loop;
	exception
	  when The_Error : others =>
	    Put_Line("Exception raised in Put_Pairs_For_Diffusion (loop)");
	    Put_Line(Ada.Exceptions.Exception_Information(The_Error));
	    raise;
	end;
      else
        begin
	  -- On gnre un seul paquet
 	  Register.Count := Pairs'Length;
	  Register.Pairs := new Segment_Pairs(1..Register.Count);
	  Register.Pairs.all := Pairs;
	  PPL.Append(List,Register);
	  Account := 1;
	exception
	  when The_Error : others =>
	    Put_Line("Exception raised in Put_Pairs_For_Diffusion (fix)");
	    Put_Line(Ada.Exceptions.Exception_Information(The_Error));
	    raise;
	end;
      end if;
      -- Comptabiliser le nombre de paquets ajouts pour chaque client
      for I in Client_Id loop
	if Clients(I) then
	  Account_By_Clients(I) := Account_By_Clients(I) + Account;
	end if;
      end loop;
    exception
      when The_Error : others =>
	Put_Line("Exception raised in Put_Pairs_For_Diffusion");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
	raise;
    end Put_Pairs_For_Diffusion;

    procedure Get_Pairs_For_Diffusion(Client : in View_Id;
				      Pairs  : out Segment_Pairs; 
				      Last   : out Integer;
				      Next   : out Boolean) is
      function Match(Register : in Register_Pairs) return Boolean is
      begin
	return Register.Clients(Client);
      end Match;
      function Find_First_Pair is new PPL.Find_First_Element(Match);
      Ptr : PPL.List_Ptr;
      Other_Active : Boolean := False;
      use PPL;
    begin
      Count_Gets := Count_Gets + 1;
      if not (Client in Client_Id) or else not(Clients(Client)) then
	-- Client non enregistre 
	Last := 0;
	Next := False;
	return; 
      end if;
      Ptr  := Find_First_Pair(List);
      if Ptr = PPL.Nil then
	-- Pas de paires pour ce client
	Last := 0;
	Next := False;
	Put_Line("For_Diffusion.Get_Pairs: pas de paires  diffuser");
	return;
      end if;
      -- retourner l'lment concern
      Last := PPL.Consult(Ptr).Count;
      Pairs(1..Last) := PPL.Consult(Ptr).Pairs(1..Last);
      Put_Line("For_Diffusion.Get_Pairs: envoi de" & Integer'Image(Last) &
	       " paires.");
      -- Declarer qu'il y a un paquet demoins  lire pour ce client
      Account_By_Clients(Client) :=  Account_By_Clients(Client) - 1;
      -- et indiquer au client s'il reste des paquets  lui diffuser
      Next := Account_By_Clients(Client) > 0;
      -- vrifier si l'lment doit etre diffuser sur d'autres clients
      for I in Client_Id'Range loop
	if I /= Client then
	  Other_Active := Other_Active or PPL.Consult(Ptr).Clients(I);
	end if;
      end loop;
      if Other_Active then
	-- indiquer que ce paquet a dj t diffus  ce client
	declare
	  Register : Register_Pairs := PPL.Consult(Ptr);
	begin
	  Register.Clients(Client) := False;
	  PPL.Replace(Ptr,Register);
	end;
      else
	-- pas de diffusion vers d'autres clients, on dtruit
	declare
	  Spa : Segment_Pairs_Access := PPL.Consult(Ptr).Pairs;
	begin
	  Free(Spa);
	end;
	PPL.Delete(List,Ptr);
      end if;
    end Get_Pairs_For_Diffusion;
  end Similar_Regions;
  
--------------------------------------------------------------------------
-- Un type protg pour mmoriser les paires qui doivent tre           --
-- enregistrees dans la base                                            --  
--------------------------------------------------------------------------
  
  protected Request_To_Register is
    entry Start;
    procedure Stop;
    -- Procedure de service
    function Is_Empty return Boolean;
    function Is_Not_Empty return Boolean;
    function Is_Active return Boolean;
    function Is_Terminated return Boolean;
    -- Dtermine si il y a des paires  enregistrer
    entry Wait_End;
    -- Attend que toutes les paires  enregistrer aient t dispatches
    entry Wait_New;
    -- Attend l'arrivee de nouvelles paires
    function Count return Integer;
    -- Indique le nombre de paires en attente
    procedure Get_Pairs(Client : out View_Id;
			Pairs : out Segment_Pairs;
			Last  : out Integer);
    entry Put_Pairs(Client : in View_Id;
		    Pairs  : in Segment_Pairs);
    -- Ajoute ou retire une paire  la pile des paires  enregistrer
    -- Si aucun lment n'est prsent dans la pile, l'erreur STACK_EMPTY
    -- est leve 
  private
    Active : Boolean := False;
    Stack : FSOPTR.Stack_Handler;
  end Request_To_Register;
  
  protected body Request_To_Register is
    -- Activation et dsactivation
    entry Start when not Active is
    begin
      Active := True;
    end Start;
    procedure Stop is
    begin
      Active := False;
    end Stop;
    -- Test de l'tat
    function Is_Empty return Boolean is
    begin
      return Is_Empty(Stack);
    end Is_Empty;
    function Is_Not_Empty return Boolean is
    begin
      return Is_Not_Empty(Stack);
    end Is_Not_Empty;
    -- 
    function Is_Active return Boolean is
    begin
      return Active;
    end Is_Active;
    function Is_Terminated return Boolean is
    begin
      return not Active and then Is_Empty(Stack);
    end Is_Terminated;
    entry Wait_New when Active or else FSOPTR.Is_Not_Empty(Stack) is
    begin
      --Put_Line("Database.Request_To_Register.Wait_New: restart");
      null;
    end Wait_New;
    entry Wait_End when not Active and then FSOPTR.Is_Empty(Stack) is
    begin
      Put_Line("Database.Request_To_Register.Wait_End");
    end Wait_End;
    function Count return Integer is
    begin
      return FSOPTR.Count(Stack);
    end Count;
    -- Ajout ou retrait d'une paire  la pile des inclusions  tester
    procedure Get_Pairs(Client : out View_Id;
			Pairs  : out Segment_Pairs;
			Last   : out Integer) is
      P : Packet_Of_Pairs;
    begin
      Pull(Stack,P);
      Client := P.Client;
      Pairs(P.Pairs'Range) := P.Pairs.all;
      Last := P.Pairs'Last;
      Free(P.Pairs);
    exception
      when The_Error : others =>
	Put_Line("Exception lev dans Request_To_Register.Get_Pair");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Get_Pairs;
    entry Put_Pairs(Client : in View_Id;
		    Pairs  : in Segment_Pairs)
    when Active and FSOPTR.Is_Not_Full(Stack) is
      P : Packet_Of_Pairs;
    begin
      P.Client := Client;
      P.Pairs := new Segment_Pairs(Pairs'Range);
      P.Pairs.all := Pairs;
      Push(Stack,P);
    exception
      when The_Error : others =>
	Put_Line("Exception lev dans Request_To_Check.Put_Pairs");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Put_Pairs;  
  end Request_To_Register;
  
--------------------------------------------------------------------------
-- Une tache prenant en charge l'allocation des paires dans la base     --
--------------------------------------------------------------------------
-- L'ajout des aires dans la base de donnes est un processus 
-- couteux en temps de calcul. Afin d'eviter d'interrompre le
-- client demandant un enregistrement de nouvelles paires
-- la requette d'ajout est transmise a cette tache qui la prend
-- en charge liberant ainsi le client
-------------------------------------------------------------------------- 
  task For_Registering_Pairs is
    entry Start;
  end For_Registering_Pairs;
   
  task body For_Registering_Pairs is
    Pairs : Segment_Pairs(1..Max_Pairs_For_Registering);
    Last : Integer;
    Client : Client_Id;
  begin
    Cpu_Counters.Add("For_Registering_Pairs");     
    Put_Line("Database.For_Registering_Pairs: Initialisation.");
    select
      accept Start;
      Request_To_Register.Start;
    or
      terminate;
    end select; 
    Put_Line("Database.For_Registering_Pairs: Activation (" &
	     Boolean'Image(Request_To_Register.Is_Active) & ").");
    loop
      while Request_To_Register.Is_Not_Empty loop
        Put_Line("Database.For_Registering_Pairs: Traitement d'un paquet (" &
		 Integer'Image(Request_To_Register.Count) & " a traiter).");
	-- On recupre un paquet de paires  traiter sur la pile
        Request_To_Register.Get_Pairs(Client,Pairs,Last);
        -- On les ajoute  la base de donnes
	Similar_Regions.Append_Pairs_In_Tree(Pairs(1..Last));
	if Similar_Regions.Diffusion_Is_Active then
	  -- On enregistre et on note le client qui a transmis
	  -- pour ne pas lui rediffuser
	  Similar_Regions.Put_Pairs_For_Diffusion(Client,Pairs(1..Last));
	end if;
      end loop;
      -- On regarde si c'est termine et on sort si oui
      exit when Request_To_Register.Is_Terminated;
      -- On attend un peu pour laisser s'excuter les autres taches
      delay 0.1;
      Request_To_Register.Wait_New;
    end loop;
    Put_Line("Database.For_Registering_Pairs: Terminaison.");
    Cpu_Counters.Remove;
  end For_Registering_Pairs;
  
--------------------------------------------------------------------------
-- Un type protg destin  grer les requetes de test d'inclusion     --
--------------------------------------------------------------------------
  protected Request_To_Check is
    entry Start;
    procedure Stop;
    -- Active et dsactive l'acceptation de paires  tester
    function Is_Not_Empty return Boolean;
    function Is_Terminated return Boolean;
    entry Wait_New;
    function Count return Integer;
    -- Dtermine si il y a des paires  tendre si le processus est actif
    -- Wait_New permet d'attendre soit la fin, soit l'arrive de nouvelle paire
    procedure Get_Pair(Pair : out Segment_Pair);
    entry Put_Pair(Pair : in Segment_Pair);
    entry Put_Pairs(Pairs : in Segment_Pairs;
		    First : in out Integer);
    -- ajoute ou retire une paire  la pile des inclusions  tester
    -- pour put_pairs, l'argument first donne l'indice de la premire
    -- paire  empiler en input et retourne l'indice de la premire paire
    -- qui devra etre entre ultrieurement en sortie. Ce mcanisme est mis
    -- en place pour tenir compte de l'ventuelle saturation de la pile
  private
    Active : Boolean := False;
    Stack  : FSOSPTC.Stack_Handler;
  end Request_To_Check;

  protected body Request_To_Check is
    -- Activation et dsactivation
    entry Start when not Active is
    begin
      Active := True;
    end Start;
    procedure Stop is
    begin
      Active := False;
    end Stop;
    -- Test de l'tat
    function Is_Not_Empty return Boolean is
    begin
      return Is_Not_Empty(Stack);
    end Is_Not_Empty;
    function Is_Terminated return Boolean is
    begin
      return not Active and then Is_Empty(Stack);
    end Is_Terminated;
    entry Wait_New when not Active or else FSOSPTC.Is_Not_Empty(Stack) is
    begin
      Put_Line("Database.Wait_New: restart");
    end Wait_New;
    function Count return Integer is
    begin
      return FSOSPTC.Count(Stack);
    end Count;
    -- Ajout ou retrait d'une paire  la pile des inclusions  tester
    procedure Get_Pair(Pair : out Segment_Pair) is
    begin
      Pull(Stack,Pair);
    exception
      when The_Error : others =>
	Put_Line("Exception lev dans Request_To_Check.Get_Pair");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Get_Pair;
    entry Put_Pair(Pair : in Segment_Pair) 
    when Active and FSOSPTC.Is_Not_Full(Stack) is
    begin
      Push(Stack,Pair);
    end Put_Pair;
    entry Put_Pairs(Pairs : in Segment_Pairs;
		    First : in out Integer)
    when Active and FSOSPTC.Is_Not_Full(Stack) is
    begin
      while First <= Pairs'Last and then FSOSPTC.Is_Not_Full(Stack) loop
	Push(Stack,Pairs(First));
	First := First + 1;
      end loop;
    exception
      when The_Error : others =>
	Put_Line("Exception lev dans Request_To_Check.Put_Pairs");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
    end Put_Pairs;  
  end Request_To_Check;

--------------------------------------------------------------------------
-- Un type protg destin  grer les requetes d'extension de graines  --
--------------------------------------------------------------------------
  protected Request_To_Expand is
    function Is_Empty return Boolean;
    -- Dtermine si il y a des paires  tendre si le processus est actif
    entry Wait_End;
    -- Attend que toutes les paires  tendre aient t dispatches 
    function Count return Integer;
    -- Indique le nombre de paires en attente
    procedure Get_Pair(Pair : out Segment_Pair);
    procedure Get_Pairs(Pairs : out Segment_Pairs;
			Last  : out Integer);
    entry Put_Pair(Pair : in Segment_Pair);
    -- Ajoute ou retire une paire  la pile des paires  tenter d'tendre
    -- Si aucun lment n'est prsent dans la pile, l'erreur STACK_EMPTY
    -- est leve 
  private
    Stack : FSOSPTE.Stack_Handler;
  end Request_To_Expand;

  protected body Request_To_Expand is
    -- Test de l'tat
    function Is_Empty return Boolean is
    begin
      return Is_Empty(Stack);
    end Is_Empty;
    function Count return Integer is
    begin
      return FSOSPTE.Count(Stack);
    end Count;
    entry Wait_End 
    when Is_Empty(Stack) and then Request_To_Check.Is_Terminated is
    begin
      null;
    end Wait_End;
    -- Ajout ou retrait d'une paire  la pile des inclusions  tenter d'tendre
    procedure Get_Pair(Pair : out Segment_Pair) is 
    begin
      Pull(Stack,Pair);
    end Get_Pair;
    procedure Get_Pairs(Pairs : out Segment_Pairs;
			Last  : out Integer) is 
    begin
      for I in Pairs'Range loop
	begin
	  Pull(Stack,Pairs(I));
	exception
	  when FSOSPTE.STACK_EMPTY => 
	    Last := I - 1;
	    return;
	end;
      end loop;
      Last := Pairs'Last;
    end Get_Pairs;
    entry Put_Pair(Pair : in Segment_Pair) 
    when FSOSPTE.Is_Not_Full(Stack) is
    begin
      Push(Stack,Pair);
    end Put_Pair;
  end Request_To_Expand;



--------------------------------------------------------------------------
-- Une tache prenant en charge les tests d'expansion                    --
--------------------------------------------------------------------------
-- Cette tache prend en charge les graines dans la pile  tester 
-- l'inclusion, teste celle-ci et en fonction du rsultat empile
-- la paire dans la pile des graines  tendre.
-- La condition de sortie de cette tache est que l'envoie de nouvelles
-- paires  tester soient dsactiver et que la pile de rception de
-- ces envois soient vides.
-- ATTENTION, la fin de cette tache est une condition  la validation
-- des fins de traitements des paires  tester l'inclusion
--------------------------------------------------------------------------
  task For_Checking_Inclusion is
    entry Start;
  end For_Checking_Inclusion;

--------------------------------------------------------------------------
-- le corps de la tache prenant en charge les tests d'expansion         --
--------------------------------------------------------------------------
  task body For_Checking_Inclusion is
    Pair  : Segment_Pair;
  begin
    Cpu_Counters.Add("For_Checking_Inclusion");     
    Put_Line("Database: Initialisation de For_Checking_Inclusion");
    select
      accept Start;
    or
      terminate;
    end select;
    Put_Line("Database: Activation de For_Checking_Inclusion");
    loop
      while Request_To_Check.Is_Not_Empty loop
        -- On recupre une paire  traiter sur la pile
        Request_To_Check.Get_Pair(Pair);
        -- On regarde si elle est inclus dans la base de donnes
	if not Similar_Regions.Included(Pair.Seg1,Pair.Seg2) then
	    -- si elle n'y est pas, on l'empile dans les paires  traiter
	  Request_To_Expand.Put_Pair(Pair);
	end if;
      end loop;
      -- On regarde si c'est termine et on sort si oui
      exit when Request_To_Check.Is_Terminated;
      -- On attend un peu pour laisser s'excuter les autres taches
      delay 0.1;
      Request_To_Check.Wait_New;
    end loop;
    Put_Line("Database: Terminaison de For_Checking_Inclusion");
    Cpu_Counters.Remove;
  end For_Checking_Inclusion;


--------------------------------------------------------------------------
-- Une tache informant le dispatcher de l'tat de la base               --
--------------------------------------------------------------------------
  task For_Informing_Dispatcher is
    entry Start;
    entry Stop;
  end For_Informing_Dispatcher;

  task body For_Informing_Dispatcher is
    use Ada.Calendar;
    use Dassirc_Common;
    State : Partition_State;
    Period : constant Duration := 2.0; 
    -- Temps entre deux mise a jour du dispatcher 
    Next_Time : Time;
    Continue : Boolean := True;
  begin
    Cpu_Counters.Add("For_Informing_Dispatcher"); 
    accept Start;
    Next_Time := Clock + Period;
    while Continue loop
      select
	accept Stop;
	Continue := False;
	--exit;
      or
	delay until Next_Time;
      end select;
      Next_Time := Next_Time + Period; 
      if Update_Childs_Required then
	Cpu_Counters.Update_Childs;
	Update_Childs_Required := False;
      end if;
      State.Database_Count := Similar_Regions.Count;
      Cpu_Counters.Elapsed_Times(State.Times);
      --Cpu_Counters.Print_Process_List;
--       Put_Line("To_Check = " & 
-- 	       Integer'Image(Request_To_Check.Count) &
-- 	       "  To_Expand = " & 
-- 	       Integer'Image(Request_To_Expand.Count) &
-- 	       "  Gets = " & 
-- 	       Integer'Image(Similar_Regions.Count_Calls_For_Diffusion));
      Dassirc_Statistics.Advancement(Database_Id,State);
    end loop;
    Cpu_Counters.Remove; 
  end For_Informing_Dispatcher;

--------------------------------------------------------------------------
-- Partie Publique                                                      --
--------------------------------------------------------------------------

--------------------------------------------------------------------------
-- Initiation et terminaison                                            --
--------------------------------------------------------------------------
  procedure Start(Excluded_Marge     : in Integer;
                  Intersection_Marge : in Integer) is
    -- Positionnement des paramtres de gestion  de la base de donnes
  begin
    Similar_Regions.Set_Parameters(Excluded_Marge,Intersection_Marge);
    For_Informing_Dispatcher.Start;
    -- On active l'enregistrement des paires via la tache
    For_Registering_Pairs.Start; 
  end Start;

  procedure Stop is
  begin
    For_Informing_Dispatcher.Stop;
    --abort For_Informing_Dispatcher;
  end Stop;

  procedure Update is
  begin
    -- Noter qu'il faudra mettre a jour la liste des processus 
    -- (rcupration des fils)
    Update_Childs_Required := True;
  end Update;

--------------------------------------------------------------------------
-- Test de l'inclusion de graines dans la base de donnes               --
--------------------------------------------------------------------------
  function Included(Seg1,Seg2 : in Segment_Type) return Boolean is
  begin
    return Similar_Regions.Included(Seg1,Seg2);
  end Included;

  function Are_Included(Pairs : in Segment_Pairs) return Answers_Array is
  begin
    --Put_Line("Database: accepting Are_Included Pairs ->" & 
    --         Integer'Image(Pairs'Length));
    return(Similar_Regions.Are_Included(Pairs));
  end Are_Included;
  
  procedure To_Check_Inclusion_Start is
  begin
    Request_To_Check.Start;
    For_Checking_Inclusion.Start;
  end To_Check_Inclusion_Start;

  procedure To_Check_Inclusion_Stop is
  begin
    Request_To_Check.Stop;
  end To_Check_Inclusion_Stop;

  function No_Pair_To_Expand return Boolean is
  begin
    return (Request_To_Check.Is_Terminated and then 
            Request_To_Expand.Is_Empty and then
            For_Checking_Inclusion'Terminated);
  end No_Pair_To_Expand;

  procedure To_Check_Inclusion(Pairs         : in Segment_Pairs;
			       Wait_For_Next : out Boolean) is
    First : Integer := Pairs'First;
  begin
    while First <= Pairs'Last loop
      Request_To_Check.Put_Pairs(Pairs,First);
    end loop;
    Wait_For_Next := Request_To_Check.Count > Threshold_Pairs_To_Check;
  end To_Check_Inclusion;

  procedure Next_Pair_To_Expand(Pair : out Segment_Pair) is
  begin
    Request_To_Expand.Get_Pair(Pair);
  exception
    when FSOSPTE.STACK_EMPTY => raise NO_PAIR_AVAILABLE; 
  end Next_Pair_To_Expand;

  procedure Next_Pairs_To_Expand(Pairs : out Segment_Pairs;
                                 Last  : out Integer) is
  begin
    Request_To_Expand.Get_Pairs(Pairs,Last);
  end Next_Pairs_To_Expand;

  procedure Wait_End_Of_Process is
  begin
    loop
      -- Attendre que les piles soient vides
      Request_To_Expand.Wait_End;
      -- mais aussi que la tache de traitement des inclusions soient
      -- termine pour viter le cas certes peu probable o on a 
      -- une paire dpile des paires  tester l'inclusion, en cours
      -- de traitement et pas encore empile dans la pile des paires
      --  tendre.
      exit when For_Checking_Inclusion'Terminated;
    end loop;
  end Wait_End_Of_Process;
  
--------------------------------------------------------------------------
-- Diffusion des paires dans les base de donnes locales                --
--------------------------------------------------------------------------
  procedure Add_Client(Client : in Client_Id) is
  begin
    Similar_Regions.Add_Client_For_Diffusion(Client);
  exception
    when The_Error : others =>
      Put_Line("Exception lev dans Add_client");
      Put_Line(Ada.Exceptions.Exception_Information(The_Error));
  end Add_Client;

  procedure Remove_Client(Client : in Client_Id) is
  begin
    Put_Line("REMOVE_CLIENT: IN");
    Similar_Regions.Remove_Client_For_Diffusion(Client);
    Put_Line("REMOVE_CLIENT: DONE");
  end Remove_Client;

  procedure Last_Pairs_From_Client(Client : in View_Id;
				   Pairs  : out Segment_Pairs; 
				   Last   : out Integer;
				   Next   : out Boolean) is
  begin
    Similar_Regions.Get_Pairs_For_Diffusion(Client,Pairs,Last,Next);
  end Last_Pairs_From_Client;

--------------------------------------------------------------------------
-- Ajout de paires dans la base de donnes                              --
--------------------------------------------------------------------------
--   procedure Append_Pair_In_Tree(Pair : in Segment_Pair) is
--   begin
--     Similar_Regions.Append_Pair_In_Tree(Pair);
--   end Append_Pair_In_Tree;

  procedure Append_Pairs_In_Tree(Pairs : in Segment_Pairs) is
  begin
    Request_To_Register.Put_Pairs(Unassigned_Id,Pairs);
    --Similar_Regions.Append_Pairs_In_Tree(Pairs);
    --if Similar_Regions.Diffusion_Is_Active then
      -- On enregistre pour diffusion sans prendre en compte
      -- l'identificateur du client qui est inconnu
      --Similar_Regions.Put_Pairs_For_Diffusion(Unassigned_Id,Pairs);
    --end if;
  end Append_Pairs_In_Tree;

  procedure Append_Pairs_In_Tree(Client : in View_Id;
				 Pairs  : in Segment_Pairs) is
  begin
    Request_To_Register.Put_Pairs(Client,Pairs);
    --Similar_Regions.Append_Pairs_In_Tree(Pairs);
    --if Similar_Regions.Diffusion_Is_Active then
      -- On enregistre et on note le client qui a transmis
      -- pour ne pas lui rediffuser
      --Similar_Regions.Put_Pairs_For_Diffusion(Client,Pairs);
    --end if;
  exception
      when The_Error : others =>
	Put_Line("Exception raised in Append_Pairs_In_Tree");
	Put_Line(Ada.Exceptions.Exception_Information(The_Error));
	raise;
  end Append_Pairs_In_Tree;
  
  procedure Registering_Is_Achieved is
  begin
    Request_To_Register.Stop;
  end Registering_Is_Achieved;
  
  procedure Wait_End_Of_Registering is
  begin
    Request_To_Register.Wait_End;
  end Wait_End_Of_Registering;

--------------------------------------------------------------------------
-- Consultation linaire de la base de donnes                          --
--------------------------------------------------------------------------
  procedure Build_Entries is
  begin
    Similar_Regions.Build_Entries;
  end Build_Entries;

  procedure Release_Entries is
  begin
    Similar_Regions.Release_Entries;
  end Release_Entries;

  procedure Next_Pair(Pair : out Segment_Pair) is
  begin
    Similar_Regions.Next_Pair(Pair);
  end Next_Pair;

  procedure Next_Pairs(Minimal_Length : in Integer;
                       Pairs : out Segment_Pairs;
                       Last  : out Integer) is
  begin
    Put_Line("Dassirc_database.next_pairs : avant Similar_Regions.Next_Pairs");
    Similar_Regions.Next_Pairs(Minimal_Length,Pairs,Last);  
    Put_Line("Dassirc_database.next_pairs : apres Similar_Regions.Next_Pairs");
  end Next_Pairs;

begin
  Cpu_Counters.Add("Database");
end Dassirc_Database;
