--------------------------------------------------------------------------
-- Dassirc_Dispatcher                       Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 01.000                 06/03/1998 --
-- Pierre VINCENS                   V 01.001                 29/11/2000 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use Strlib;
with Cpu_Times.Counters; use Cpu_Times.Counters;
with Options_Util; use Options_Util;

with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;

with Dassirc_Channel; use Dassirc_Channel;

with Assirc_Global_Variables; use Assirc_Global_Variables;

with Dassirc_Dispatcher_Management; use Dassirc_Dispatcher_Management;
with Dassirc_Database;

pragma Elaborate_All(Cpu_Times.Counters);

package body Dassirc_Dispatcher is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
  use Type_Specifications.IO.Real_Io;
  use Type_Specifications.Real_Lib;

--------------------------------------------------------------------------
-- Procdures accessibles au client                                     --
--------------------------------------------------------------------------

  procedure Start is
  begin
    -- Initialiser le type protg Dispatcher
    Dispatcher.Start;
  end Start;

  procedure Stop is
  begin
    Dispatcher.Stop;
    Cpu_Counters.Remove; 
  end Stop;

  procedure Wait_The_Last_Client is
  begin
    Dispatcher.Wait_The_Last;
  end Wait_The_Last_Client;

  procedure Client_Registering
    (Client           : out Client_Id;
     Supported_Action : in out Action_List;
     Computing_Method : in Method_Of_Seed_Computing)
  is
    Error : Exception_Id;
  begin
    Dispatcher.New_Client(Client,Supported_Action,Computing_Method,Error);
    Reraise_Exception(Error);
    Dassirc_Database.Update;
    Put_Line("In Dassirc_Dispatcher: Registering a new lient");
  end Client_Registering;

  procedure Client_Aborting(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.View_State;
    Dispatcher.Abort_Client(Client,Error);
    Reraise_Exception(Error);
    Put_Line("Dassirc_Dispatcher: Aborting client" & Client_Id'Image(Client));
  end Client_Aborting;
  
  procedure Client_Disconnecting(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.View_State;
    Dispatcher.Close_Client(Client,Error);
    Reraise_Exception(Error);
    Put_Line("Dassirc_Dispatcher: Disconnecting client" & Client_Id'Image(Client));
  end Client_Disconnecting;
  
  function Get_Index_Of_Last_Sequences return Accession_Type is
  begin
    return Sequences'Last;
  end Get_Index_Of_Last_Sequences;

  procedure Get_Sequences_Info(Index   : in Accession_Type;
                               Id      : out Accession_Type;
                               Strand  : out Strand_Type;
                               From    : out Integer;
                               To      : out Integer;
			       Length  : out Integer) is
  begin
    Dispatcher.Ready;
    Id := Sequences(Index).Id;
    Strand := Sequences(Index).Strand;
    From := Sequences(Index).Nucleotides'First;
    To  := Sequences(Index).Nucleotides'Last;
    Length  := Sequences(Index).Length;
  end Get_Sequences_Info;

  function Get_Sequences_Identifier(Index   : in Accession_Type)
				    return String is
  begin
    Dispatcher.Ready;
    return To_String(Sequences(Index).Identifier);
  end Get_Sequences_Identifier;
  
  procedure Get_Sequence(Client  : in Accession_Type; 
                         From,To : in Natural;
                         Host    : in String;
                         Port    : in Positive) is
  begin
    Dispatcher.Ready;
    Send_Sequence(Sequences(Client).Nucleotides,From,To,Host,Port);
  end Get_Sequence;

  function Get_Parameters return Searching_Parameters is
    Parameters : Searching_Parameters := 
      (Xonly                => Options.Xonly,
       Seed_Size            => Options.Seed_Size,
       Mask_Size            => Options.Mask_Size,  
       Final_Length         => Options.Final_Length,
       Recorded_Length      => Options.Recorded_Length,
       Distance             => Options.Distance,
       Tolerance            => Options.Tolerance,
       Skip                 => Options.Skip,
       Common_Max_Dist      => Options.Common_Max_Dist,
       Excluded_Marge       => Options.Excluded_Marge,
       Intersection_Marge   => Options.Intersection_Marge,
       Alignment            => Options.Alignment,
       Method               => Options.Alignment_Method,
       Minimal_Similarity   => Percentage(Options.Min_Similarity),
       Match_Weight         => Options.Match_Weight,
       Mismatch_Weight      => Options.Mismatch_Weight,
       Gap_Weight           => Options.Gap_Weight,
       Gap_Length_Weight    => Options.Gap_Length_Weight,
       With_End_Weight      => Options.With_End_Weight,
       Similarity_Threshold => Options.Similarity_Threshold,
       Randomization        => Options.Randomization,
       Preserve             => Options.Preserve,
       Max_Length           => Options.Max_Length,
       Adjust               => Options.Adjust);
  begin
    Dispatcher.Ready;
    return Parameters;
  end Get_Parameters;
  
  function Get_Searching_Shift return Shift_Type is
  begin
    Dispatcher.Ready;
    return Shift.all;
  end Get_Searching_Shift;

  procedure Get_Region(Client    : in Client_Id;
                       Zone1     : out Segment_Type;
		       Zone2     : out Segment_Type;
		       Z1_Xonly  : out Boolean;
		       Z2_Xonly  : out Boolean) is
    Error : Exception_Id;
  begin
    Dispatcher.Assign_Region(Client,Zone1,Zone2,Z1_Xonly,Z2_Xonly,Error);
    Reraise_Exception(Error);
  end Get_Region;

  procedure Region_Completed(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.Terminate_Region(Client,Error);
    Reraise_Exception(Error);
  end Region_Completed;

  function Get_Seed_Mask(Client : in Client_Id) return Integer is
    Mask : Integer;
    Error : Exception_Id;
  begin
    Dispatcher.Assign_Mask(Client,Mask,Error);
    Reraise_Exception(Error);
    Put_Line("Dispatcher: assigning mask" & Integer'Image(Mask) &
	     " to client " & Client_Id'Image(Client));
    return Mask;
  end Get_Seed_Mask;

  procedure Mask_Completed(Client : in Client_Id; Mask : in Integer) is
    Error : Exception_Id;
  begin
    Dispatcher.Terminate_Mask(Client,Mask,Error);
    Reraise_Exception(Error);
  end Mask_Completed;

  procedure Expand_Computing_Start(Client : in Client_Id) is
    Error : Exception_Id ;
  begin
    Dispatcher.Assign_Expand_Computing(Client,Error);
    Reraise_Exception(Error);
    Put_Line("Dispatcher: assigning expand computing to client " & 
	     Client_Id'Image(Client));
  end Expand_Computing_Start;

  procedure Expand_Computing_Stop(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.Terminate_Expand_Computing(Client,Error);
    Reraise_Exception(Error);
  end Expand_Computing_Stop;

  procedure Alignment_Computing_Start(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.Assign_Alignment_Computing(Client,Error);
    Reraise_Exception(Error);
  end Alignment_Computing_Start;

  procedure Alignment_Computing_Stop(Client : in Client_Id) is
    Error : Exception_Id;
  begin
    Dispatcher.Terminate_Alignment_Computing(Client,Error);
    Reraise_Exception(Error);
  end Alignment_Computing_Stop;

  procedure Print_Database(Name           : in String;
                           Minimal_Length : in Integer) is
    -- Le contenu de cette procedure est repris dans assirc_database
    -- Les modifications eventuelles doivent y etre reportees. Le fait
    -- que les pointeurs de fichiers soient limites prives impose  

    File : Ada.Text_Io.File_Type;
    -- Obligatoirement declare global a cette procedure pour
    -- pouvoir fermer le fichier en fin de traitement. 

    function Create_File(Name : in String) return File_Type is
      -- ouvre un fichier name avec terminaison ".sil"
      -- si cela n'est pas possible, on ouvre le fichier dans /tmp
      -- et sinon on ecrit sur la sortie standard
      -- Cette fonction retourne le pointeur sur le fichier
      -- (type limite privee donc affectation interdite)
      Suffix : constant String := ".sil";
    begin 
      if Name'Length > 5 and then Name(Name'Last-4..Name'Last) = Suffix then
	Create(File,Out_File,Name);
      else
	Create(File,Out_File,Name & Suffix);
      end if;
      return File;
    exception
      when Name_Error =>
	declare
	  New_Name : constant String := "/tmp/" & 
	    Root_Name(To_Varstring(Name)) & Suffix;
	begin
	  Put_Line("WARNING: Output data are write in " & New_Name);
	  Create(File,Out_File,New_Name);
	  return File;
	exception
	  when Name_Error =>
	    Put_Line("WARNING: Output data are write to the screen");
	    return Standard_Output;
	end;
    end Create_File;

    procedure Print_The_Pairs(File : in Ada.Text_Io.File_Type) is
      Max_Pairs : constant := Dassirc_Database.Max_Pairs_For_Registering;
      Pairs : Segment_Pairs(1..Max_Pairs);
      Last : Integer;

      procedure Print_A_Pair(Pair : in out Segment_Pair) is      
      begin
	if Pair.Length < Minimal_Length then return; end if;
	Put(File,To_String(Sequences(Pair.Seg1.Id).Identifier));
	Set_Col(File,12);
	Put(File,Image(Pair.Seg1.Strand));
	if Pair.Seg1.Strand = Watson then
	  Put(File,Pair.Seg1.Start,10);
	  Put(File,Pair.Seg1.Stop,10);
	else -- Crick on inscrit desormais dans le meme ordre
	  Put(File,Pair.Seg1.Start,10);
	  Put(File,Pair.Seg1.Stop,10);
	end if;
	Put(File," ");
	Put(File,To_String(Sequences(Pair.Seg2.Id).Identifier));
	Set_Col(File,45);
	Put(File,Image(Pair.Seg2.Strand));
	if Pair.Seg2.Strand = Watson then
	  Put(File,Pair.Seg2.Start,10);
	  Put(File,Pair.Seg2.Stop,10);
	else -- Crick on inscrit desormais dans le meme ordre
	  Put(File,Pair.Seg2.Start,10);
	  Put(File,Pair.Seg2.Stop,10);
	end if;
	Put(File,Pair.Length,10);
	Put(File,Sqrt(Real(Pair.Distance)),8,3,0);
	New_Line(File);
      end Print_A_Pair;

    begin
      Dassirc_Database.Build_Entries;
      Printing: loop
	-- La sortie de la boucle Printing se fait sur 
	-- l'exception OUT_OF_ENTRIES
	Dassirc_Database.Next_Pairs(Minimal_Length,Pairs,Last);
	if Last = 0 then 
	  -- A cause du bug dans les transmissions des exceptions entre
	  -- partition sous Linux
	  -- raise OUT_OF_ENTRIES;
	  exit Printing;
	end if;
	for I in Pairs'First..Last loop
	  Print_A_Pair(Pairs(I));
	end loop;
      end loop Printing;
      Dassirc_Database.Release_Entries;
      Put_Line("Dispatcher: Putative pairs have been printed.");
    end Print_The_Pairs;

  begin
    -- On doit etre prt  imprimer
    Print_The_Pairs(Create_File(Name));
    if Is_Open(File) then
      -- On ne doit pas fermer Standard_Output donc c'est ok
      -- car le pointeur de fichier global file n'a pas ete ouvert
      Close(File);
    end if;
  end Print_Database;
  
  procedure Wait_End_Of_Search is
  begin
    -- Attendre que toutes les paires aient t testes
    Put_Line("Dispatcher.Wait_End_Of_Search: " &
	     "Attendre la fin de la recherche.");
    Dispatcher.Ready_To_Align;
    Put_Line("Dispatcher.Wait_End_Of_Search: " &
	     "toutes les paires ont ete decouvertes.");
    -- Attendre que toutes les paires aient ete enregistrees 
    -- dans la base de donnees
    Dassirc_Database.Registering_Is_Achieved;
    Put_Line("Dispatcher.Wait_End_Of_Search: " &
	     "notification de la fin a la base.");
    Dassirc_Database.Wait_End_Of_Registering;
    Put_Line("Dispatcher.Wait_End_Of_Search: " &
	     "la base de donnees a terminee ses enregistrements.");
  end Wait_End_Of_Search;

  procedure Start_Alignment is
  begin
    -- Dire que l'on est pret  recevoir
    Dispatcher.Activate_Alignment;
  end Start_Alignment;

  procedure Append_Summaries(Summaries : in Alignment_Summaries_Array) is
  begin
    Dispatcher.Append_Summaries(Summaries);
  end Append_Summaries;
  
  procedure Stop_Alignment is
  begin
    -- Dire que l'alignement est termin et que l'on peut 
    Dispatcher.Terminate_Alignment;
  end Stop_Alignment;

  procedure Print_Valid_Pairs(Name : in String) is
    -- Cette procedure imprime les paires similaires trouves  
    File : Ada.Text_Io.File_Type;
    -- Obligatoirement declare global a cette procedure pour
    -- pouvoir fermer le fichier en fin de traitement. 

    function Create_File(Name : in String) return File_Type is
      -- ouvre un fichier name avec terminaison ".sim"
      -- si cela n'est pas possible, on ouvre le fichier dans /tmp
      -- et sinon on ecrit sur la sortie standard
      -- Cette fonction retourne le pointeur sur le fichier
      -- (type limite privee donc affectation interdite)
      Suffix : constant String := ".sim";
    begin 
      if Name'Length > 5 and then Name(Name'Last-4..Name'Last) = Suffix then
	Create(File,Out_File,Name);
      else
	Create(File,Out_File,Name & Suffix);
      end if;
      return File;
    exception
      when Name_Error =>
	declare
	  New_Name : constant String := "/tmp/" & 
	    Root_Name(To_Varstring(Name)) & Suffix;
	begin
	  Put_Line("WARNING: Output data are write in " & New_Name);
	  Create(File,Out_File,New_Name);
	  return File;
	exception
	  when Name_Error =>
	    Put_Line("WARNING: Output data are write to the screen");
	    return Standard_Output;
	end;
    end Create_File;

  begin
    -- On est donc maintenat prt  imprimer
    Put_Line("Dispatcher: activation de l'impression des paires valides");
    Dispatcher.Print_Summaries(Create_File(Name));
    if Is_Open(File) then
      -- On ne doit pas fermer Standard_Output donc c'est ok
      -- car le pointeur de fichier global file n'a pas ete ouvert
      Close(File);
    end if;
    Put_Line("Dispatcher: fin de l'impression");
  end Print_Valid_Pairs;
  
  function Terminated return Boolean is
  begin
    return Dispatcher.Program_Terminated;
  end Terminated;

begin
  Cpu_Counters.Add("Dispatcher");
end Dassirc_Dispatcher;
