--------------------------------------------------------------------------
-- Dassirc_Dispatcher_Management            Copyrigth ENS ULM 1996-2000 --
-- Pierre VINCENS                   V 01.000                 29/11/2000 --
--------------------------------------------------------------------------
with Ada.Exceptions;
with Cpu_Times.Counters; use Cpu_Times.Counters;
with Generic_Quicksort;
with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Alignment.Print; use Sequence_Alignment.Print;
with Assirc_Global_Variables; use Assirc_Global_Variables;
with Dassirc_Database;
with Dassirc_Statistics;

pragma Elaborate_All(Cpu_Times.Counters);

package body Dassirc_Dispatcher_Management is

--------------------------------------------------------------------------
-- Procdures de reprises des exceptions (suite a un bug sous linux)    --
--------------------------------------------------------------------------
  procedure Reraise_Exception(Id : in Exception_Id) is
  begin
    case Id is
      when Null_Exception              => return;
      when Not_Connected_Exception     => raise NOT_CONNECTED;
      when Invalid_Status_Exception    => raise INVALID_STATUS;
      when Invalid_Mask_Exception      => raise INVALID_MASK;
      when Too_Many_Clients_Exception  => raise TOO_MANY_CLIENTS;
      when No_Seed_Available_Exception => raise NO_SEED_AVAILABLE;
    end case;
  end Reraise_Exception;

--------------------------------------------------------------------------
-- Gestion du dispatcher  l'aide d'un type protg                     --
--------------------------------------------------------------------------
  protected body Dispatcher is

    procedure Set_From(Position : in out Position_Type);
    procedure Set_To(Position : in out Position_Type);
    pragma Inline(Set_From,Set_To);
      
    procedure Set_From(Position : in out Position_Type) is
    begin
      Put("Mode region - Set_From - Id :" & 
	  Accession_Type'Image(Position.Id) &
	  " From : avant: " & Integer'Image(Position.From));
      Position.From := Position.From - Options.Regions_Overlap;
      if Position.From < Sequences(Position.Id).Start then
	Position.From := Sequences(Position.Id).Start;
      end if;
      Put_Line(" apres : " & Integer'Image(Position.From));
    end Set_From;
      
    procedure Set_To(Position : in out Position_Type) is
    begin
      Put("Mode region - Set_To - Id :" & 
	  Accession_Type'Image(Position.Id) &
	  " To : avant: " & Integer'Image(Position.To));
      Position.To := Position.From +
	Options.Regions_Length - 1;
      if Position.To + Minimal_Region > Sequences(Position.Id).Stop then
	-- On s'assure que la region  traiter ne dpasse pas la taille
	-- de la squence et aussi que l'on ne va pas laisser moins que 
	-- minimal_region  traiter la prochaine fois
	Position.To := Sequences(Position.Id).Stop;
      end if;
      Put_Line(" apres : " & Integer'Image(Position.To));
    end Set_To;

    procedure Check_Terminating_Region;
    pragma Inline(Check_Terminating_Region);

    procedure Check_Terminating_Region is
      -- voir s'il y en a encore pour lesquelles on attend les rsultats
      -- ATTENTION, on doit faire ce test a la fois dans la partie
      -- de distribution (assign_region) et dans la partie de
      -- validation des zones (terminate_region). La raison est que
      -- l'on ne peut etablir le fait que tout a ete traite que par
      -- un appel a assign_region generant une exception
      -- NO_REGION_AVAILABLE. Cet appel ne donnera pas lieu a une
      -- validation par Terminate_Region. Donc si toutes les regions
      -- sont achevees avant que cet appel ait lieu (cas d'une seule 
      -- partition active de traitement), terminate_region ne sera jamais
      -- appele ulterieurement. Inversement, si cet appel alors que 
      -- des regions sont en cours de calcul, le test devra etre effectue
      -- dans le terminate_region.
    begin
      if Position_1.Id > Sequences'Last then
        -- dans le cas contraire, il reste des regions  dispatcher
        for I in Clients'Range loop
          if Clients(I).Running(Region_Computing) then
            return; -- il y a donc une tache inacheve
          end if;
        end loop;
        -- Les calculs sont finis
        State := Completed;
      end if;
    end Check_Terminating_Region;

    entry Start when State = Waiting is
    begin
      -- On est pret a dispatcher
      State := Dispatching;
      -- Initialiser les variables regissant la distribution par masque
      Last_Mask := (Nb_Symbol ** Options.Mask_Size) - 1;
      -- Initialiser les variables regissant la distribution des expansions
      Expand_Computing_Required := False;
      -- Initialiser les variables regissant la distribution par region
      if Options.Regions_Mode then
	Put_Line("Activation du mode de recherche par region");
	Minimal_Region := Options.Regions_Overlap * 2;
	-- Premier region 1  traiter
        Position_1.Id := Sequences'First;
	Position_1.From := Sequences(Position_1.Id).Start;
        Set_To(Position_1);
	-- La seconde rgion est identique  la premire
	-- sauf s'il y a un seul segment
	Position_2.Id := Sequences'First;
	if Sequences'Length = 1 and then
	  Position_1.To = Sequences(Position_1.Id).Stop then
	  -- 1 seul segment, donc le second est positionne
	  -- avant pour quau bout du compte, les deux soient
	  -- superposs. Cette situation ne devrait pas
	  -- arriver
	  Position_2.To := Sequences(Position_2.Id).Start - 1;
	else
	  -- La region 2 devrait se terminer avec le decalage
	  -- de chevauchement par rapport  la region 1
	  -- On ajoute donc Minimal_Region  la Position de fin
	  -- deregion 1
	  Position_2.To := Position_1.To;-- + Minimal_Region;
	end if;
      end if;
      Put_Line("In dispatcher: start completed");
    end Start;

    entry Stop when State >= Validated is
    begin
      State := Terminated;
    end Stop;

    entry Ready when State >= Dispatching is
    begin
      null;
    end Ready;

    entry New_Client(Client : out Client_Id;
                     Action : in out Action_List;
                     Method : in Method_Of_Seed_Computing;
		     Error  : out Exception_Id) 
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      Put_Line("In dispatcher: new client");
      for I in Clients'Range loop 
	Put_Line("New_client - Test:" & Client_Id'Image(I));
        if not Clients(I).Connect and then
	  Dassirc_Statistics.Get_Status(I) = Unknown 
	then
          Put_Line("New_client - Assign: " & Client_Id'Image(I));
	  -- premire place libre
	  -- On declare maintenant que ce client est connecte
	  Clients(I).Connect := True;
	  Dassirc_Statistics.Connect(I);
	  -- On definit maintenant ses capacites
	  Clients(I).Method  := Method;
	  Clients(I).Running := No_Action;
	  Put_Line("New_client - 1");
	  if Options.Regions_Mode then
	    -- Peu importe l'option du client, le serveur use de sa
	    -- prerogative pour forcer le mode
            Action := By_Regions_Action;
          else
            -- sinon on prend en compte les desiderata du client
            -- on interdit cependant le mode par region (exclusif)
            Action(Region_Computing) := False;
          end if;
	  -- Quel que soit le client, on le force a participer  l'alignement
	  -- de validation (a revoir)
	  Action(Alignment_Computing) := True;
          Put_Line("New_client - 2");
	  Clients(I).Accepted := Action;
          -- On tiend a jour son status
          Dassirc_Statistics.Set_Status(I,Waiting);
          -- Le client est connect mais n'a pas de taches assignes
          Client := I;    
          Put_Line("In dispatcher: new client accepted");
          return;
        end if;
      end loop;
      Put_Line("In dispatcher: Too many clients raised");
      --raise TOO_MANY_CLIENTS;
      Error := Too_Many_Clients_Exception;
    end New_Client;

    procedure Abort_Client(Client : in Client_Id;
			   Error  : out Exception_Id) is
      Job : Job_Aborted_Type;
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
	Put_Line("Dassirc_Dispatcher: Client" & client_id'Image(Client) &
		 " not connected. Exception raised.");
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      -- il reste  traiter le cas du client RUNNING
      for I in Clients(Client).Running'Range loop
        if Clients(Client).Running(I) then
          -- Ce client a une tache en cours non close
	  if I = Region_Computing then
	    -- Gestion du decoupage par region, on memorise
	    -- la region pour une reaffectation ulterieure
	    -- WARNING: On suppose que cet etat est le seul valide
	    Job.Action := Region_Computing;
	    Job.Method := Clients(Client).Method;
	    Job.Pairs  := Clients(Client).Pairs;
	    Push(Jobs_Aborted,Job);
	    Clients(Client).Running(Region_Computing) := False;
	    -- Il n'y a plus d'actions supportees
	    Clients(Client).Accepted := No_Action;
	    -- L'etat du client est considere comme aborte
	    Dassirc_Statistics.Set_Status(Client,Aborted);
	  else
	    -- C'est une tache que l'on se sait pas aborte
	    Put_Line("Dassirc_Dispatcher: Client" & client_id'Image(Client) &
		     " has at least a running task. Exception raised.");
	    --raise INVALID_STATUS;
	    Error := Invalid_Status_Exception;
	    return;
	  end if;
        end if;
      end loop;
      -- On tiend a jour son status
      Dassirc_Statistics.Set_Status(Client,Aborted);
      -- et noter l'information de sa decnnection
      Dassirc_Statistics.Disconnect(Client);
      Clients(Client).Connect := False;
      -- On regarde s'il y a encore un client connecte
      Clients_Connected := False;
      for I in Clients'Range loop
	Clients_Connected := Clients_Connected or Clients(I).Connect;
      end loop;
      Put_Line("Clients_connected: " & Boolean'Image(Clients_Connected));
    end Abort_Client; 

    entry Close_Client(Client : in Client_Id;
		       Error  : out Exception_Id) 
    when State >= Dispatching is
    begin
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
	Put_Line("Dassirc_Dispatcher: Client" & client_id'Image(Client) &
		 " not connected. Exception raised.");
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      -- il reste  traiter le cas du client RUNNING
      for I in Clients(Client).Running'Range loop
        if Clients(Client).Running(I) then
          -- Ce client a une tache en cours non close
          Put_Line("Dassirc_Dispatcher: Client" & client_id'Image(Client) &
		   " has at least a running task. Exception raised.");
	  --raise INVALID_STATUS;
	  Error := Invalid_Status_Exception;
	  return;
        end if;
      end loop;
      -- et noter l'information
      Dassirc_Statistics.Disconnect(Client);
      Clients(Client).Connect := False;
      -- On tiend a jour son status
      Dassirc_Statistics.Set_Status(Client,Terminated);
      -- On regarde s'il y a encore un client connecte
      Clients_Connected := False;
      for I in Clients'Range loop
	Clients_Connected := Clients_Connected or Clients(I).Connect;
      end loop;
      Put_Line("Clients_connected: " & Boolean'Image(Clients_Connected));
    end Close_Client; 

    entry Wait_The_Last when not Clients_Connected is
    begin
      null;
    end Wait_The_Last;

    entry Assign_Region(Client    : in Client_Id;
			Zone1     : out Segment_Type;
                        Zone2     : out Segment_Type;
                        Z1_Xonly  : out Boolean;
                        Z2_Xonly  : out Boolean;
			Error     : out Exception_Id)
    when State >= Dispatching and Position_1.Id /= Accession_Type'First is
      
      
      
      procedure Increment_Id2 is
        -- Incrementation de Current_Id2 et par voie de consequence
        -- si necessaire Current_Id1
      begin
        Position_2.Id := Position_2.Id + 1;
        if Position_2.Id > Sequences'Last then
          -- On a atteind pour le deuxieme indice la derniere sequence
          P2_Xonly := True;
	  -- Tous les segments 2 ont maintenant ete traites contre eux-memes
	  -- cette variable est desormais mise a faux pour toute la suite
	  -- des couples dfe regions a traiter.
    	  -- On progresse maintenant sur la sequence 1
	  if not (Position_1.Id in Sequences'Range) or else
	    Position_1.To >= Sequences(Position_1.Id).Stop then
	    -- On a atteind la fin de la sequence 1
	    -- On incremente l'identificateur de la sequence 
	    Position_1.Id := Position_1.Id + 1;
	    if Position_1.Id > Sequences'Last then
	      -- On a fini, toutes les taches ont t affectes
	      -- ATTENTION, il faut verifier s'il y a des clients actifs
	      -- si ce n'est pas le cas, il faut indiquer que c'est fini
	      -- en position completed_region, sinon personne ne le
	      -- positionne. Ce probleme se pose seulement s'il y a un 
	      -- seul client il faut revoir cela aussi dans les autres cas.
	      Check_Terminating_Region;
	      raise NO_REGION_AVAILABLE;
	    else
	      -- On reprend le traitement sur la sequence Position_1.Id  
	      Position_1.From := Sequences(Position_1.Id).Start;
	      Position_1.To   := Position_1.From +
		Options.Regions_Length - 1;
	      if Position_1.To + Minimal_Region > 
		Sequences(Position_1.Id).Stop 
	      then
		-- On s'assure que la region  traiter ne dpasse pas la 
		-- taille de la squence et aussi que l'on ne va pas 
		-- laisser moins que minimal_region  traiter la 
		-- prochaine fois
		Position_1.To := Sequences(Position_1.Id).Stop;
		-- Dans ce cas, le traitement se fait sur la sequence 
		-- complete. On demarre donc la sequence 2 a 
		-- l'identificateur au dessus
		Position_2.Id := Position_1.Id + 1;
		if Position_2.Id > Sequences'Last then
		  -- Si Position_1.Id > Sequences'First, cette sequence 
		  -- a deja ete consideree dans son entier en tant
		  -- que sequence2, il est donc inutile de la traiter
		  -- (On a alors Position_1.Id egal Sequences'Last-1) 
		  if Position_1.Id = Sequences'First then
		    -- On ne doit traiter qu'une sequence
		    -- On affecte un numero invalide a Position_2.Id
		    Position_2.Id := Sequences'Last + 1;
		  else
		    -- C'est donc fini
		    Position_1.Id := Sequences'Last + 1;
		    -- On regarde s'il reste des regions a tester
		    Check_Terminating_Region;
		    raise NO_REGION_AVAILABLE;
		  end if;
		else
		  Position_2.From := Sequences(Position_2.Id).Start;
		  Set_To(Position_2);
		end if;
	      else
		Position_2.Id := Position_1.Id;
		Position_2.From := Position_1.To + 1;
		Set_From(Position_2);
		Set_To(Position_2);      
	      end if;
	    end if;
	  else
	    -- On determine la nouvelle position sur la sequence 1
	    -- par incrementation. Cette region ne peut donc en aucun 
	    -- cas etre la premiere
	    Position_1.From := Position_1.To + 1;
	    Set_From(Position_1);
	    Set_To(Position_1);
	    -- Et on reinitialise la sequence 2
	    if Position_1.To >= Sequences(Position_1.Id).Stop then
	      -- On est sur le dernier segment de cette sequence
              -- qui a deja ete traite contre lui-meme. On passe donc
              -- a la sequence suivante pour Position_2
	      Position_2.Id := Position_1.Id + 1;
	      if Position_2.Id > Sequences'Last then
		-- Il n'y a plus de region a traiter, Position_1.id
		-- etait la derniere et on est sur le dernier segment
		-- donc c'est fini
		Position_1.Id := Sequences'Last + 1;
		Check_Terminating_Region;
		raise NO_REGION_AVAILABLE;
	      else
		-- On reprend au premier segment de cette nouvelle sequence
		Position_2.From := Sequences(Position_2.Id).Start;
		Set_To(Position_2);
	      end if;
	    else
	      -- On n'est pas sur le dernier segment de id1. On reprend
	      -- donc a partir du segment suivant sur cette meme sequence
	      Position_2.Id := Position_1.Id;
	      Position_2.From := Position_1.To + 1;
	      Set_From(Position_2);
	      Set_To(Position_2);
	    end if;
	  end if;
        else
          -- On debute en debut de sequences 2
          Position_2.From := Sequences(Position_2.Id).Start;
          Set_To(Position_2);
        end if;
      end Increment_Id2;

      procedure Next_Region is
	-- Determination de la region suivante a traiter
      begin
	-- On progresse sur la sequence 2
	if not (Position_2.Id in Sequences'Range) or else
	  Position_2.To >= Sequences(Position_2.Id).Stop then
	  -- On a atteind la fin de la sequence 2
	  Increment_Id2;
	else
	  -- On determine la nouvelle position sur la sequence 2
	  Position_2.From := Position_2.To + 1;
	  Set_From(Position_2);
	  Set_To(Position_2);
	end if;
      end Next_Region;

    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if Clients(Client).Running(Region_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier que ce client accepte ce type de tche
      if not Clients(Client).Accepted(Region_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      
      -- Attribuer la region a prendre en compte
      if Is_Not_Empty(Jobs_Aborted) then
	-- il y a une region a reaffecter
	declare
	  Job : Job_Aborted_Type := Top(Jobs_Aborted);
	begin
	  if Job.Action = Region_Computing then
	    -- On recupere les caracteristique du job a effectuer
	    Clients(Client).Pairs := Job.Pairs;
	    -- On depile le job
	    Pull(Jobs_Aborted,Job);
	  else
	    -- Ce n'est pas un job de la categorie attendue
	    -- on ne fait rien
	    null;
	  end if;
	end;
      else
	-- On prend en compte une nouvelle region
	-- vrifier l'tt des taches
	-- Determination du debut et de la fin des zones a traiter
	if Options.Xonly then
	  -- on recherche la region suivante appartenant a 2 sequences
	  -- differentes
	  loop
	    Next_Region;
	    exit when Position_1.Id /= Position_2.Id;
	  end loop;
	else
	  -- On traite tout donc on prend la region suivante
	  Next_Region;
	end if;
	-- On va donc traiter ce nouveau couple de regions
	Computerized_Pairs_Of_Regions := Computerized_Pairs_Of_Regions + 1;     
	-- Determination des brins a considerer
	Clients(Client).Pairs.Zone1.Id := Position_1.Id;
	Clients(Client).Pairs.Zone1.Start := Position_1.From;
	Clients(Client).Pairs.Zone1.Stop := Position_1.To;
	Clients(Client).Pairs.Zone2.Id := Position_2.Id;
	Clients(Client).Pairs.Zone2.Start := Position_2.From;
	Clients(Client).Pairs.Zone2.Stop := Position_2.To;
	Clients(Client).Pairs.Z1_Xonly := Computerized_Pairs_Of_Regions > 1;
	Clients(Client).Pairs.Z2_Xonly := P2_Xonly;
      end if;
      -- On affecte les parametres a passer a la procedure appellante
      Zone1 := Clients(Client).Pairs.Zone1;
      Zone2 := Clients(Client).Pairs.Zone2;
      Z1_Xonly := Clients(Client).Pairs.Z1_Xonly;
      Z2_Xonly := Clients(Client).Pairs.Z2_Xonly;
      -- On indique l'affectation
      Put_Line("Affecte au client" & Client_Id'Image(Client) &
	       " la region" & Accession_Type'Image(Zone1.Id) &
	       " de" & Integer'Image(Zone1.Start) & 
	       " a " & Integer'Image(Zone1.Stop) &
	       " contre " & Accession_Type'Image(Zone2.Id) &
	       " de" & Integer'Image(Zone2.Start) & 
	       " a " & Integer'Image(Zone2.Stop) &
	       " Z1_Xonly = " & Boolean'Image(Z1_Xonly) &
      	       " Z2_Xonly = " & Boolean'Image(Z2_Xonly));
      -- et a positionner le statut du client
      Clients(Client).Running(Region_Computing) := True;
    exception
      when NO_REGION_AVAILABLE =>
	-- On indique qu'il n'y a plus rien a traiter en
	-- positionnnant from1 et to1 a la meme valeur la plus negative
	-- Ceci est fait car il semblerait qu'il y ait un bug dans la 
	-- transmission de l'exception
	Zone1.Start := Integer'First;
    	Zone1.Stop := Integer'First;
    end Assign_Region;

    entry Terminate_Region(Client : in Client_Id;
			   Error  : out Exception_Id)
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        Put_Line("Terminate_region: lev NOT_CONNECTED");
	--raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if not Clients(Client).Running(Region_Computing) then
        Put_Line("Terminate_region: lev INVALID_STATUS");
	-- raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- noter la terminaison
      Clients(Client).Running(Region_Computing) := False;
      -- et voir s'il y en a encore pour lesquelles on attend les rsultats
      Check_Terminating_Region;
      Put_Line("Validation de la region enregistree pour le client" & 
	       Client_Id'Image(Client));
    end Terminate_Region;

    entry Assign_Mask(Client : in Client_Id; 
		      Mask : out Integer;
		      Error  : out Exception_Id) 
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if Clients(Client).Running(Seed_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier que ce client accepte ce type de tche
      if not Clients(Client).Accepted(Seed_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier l'tt des taches
      if Current_Mask > Last_Mask then
        -- toutes les taches ont t affectes
	-- En raison d'un bug linux, on transmet une valeur negative
	-- au lieu de lever l'exception
        --raise NO_MASK_AVAILABLE;
	Mask := -1;
	return;
      end if;
      -- affecter la tache au client
      Mask := Current_Mask;
      Clients(Client).Running(Seed_Computing) := True;
      Clients(Client).Mask := Current_Mask;
      -- incrmenter le compteur des tches dispatches
      Current_Mask := Current_Mask + 1;
      -- En fonction de la mthode utilise pour l'extension des graines
      -- on doit activer le flux d'entre pour les paires dont on veut tester
      -- l'inclusion 
      if Clients(Client).Method = Simple and not Expand_Computing_Required then
        Dassirc_Database.To_Check_Inclusion_Start;
        Expand_Computing_Required := True;
      end if;
    end Assign_Mask;

    entry Terminate_Mask(Client : in Client_Id; 
			 Mask   : in Integer;
			 Error  : out Exception_Id)
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if not Clients(Client).Running(Seed_Computing) then
        --raise INVALID_STATUS;;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier l'affectation de la tache  ce client
      if Clients(Client).Mask /= Mask then
        --raise INVALID_MASK;;
	Error := Invalid_Mask_Exception;
	return;
      end if;
      -- noter la terminaison
      Clients(Client).Running(Seed_Computing) := False;
      -- et voir s'il y en a encore pour lesquelles on attend les rsultats
      if Current_Mask > Last_Mask then
        -- dans le cas contraire, il reste des taches  dispatcher
        for I in Clients'Range loop
          if Clients(I).Running(Seed_Computing)  then
            return; -- il y a donc une tache de ce type inacheve
          end if;
        end loop;
        Put_Line("Dispatcher: Fin du traitement des masques");
        -- Il n'y aura plus de gnration de graines
        if Expand_Computing_Required then
          -- On informe la base de donnes qu'elle ne recevra plus
          -- de graines pour valuation de l'inclusion si ncessaire
          -- ATTENTION, il peut rester dans la base des graines
          -- pour lesquelles l'extension n'a pas t ralise
          Dassirc_Database.To_Check_Inclusion_Stop;
        else
          -- Tous les calculs se sont fait avec gnration des
          -- graines et extension par le mme client
          -- donc tout est fini
          State := Completed;
        end if;
      end if;
    end Terminate_Mask;

    entry Assign_Expand_Computing(Client : in Client_Id;
				  Error  : out Exception_Id)
    -- pour affecter a un client le traitement des extensions
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
	Put_Line("Dispatcher: Assign_Expand_Computing - NOT_CONNECTED"); 
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if Clients(Client).Running(Expand_Computing) then
        Put_Line("Dispatcher: Assign_Expand_Computing - INVALID_STATUS/1"); 
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier que ce client accepte ce type de tche
      if not Clients(Client).Accepted(Expand_Computing) then
        Put_Line("Dispatcher: Assign_Expand_Computing - INVALID_STATUS/2"); 
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- affecter la tache au client
      Clients(Client).Running(Expand_Computing) := True;
    end Assign_Expand_Computing;

    entry Terminate_Expand_Computing(Client : in Client_Id;
				     Error  : out Exception_Id)
    when State >= Dispatching is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if not Clients(Client).Running(Expand_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- noter la terminaison
      Clients(Client).Running(Expand_Computing) := False;
      -- et voir s'il y en a encore des clients ralisant une tache quivalente
      for I in Clients'Range loop
        if Clients(I).Running(Expand_Computing) then
          -- il y en a encore un
          return;
        end if;
      end loop;
      -- il n'y a plus de clients ralisant cette tache
      -- Par scurit, on teste l'tat de la base de donnes
      if Dassirc_Database.No_Pair_To_Expand then
        Put_Line("Dispatcher: fin de traitement des expansions");
        -- Il n'y a plus de paires  traiter
        Expand_Computing_Required := False;
        -- et tout est fini
        State := Completed;
      end if;
    end Terminate_Expand_Computing;

    entry Assign_Alignment_Computing(Client : in Client_Id;
				     Error  : out Exception_Id)
    when State >= Aligning is
      -- pour affecter a un client le traitement des validations par
      -- par alignement
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if Clients(Client).Running(Alignment_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- vrifier que ce client accepte ce type de tche
      if not Clients(Client).Accepted(Alignment_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- affecter la tache au client
      Clients(Client).Running(Alignment_Computing) := True;
    end Assign_Alignment_Computing;

    entry Terminate_Alignment_Computing(Client : in Client_Id;
					Error  : out Exception_Id)
    when State >= Aligning is
    begin
      Error := Null_Exception;
      -- vrifier l'tt du client
      if not Clients(Client).Connect then
        --raise NOT_CONNECTED;
	Error := Not_Connected_Exception;
	return;
      end if;
      if not Clients(Client).Running(Alignment_Computing) then
        --raise INVALID_STATUS;
	Error := Invalid_Status_Exception;
	return;
      end if;
      -- noter la terminaison
      Clients(Client).Running(Alignment_Computing) := False;
      -- et voir s'il y en a encore des clients ralisant 
      --une tache quivalente
      for I in Clients'Range loop
        if Clients(I).Running(Alignment_Computing) then
          -- il y en a encore un
          return;
        end if;
      end loop;
      -- il n'y a plus de clients ralisant cette tache
      -- tout est fini
      State := Validated;
    end Terminate_Alignment_Computing;

    function Program_Terminated return Boolean is
    begin
      return Dispatcher.State = Terminated;
    end Program_Terminated;

    entry Ready_To_Align when State = Completed is
    begin
      null;
    end Ready_To_Align;

    entry Activate_Alignment when State = Completed is
    begin
      -- Librer la liste pour accumuler les rsultats
      Reset(Summaries_List);
      -- Demander  la base de prparer les donnes pour pouvoir les rcuprer
      Put_Line("Dispatcher.Activate_Alignment: avant build_entries");
      Dassirc_Database.Build_Entries;
      Put_Line("Dispatcher.Activate_Alignment: apres build_entries");
      -- Dire que l'on est pret  valider par alignement
      State := Aligning;
    end Activate_Alignment;

    entry Terminate_Alignment when State >= Validated is
    begin
      -- Demander  la base de librer les donnes pour pouvoir les rcuprer
      Put_Line("Dispatcher Terminate_Alignment: avant release_entries");
      Dassirc_Database.Release_Entries;
      Put_Line("Dispatcher Terminate_Alignment: apres release_entries");
      -- Librer la liste pour accumuler les rsultats
      Reset(Summaries_List);
      -- Dire que l'on est pret  valider par alignement
      State := Terminated;
    end Terminate_Alignment;

    entry Append_Summaries(Summaries : in Alignment_Summaries_Array) 
    when State = Aligning is
    begin
      for I in Summaries'Range loop
	Append(Summaries_List,Summaries(I));
      end loop;
    end Append_Summaries;

    entry Print_Summaries(File : in Ada.Text_Io.File_Type) 
    when State = Validated is
      Entries : List_Entries;
      function Inf(Left  : in Alignment_Summary_Model;
		   Right : in Alignment_Summary_Model) 
      return Boolean;
      pragma Inline(Inf);
      function Inf(Left  : in List_Ptr;
		   Right : in List_Ptr)
      return Boolean is
      begin
	if Consult(Left).Name1 < Consult(Right).Name1 then
	  return True;
	elsif  Consult(Left).Name1 > Consult(Right).Name1 then
	  return False;
	else -- Consult(Left).Name1 = Consult(Right).Name1
	  if Consult(Left).Strand1 < Consult(Right).Strand1 then
	    return True;
	  elsif  Consult(Left).Strand1 > Consult(Right).Strand1 then
	    return False;
	  else -- Consult(Left).Strand1 = Consult(Right).Strand1
	    if Consult(Left).Name2 < Consult(Right).Name2 then
	      return True;
	    elsif  Consult(Left).Name2 > Consult(Right).Name2 then
	      return False;
	    else -- Consult(Left).Name2 = Consult(Right).Name2
	      if Consult(Left).Strand2 < Consult(Right).Strand2 then
		return True;
	      elsif  Consult(Left).Strand2 > Consult(Right).Strand2 then
		return False;
	      else -- Consult(Left).Strand2 = Consult(Right).Strand2
		if Consult(Left).Start1 < Consult(Right).Start1 then
		  return True;
		elsif  Consult(Left).Start1 > Consult(Right).Start1 then
		  return False;
		else -- Consult(Left).Start1 = Consult(Right).Start1
		  if Consult(Left).Start2 < Consult(Right).Start2 then
		    return True;
		  elsif  Consult(Left).Start2 > Consult(Right).Start2 then
		    return False;
		  else -- Consult(Left).Start2 = Consult(Right).Start2
		    if Consult(Left).Stop1 < Consult(Right).Stop1 then
		      return True;
		    elsif  Consult(Left).Stop1 > Consult(Right).Stop1 then
		      return False;
		    else -- Consult(Left).Stop1 = Consult(Right).Stop1
		      if Consult(Left).Stop2 < Consult(Right).Stop2 then
			return True;
		      else  -- Consult(Left).Stop2 >= Consult(Right).Stop2
			return False;
		      end if;
		    end if;
		  end if;
		end if;
	      end if;
	    end if;
	  end if;
	end if;
      end Inf;

      package Summaries_Sort is new Generic_Quicksort(List_Ptr,
						      Integer,
						      List_Entries_Array,
						      Inf);
						      
    begin
      Get_Entries(Summaries_List,Entries);
      Summaries_Sort.Quicksort(Entries.all);
      for I in Entries'Range loop
	Print_Summary(File,Consult(Entries(I)));
	New_Line(File);
      end loop;
      Free(Entries);
    end Print_Summaries;
    
    procedure View_State is   
    begin
      Put_Line("Dispatcher State : " & Dispatcher_State'Image(Dispatcher.State));
      Put("Clients state: ");
      for I in Client_Id loop
        case Dassirc_Statistics.Get_Status(I) is
	  when Unknown    => Put('U');
	  when Waiting    => Put('W');
	  when Running    => Put('R');
	  when Terminated => Put('T');
	  when Aborted    => Put('A');
	end case;
      end loop;
      New_Line;
    end View_State;
    
  end Dispatcher;


end Dassirc_Dispatcher_Management;
