--------------------------------------------------------------------------
-- Fifo                                     Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 04/05/1993 --
-- Pierre VINCENS                   V 01.001                 31/03/1998 --
-- Pierre VINCENS                   V 01.002                 28/12/1998 --
-- Pierre VINCENS                   V 01.003                 04/03/1999 --
--------------------------------------------------------------------------
with Unchecked_Deallocation;

package body Fifo is

  procedure Free is new Unchecked_Deallocation(Stack_Node,Stack_Ptr);

  procedure Push(Stack : in out Stack_Handler; Element : in Element_Type) is
    Ptr : Stack_Ptr := new Stack_Node'(null,Element);
  begin 
    if Stack.Bottom /= null then
      Stack.Bottom.Adr := Ptr;  
    end if;
    Stack.Bottom := Ptr;
    if Stack.Top = null then
      Stack.Top := Ptr;
    end if;
    Stack.Size := Stack.Size + 1;
  end Push;

  procedure Pull(Stack : in out Stack_Handler; Element : out Element_Type) is
    Ptr : Stack_Ptr := Stack.Top;
  begin
    if Ptr = null then
      raise STACK_EMPTY;
    end if;
    Element := Ptr.Element;
    Stack.Top := Ptr.Adr;
    if Ptr = Stack.Bottom then
      Stack.Bottom := null;
    end if;
    Stack.Size := Stack.Size - 1;
    Free(Ptr);
  end Pull;

  procedure Top(Stack : in Stack_Handler; Element : out Element_Type) is
  begin
    if Stack.Top = null then
      raise STACK_EMPTY;
    end if;
    Element := Stack.Top.Element;
  end Top;

  function Top(Stack : in Stack_Handler) return Element_Type is
  begin
    if Stack.Top = null then
      raise STACK_EMPTY;
    end if;
    return Stack.Top.Element;
  end Top;

  function Is_In(Stack : in Stack_Handler;
		 Value : in Element_Type) return Boolean is
    Ptr : Stack_Ptr := Stack.Top;
  begin
    while Ptr /= null loop
      if Feature(Ptr.Element,Value) then
	return True;
      end if;
      Ptr := Ptr.Adr;
    end loop;
    return False;
  end Is_In;

  procedure Reset(Stack : in out Stack_Handler) is
    Ptr : Stack_Ptr := Stack.Top;
  begin
    while Ptr /= null loop
      Stack.Top := Ptr.Adr;
      Free(Ptr);
      Ptr := Stack.Top;
    end loop;
    Stack.Bottom := null;
    Stack.Size := 0;
  end Reset;

  procedure Clean_And_Reset(Stack : in out Stack_Handler) is
    Ptr : Stack_Ptr := Stack.Top;
  begin
    while Ptr /= null loop
      Stack.Top := Ptr.Adr;
      Clean(Ptr.Element);
      Free(Ptr);
      Ptr := Stack.Top;
    end loop;
    Stack.Bottom := null;
    Stack.Size := 0;
  end Clean_And_Reset;
  
  function Is_Empty(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Top = null;
  end Is_Empty;

  function Is_Not_Empty(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Top /= null;
  end Is_Not_Empty;

  function Count(Stack : in Stack_Handler) return Natural is
  begin
    return Stack.Size;
  end Count;

  function Force_Count(Stack : in Stack_Handler) return Natural is
    Ptr : Stack_Ptr := Stack.Top;
    Nb  : Natural := 0;
  begin
    while Ptr /= null loop
      Ptr := Ptr.Adr;
      Nb := Nb + 1;
    end loop;
    return Nb;
  end Force_Count;

end Fifo;
  
