--------------------------------------------------------------------------
-- Fifo                                     Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 04/05/1993 --
-- Pierre VINCENS                   V 01.001                 31/03/1998 --
-- Pierre VINCENS                   V 01.002                 28/12/1998 --
-- Pierre VINCENS                   V 01.003                 04/03/1999 --
--------------------------------------------------------------------------
-- Gestion d'une pile FIFO (First In First Out)
-- Ce gnrique doit etre instancie avec le type de donnees a stocker
--------------------------------------------------------------------------
-- V 01.001
-- Modifications mineures
--   o Le nombre d'lments de la pile est maintenu dans un champ du 
--   o handler afin d'viter d'avoir  la recalculer.
--------------------------------------------------------------------------
-- V 01.002
-- Modifications mineures 
--   o Ajout de la fonction top
--   o Suppression du mode in out pour le paramtre stack dans la procedure 
--     top
--------------------------------------------------------------------------
-- V 01.003
-- Modifications mineures 
--   o Ajout de la fonction generique is_in
--------------------------------------------------------------------------
generic
  type Element_Type is private;
package Fifo is

  type Stack_Handler is limited private;

  STACK_EMPTY: exception;

  procedure Push(Stack : in out Stack_Handler; Element : in Element_Type);
  pragma Inline(Push);
  -- empiler un element
  
  procedure Pull(Stack : in out Stack_Handler; Element : out Element_Type);
  pragma Inline(Pull);
  -- depiler un element
  
  procedure Top(Stack : in Stack_Handler; Element : out Element_Type);
  pragma Inline(Top);
  -- recopier l'element en haut de la pile
  
  function Top(Stack : in Stack_Handler) return Element_Type;
  pragma Inline(Top);
  -- consulter l'element en haut de la pile

  generic
    with function Feature(A,B : in Element_Type) return Boolean;
  function Is_In(Stack : in Stack_Handler;
		 Value : in Element_Type) return Boolean;
  pragma Inline(Is_In);
  -- applique la fonction feature  chaque lment de la pile
  -- et retourne vrai  la premire occurence trouv pour laquelle
  -- feature retourne vrai. La principale application est de vrifier
  -- si un lment est dj dans la pile
    
  procedure Reset(Stack : in out Stack_Handler);
  pragma Inline(Reset);
  -- reinitialiser la pile
  
  generic
    with procedure Clean(Element : in out Element_Type);
  procedure Clean_And_Reset(Stack : in out Stack_Handler);
  pragma Inline(Clean_And_Reset);
  -- reinitialiser la pile apres avoir applique la procedure Clean
  -- sur chacun des elements
  
  function Is_Empty(Stack : in Stack_Handler) return Boolean;
  pragma Inline(Is_Empty);
  function Is_Not_Empty(Stack : in Stack_Handler) return Boolean;
  pragma Inline(Is_Not_Empty);
  -- tester si la pile est vide ou non vide
  
  function Count(Stack : in Stack_Handler) return Natural;
  pragma Inline(Count);
  -- compte le nombre d'elements present de la pile.

private

  type Stack_Node;
  type Stack_Ptr is access Stack_Node;
  type Stack_Node is record
    Adr     : Stack_Ptr;
    Element : Element_Type;
  end record;

  type Stack_Handler is record
    Top,Bottom : Stack_Ptr := null;
    Size : Integer := 0;
  end record;

end Fifo;
  
