--------------------------------------------------------------------------
-- Linked_List                              Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 22/10/1993 --
-- Pierre VINCENS                   V 01.001                 25/05/1998 --
-- Pierre VINCENS                   V 01.002                 30/06/1999 --
--------------------------------------------------------------------------
-- Gestion d'une liste en chainage double
-- Ce generique doit etre instancie avec le type de donnees a stocker
--------------------------------------------------------------------------
-- V 01.001
-- Le paquetage est modifi pour pouvoir etre preelabore.
--------------------------------------------------------------------------
-- V 01.002
-- Correction d'un bug dans la fonction delete. Lors de la suppression
-- du dernier lment de la liste, le champ last du handler pouvait ne
-- pas etre mis  jour correctement
--------------------------------------------------------------------------
generic
  type Element_Type is private;
package Linked_List is

  pragma Preelaborate(Linked_List);

  -- L'affectation d'objet des types suivant doit etre faite avec
  -- precautions. Il s'agit d'adresse vers des objets. Ces adresses
  -- sont amenees a evoluer au cours du temps. La declaration de ces
  -- objets comme limited private limiterait ces risques mais
  -- introduit plusieurs contraintes majeures.
  type List_Handler is private;
  type List_Ptr is private;
  type List_Entries_Array is array(Integer range <>) of List_Ptr;
  type List_Entries is access List_Entries_Array;

  Nil : constant List_Ptr;

  procedure Free(Entries : in out List_Entries);

  procedure Append(List    : in out List_Handler;
                   Element : in Element_Type;
                   Address : in List_Ptr := Nil);
  pragma Inline(Append);
    -- Ajouter un element apres l'element pointe par l'adresse specifiee
    -- Si l'adresse vaut null, l'element est ajoute en fin de liste

  procedure Insert(List    : in out List_Handler;
                   Element : in Element_Type;
                   Address : in List_Ptr := Nil);
  pragma Inline(Insert);
    -- Inserer un element apres l'element pointe par l'adresse specifiee
    -- Si l'adresse vaut null, l'element est ajoute en debut de liste

  function First(List    : in List_Handler) return Element_Type;
  pragma Inline(First);
    -- retourne le premier element de la liste
  function Last(List    : in List_Handler) return Element_Type;
  pragma Inline(Last);
    -- retourne le dernier element de la liste

  function First(List      : in List_Handler) return List_Ptr;
  pragma Inline(First);
    -- retourne l'adresse du premier element de la liste
  function Last(List       : in List_Handler) return List_Ptr;
  pragma Inline(Last);
    -- retourne l'adresse du dernier element de la liste

  function Consult(Address : in List_Ptr) return Element_Type;
  pragma Inline(Consult);
    -- Retourne l'element dont l'adresse a ete specifiee

  procedure Delete(List    : in out List_Handler;
                   Address : in out List_Ptr);
  pragma Inline(Delete);
    -- Supprime l'element dont l'adresse a ete specifiee

  procedure Replace(Address : in List_Ptr;
                    Element : in Element_Type);
  pragma Inline(Replace);
    -- remplace l'ancienne valeur de element par la nouvelle

  generic
    type Data_Command is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
  procedure Modify(Address : in List_Ptr;
                   Data    : in out Data_Command);
  pragma Inline(Modify);
        -- Cette procedure applique la procedure command a la donnee pointee
        -- par address

  generic
    type Data_Command is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in Data_Command);
  procedure Assign(Address : in List_Ptr;
                   Data    : in Data_Command);
  pragma Inline(Assign);
        -- Cette procedure applique la procedure command a la donnee pointee
        -- par address

  generic
    with function Feature(Element : in Element_Type) return Boolean;
  function Find_First_Element(List : in List_Handler) return List_Ptr;
  pragma Inline(Find_First_Element);
    -- Recherche le premier element de la liste verifiant le critere
    -- specifie. Si aucun element n'est trouve, un pointeur null est retourne

  generic
    type Data_Type is private;
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Type) return Boolean;
  function Find_Element(List : in List_Handler;
                        Data : in Data_Type) return List_Ptr;
  pragma Inline(Find_Element);
    -- Recherche le premier element de la liste verifiant le critere
    -- specifie. Si aucun element n'est trouve, un pointeur null est retourne

  generic
    with procedure Command(Element : in out Element_Type);
  procedure Apply(List : in List_Handler);
  pragma Inline(Apply);
    -- Applique la procedure command specifie a l'ensemble des elements
    -- de la liste.

  generic
    type Data_Command is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
  procedure Apply_With_Data(List : in List_Handler;
                            Data : in out Data_Command);
  pragma Inline(Apply_With_Data);
    -- Applique la procedure command specifie a l'ensemble des elements
    -- de la liste. La variable data est passe a chaque appel de command.

  generic
    with procedure Command(Element : in out Element_Type;
                           Ptr     : in List_Ptr);
  procedure Apply_With_Address(List : in List_Handler);
  pragma Inline(Apply_With_Address);
    -- Applique la procedure command specifie a l'ensemble des elements
    -- de la liste. L'adresse de l'element est passee a chaque
    -- appel de command.

  generic
    type Data_Command is private;
    type Data_Feature is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Feature) return Boolean;
  procedure Apply_If(List      : in List_Handler;
                     Condition : in Data_Feature;
                     Data      : in out Data_Command);
  pragma Inline(Apply_If);
    -- Applique la procedure command specifie a l'ensemble des elements
    -- de la liste qui verifie la condition.

  generic
    type Data_Command is private;
    type Data_Feature is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Feature) return Boolean;
  procedure Apply_While(List      : in List_Handler;
                        Condition : in Data_Feature;
                        Data      : in out Data_Command);
  pragma Inline(Apply_While);
    -- Applique la procedure command specifie aux elements
    -- de la liste tant que la condition specifiee est vrai.
        -- La procedure command est appliquee apres le test de la condition

  generic
    type Data_Command is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Command) return Boolean;
  procedure Apply_Until(List      : in List_Handler;
                        Data      : in out Data_Command);
  pragma Inline(Apply_Until);
    -- Applique la procedure command specifie aux elements
    -- de la liste jusqu'a ce que la condition specifiee soit vrai.
        -- La procedure command est appliquee avant le test de la condition

  generic
    type Data_Command is private;
    type Data_Feature is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Command);
    with function Feature(Element : in Element_Type;
                          Data    : in Data_Feature) return Boolean;
  procedure Apply_If_To_First(List      : in List_Handler;
                              Condition : in Data_Feature;
                              Data      : in out Data_Command;
                              Apply     : out Boolean);
  pragma Inline(Apply_If_To_First);
    -- Applique la procedure command specifie au premier element
    -- de la liste qui verifie la condition et retourne apply a true
    -- dans ce cas.

  generic
    type Data_Type is private;
    with procedure Command(Element : in out Element_Type;
                           Data    : in out Data_Type;
                           Remove  : out Boolean);
  procedure Apply_And_Remove_When(List : in out List_Handler;
                                  Data : in out Data_Type);
  pragma Inline(Apply_And_Remove_When);
    -- Applique la procedure command specifie a l'ensemble des elements
    -- de la liste. L'element est supprime si delete est retourne vrai

  procedure Reset(List : in out List_Handler);
  pragma Inline(Reset);
    -- reinitialiser la liste (efface tous les elements)
    -- ATTENTION, cela ne libere pas la place occupee par les elements
    -- eux memes qui seraient pointes par la structure, mais seulement
    -- la place occupee par la liste chainee et l'enregistrement

  generic
    with procedure Clean(Element : in out Element_Type);
  procedure Clean_And_Reset(List : in out List_Handler);
  pragma Inline(Clean_And_Reset);
    -- reinitialiser la liste (efface tous les elements)
    -- la procedure clean est appelee sur chaque element avant sa destruction

  function Is_Empty(List : in List_Handler) return Boolean;
  pragma Inline(Is_Empty);
    -- tester si la liste est vide

  function Count(List : in List_Handler) return Natural;
  pragma Inline(Count);
    -- compte le nombre d'elements presents dans la liste.

  procedure Get_Entries(List : in List_Handler; Entries : in out List_Entries);
  pragma Inline(Get_Entries);
    -- construit un tableau contenant la liste des adresses des elements
    -- ATTENTION: ce tableau n'a de sens que tant qu'aucun element de la
    -- liste n'a ete modifie.


private

  type List_Node;
  type List_Ptr is access List_Node;
  type List_Node is record
    Previous,Next : List_Ptr := null;
    Element       : Element_Type;
  end record;
  Nil : constant List_Ptr := null;

  type List_Handler is record
    First,Last : List_Ptr := null;
    Count      : Natural := 0;
  end record;

end Linked_List;

