--------------------------------------------------------------------------
-- Sequence_Alignment                       Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 01.000                 29/12/1998 --
-- Pierre VINCENS                   V 01.001                 02/06/1999 --
--------------------------------------------------------------------------
-- Dclarations de base utilises pour l'alignement de squence
--------------------------------------------------------------------------
-- V1.001
-- Ce paquetage est dcoup en deux et donne naissance a 
-- Sequence_alignement.Basic. Le but etant de rendre "Pur" ce paquetage 
--------------------------------------------------------------------------
with Type_specifications; use Type_Specifications;
with Sequence_specifications; use Sequence_Specifications;

package Sequence_Alignment is

  pragma Pure;

  Complete : constant Integer := -99999999;

  INCONSISTENT_MODEL : exception;
    -- leve si l'Alignment demande est incompatible (DNA <-> Protein)

  INCONSISTENT_RANGE : exception;
    -- leve si les bornes de l'intervalle des squences ne sont pas
    -- raisonnables.

  INVALID_FORMAT : exception;
    -- leve si les squences ne sont memorises dans le champ
    -- nucleotide ou protein

  subtype Percentage is Real range 0.0..100.0;

  type Alignment_Method is (Global_Alignment,Local_Alignment);
  type Building_Method is (One,All_Firsts,Full);
  type Optimize_Method is (No,Short,Large);
  type Progression_Method is (DVH,DHV,VDH,VHD,HDV,HVD);

  
--------------------------------------------------------------------------
-- Type associ aux rsultats minimal d'un alignement
-- Le type alignment_model donnant le rsultat complet est dans 
-- sequence_alignment-basic. Ce paquetage contient un utilitaire 
-- de conversion. Ce type est essentiellement utilis en environnement
-- distribu.
--------------------------------------------------------------------------
  Summary_Id_Length: constant integer := 15;
  
  type Alignment_Summary_Model is record
    -- identificateur des sequences
    Name1, Name2         : String(1..10);
    Strand1, Strand2     : Strand_Type;
    -- methode utilise pour realiser l'Alignment
    Method               : Alignment_Method := Global_Alignment;
    -- quantificateur de l'Alignment realise
    Quality              : Integer     := 0;
    Quality_Mean         : Real        := 0.0;
    Quality_Std          : Real        := 0.0;
    Quality_Zscore       : Real        := 0.0;
    Quality_Sng          : Real        := 0.0;
    Ratio                : Real        := 0.0;
    Similarity           : Percentage  := 0.0;
    Identity             : Percentage  := 0.0;
    Length               : Integer     := 0;
    Nb_Match             : Integer     := 0;
    Nb_Mismatch          : Integer     := 0;
    Nb_Gaps              : Integer     := 0;
    Gaps_Length          : Integer     := 0;
    Max_Continuous_Match : Integer     := 0;
    -- les positions de depart et de fin dans les squence initiales
    Start1,Stop1         : Integer;
    Start2,Stop2         : Integer;
  end record;
  type Alignment_Summaries_Array is array(Integer range <>) 
    of Alignment_Summary_Model;
  
--------------------------------------------------------------------------
-- Types associs aux matrices de score
--------------------------------------------------------------------------
  subtype Match_Counter is Integer;-- range 0..Integer'Last
  
  type Nucleotide_Match_Matrix is
    array(Nucleotide_Type,Nucleotide_Type) of Match_Counter;

  type Protein_Match_Matrix is
    array(Amino_Acid_Type,Amino_Acid_Type) of Match_Counter;
  

end Sequence_Alignment;
