--------------------------------------------------------------------------
-- Sequence_IO                              Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 05/03/1999 --
--------------------------------------------------------------------------
with Ada.Unchecked_Conversion;
with Ada.Sequential_IO;
with Ada.Text_Io; use Ada.Text_Io;
with Type_Specifications;use Type_Specifications;
with Type_Specifications.Io;use Type_Specifications.Io;
with Strlib,Strlib.Io; use Strlib,Strlib.Io;

package body Sequence_Io is

  Max_Sequence_Length : constant Integer := 100000;
  Max_Line_Length     : constant Integer := 100000;

  use Int_Io;

  package Char_Io is new Ada.Sequential_IO(Character);
  use Char_Io;
  
  type Tabchar is array(Integer range <>) of Character;
  subtype Char_2 is Tabchar(1..2);
  function To_Short is new Ada.Unchecked_Conversion(Char_2,Integer_16);
    
  function Find(Texte : in String; Value : in String) return Boolean is
    Position : Integer;
  begin
    Position := Search(Texte,Value);
    return True;
  exception
    when STRING_NOT_FOUND => return False;
  end Find;

--------------------------------------------------------------------------
-- Lecture des squences nucliques
--------------------------------------------------------------------------
  procedure Read_Nuc_Seq_Text(Name : in String;
                              Sequence: in out Sequence_Data'Class;
			      Field    : in Field_Mode) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    File      : Ada.Text_Io.File_Type;
  begin
    -- determination du nombre de nucleotide
    Open(File, In_File, Name);
    while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      if Length > 0 and then not Is_In_List(Line(1),"->:") then
        for I in 1..Length loop
          if Is_A_Nucleotide(Line(I)) then
            Lseq := Lseq + 1;
          end if;
        end loop;
      end if;
    end loop;
    Close(File);
    -- Lecture effective de la sequence
    Open(File, In_File, Name);
    case Field is
      when Internal =>
	-- Lecture dans Nucleotides comme un tableau de nuclotides
	declare
	Seq : Nucleotidic_Sequence renames Sequence.Nucleotides;
	begin
	  Sequence.Length := Lseq;
	  Seq := new Nucleotidic_Array(1..Lseq);
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 and then not Is_In_List(Line(1),"->:") then
	      for I in 1..Length loop
		if Is_A_Nucleotide(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Nucleotide(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	Seq       : String_Access renames Sequence.Composition.Value;
	begin
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 and then not Is_In_List(Line(1),"->:") then
	      for I in 1..Length loop
		if Is_A_Nucleotide(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Nucleotide(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case;
    Sequence.Name  := To_Varstring(Name);
    --
    Close(File);
  end Read_Nuc_Seq_Text;

  procedure Read_Nuc_Seq_GCG(Name : in String;
                             Sequence: in out Sequence_Data'Class;
			     Field    : in Field_Mode) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Seq       : String_Access renames Sequence.Composition.Value;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    Position  : Integer := 1;
    Mot       : Varstring;
    File      : Ada.Text_Io.File_Type;
  begin
    -- ouverture du fichier
    Open(File, In_File, Name);
    -- determination du nombre de nucleotide
    En_Tete: while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      -- Est-ce que la ligne se termine par ..
      if Length > 10 and then Line(Length-1..Length) = ".." then
        -- on a atteind la ligne de definition de la sequence
        -- le premier mot est le nom de la sequence
        Position := 1;
        Next_Word(Line(Position..Length),Sequence.Name,Position," ");
        Next_Word(Line(Position..Length),Mot,Position);
        if Mot.Value.all = "Length" then
          -- on a bien ce qui est attendu
          Position := Position+1;
          Get(Line(Position..Length),Lseq,Position);
          exit En_Tete;
        end if;
      end if;
    end loop En_Tete;
    -- Lecture effective
    case Field is
      when Internal =>
	-- Lecture dans Nucleotides comme un tableau de nuclotides
	declare
	  Seq : Nucleotidic_Sequence renames Sequence.Nucleotides;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Nucleotidic_Array(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    for I in 11..Length loop
	      if Is_A_Nucleotide(Line(I)) then
		Aseq := Aseq + 1;
		Seq(Aseq) := To_Nucleotide(Line(I));
	      end if;
	    end loop;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq       : String_Access renames Sequence.Composition.Value;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    for I in 11..Length loop
	      if Is_A_Nucleotide(Line(I)) then
		Aseq := Aseq + 1;
		Seq(Aseq) := To_Nucleotide(Line(I));
	      end if;
	    end loop;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case; 
    --
    Close(File);
  end Read_Nuc_Seq_GCG;

  procedure Read_Nuc_Seq_Genbank(Name : in String;
				 Sequence: in out Sequence_Data'Class;
				 Field    : in Field_Mode) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Seq       : String_Access renames Sequence.Composition.Value;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    Position  : Integer := 1;
    Mot       : Varstring;
    File      : Ada.Text_Io.File_Type;
  begin
    -- ouverture du fichier
    Open(File, In_File, Name);
    -- determination du nombre de nucleotide
    En_Tete : while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      -- Est-ce que la ligne commence par "LOCUS"
      if Length > 11 and then Line(1..5) = "LOCUS" then
        -- on a atteind la ligne de definition de la sequence
        -- le premier mot est le nom de la sequence
        Position := 11;
	-- On lit le nom de la squence
        Next_Word(Line(Position..Length),Sequence.Name,Position," ");
	-- On lit la taille
	Position := Position+1;
	Get(Line(Position..Length),Lseq,Position);
	-- On verifie que c'est bien ce qui est attendue
	Position := Position + 1;
        Next_Word(Line(Position..Length),Mot,Position);
        if Mot.Value.all = "bp" then
          -- on a bien ce qui est attendu
	  exit En_Tete;
	else
	  -- Ce n'est pas ce qui tait attendu, on ignore
	  raise INVALID_SEQUENCE_FORMAT;
        end if;
      end if;
    end loop En_Tete;
    -- On saute toute les lignes jusqu'au dbut de la sequence
    Info : while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      if Length >= 6 and then Line(1..6) = "ORIGIN" then
        -- on a atteind la dernire ligne contenant des informations
        exit Info;
      end if;
    end loop Info;
    -- Lecture effective
    case Field is
      when Internal =>
	-- Lecture dans Nucleotides comme un tableau de nuclotides
	declare
	  Seq : Nucleotidic_Sequence renames Sequence.Nucleotides;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Nucleotidic_Array(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    for I in 11..Length loop
	      if Is_A_Nucleotide(Line(I)) then
		Aseq := Aseq + 1;
		Seq(Aseq) := To_Nucleotide(Line(I));
	      end if;
	    end loop;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq       : String_Access renames Sequence.Composition.Value;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    for I in 11..Length loop
	      if Is_A_Nucleotide(Line(I)) then
		Aseq := Aseq + 1;
		Seq(Aseq) := To_Nucleotide(Line(I));
	      end if;
	    end loop;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case; 
    --
    Close(File);
  end Read_Nuc_Seq_Genbank;

  procedure Read_Nuc_Seq_Dna_Strider(Name : in String;
                                     Sequence: in out Sequence_Data'Class;
				     Field    : in Field_Mode) is
    package Char_Io is new Ada.Sequential_IO(Character);
    use Char_Io;
    Header_Length : constant Integer := 112;
    type Tabchar is array(Integer range <>) of Character;
    subtype Char_2 is Tabchar(1..2);
    function To_Short is new Ada.Unchecked_Conversion(Char_2,Integer_16);
    Header    : Tabchar(1..Header_Length);
    File      : Char_Io.File_Type;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    Value     : Character;
    Ch2       : Char_2;
  begin
    -- determination du nombre de nucleotide
    Open(File, In_File, Name);
    -- lire les 112 premiers octets dont on ignore la valeur
    for I in Header'range loop
      Read(File,Header(I));
    end loop;
    Ch2 := Header(31..32);
    Lseq := Integer(To_Short(Ch2));
    -- Enregistrement de la squence
    case Field is
      when Internal =>
	-- Lecture dans Nucleotides comme un tableau de nuclotides
	declare
	  Seq : Nucleotidic_Sequence renames Sequence.Nucleotides;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Nucleotidic_Array(1..Lseq);
	  -- lecture effective de la sequence
	  for I in Seq'Range loop
	    Read(File,Value);
	    if Is_A_Nucleotide(Value) then
	      Aseq := Aseq + 1;
	      Seq(Aseq) := To_Nucleotide(Value);
	    else
	      Put("Lecture d'un caractere anormal. code: ");
	      Put(Character'Pos(Value),4);
	      if Character'Pos(Value) > 32 then
		Put(" Valeur: ");
		Put(Value);
	      end if;
	      New_Line;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq       : String_Access renames Sequence.Composition.Value;
    	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  -- lecture effective de la sequence
	  for I in Seq'Range loop
	    Read(File,Value);
	    if Is_A_Nucleotide(Value) then
	      Aseq := Aseq + 1;
	      Seq(Aseq) := To_Nucleotide(Value);
	    else
	      Put("Lecture d'un caractere anormal. code: ");
	      Put(Character'Pos(Value),4);
	      if Character'Pos(Value) > 32 then
		Put(" Valeur: ");
		Put(Value);
	      end if;
	      New_Line;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case;
    --
    Sequence.Name  := To_Varstring(Name);
    Close(File);
  end Read_Nuc_Seq_Dna_Strider;

--------------------------------------------------------------------------
-- Lecture des squences protiques
--------------------------------------------------------------------------
  procedure Read_Prot_Seq_Text(Name : in String;
			       Sequence: in out Sequence_Data'Class;
			      Field    : in Field_Mode) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    File      : Ada.Text_Io.File_Type;
  begin
    -- determination du nombre de nucleotide
    Open(File, In_File, Name);
    while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      if Length > 0 and then not Is_In_List(Line(1),"->:") then
        for I in 1..Length loop
          if Is_An_Amino_Acid(Line(I)) then
            Lseq := Lseq + 1;
          end if;
        end loop;
      end if;
    end loop;
    Close(File);
    -- lecture effective de la sequence
    Open(File, In_File, Name);
    case Field is
      when Internal =>
	-- Lecture dans Amino_Acids comme un tableau d'acides amins
	declare
	  Seq : Amino_Acids_Sequence renames Sequence.Amino_Acids;
	begin-- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Amino_Acids_Array(1..Lseq);
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 and then not Is_In_List(Line(1),"->:") then
	      for I in 1..Length loop
		if Is_An_Amino_Acid(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Amino_Acid(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
    	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq       : String_Access renames Sequence.Composition.Value;
	begin
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 and then not Is_In_List(Line(1),"->:") then
	      for I in 1..Length loop
		if Is_An_Amino_Acid(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Amino_Acid(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case;
    --
    Sequence.Name  := To_Varstring(Name);
    Close(File);
  end Read_Prot_Seq_Text;

  procedure Read_Prot_Seq_GCG(Name     : in String;
			      Sequence : in out Sequence_Data'Class;
			      Field    : in Field_Mode) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    Position  : Integer := 1;
    Mot       : Varstring;
    File      : Ada.Text_Io.File_Type;
  begin
    -- ouverture du fichier
    Open(File, In_File, Name);
    -- determination du nombre de nucleotide
    En_Tete: while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      -- Est-ce que la ligne se termine par ..
      if Length > 10 and then Line(Length-1..Length) = ".." then
        -- on a atteind la ligne de definition de la sequence
        -- le premier mot est le nom de la sequence
        Position := 1;
        Next_Word(Line(Position..Length),Sequence.Name,Position," ");
        Next_Word(Line(Position..Length),Mot,Position);
        if Mot.Value.all = "Length" then
          -- on a bien ce qui est attendu
          Position := Position+1;
          Get(Line(Position..Length),Lseq,Position);
          exit En_Tete;
        end if;
      end if;
    end loop En_Tete;
    -- Lecture de la sequence
    case Field is
      when Internal =>
	-- Lecture dans Amino_Acids comme un tableau d'acides amins
	declare
	  Seq : Amino_Acids_Sequence renames Sequence.Amino_Acids;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Amino_Acids_Array(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 then
	      for I in 11..Length loop
		if Is_An_Amino_Acid(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Amino_Acid(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq       : String_Access renames Sequence.Composition.Value;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  -- lecture de la sequence
	  while not End_Of_File(File) loop
	    Get_Line(File,Line,Length);
	    if Length > 0 then
	      for I in 11..Length loop
		if Is_An_Amino_Acid(Line(I)) then
		  Aseq := Aseq + 1;
		  Seq(Aseq) := To_Amino_Acid(Line(I));
		end if;
	      end loop;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;	  
    end case;
    --
    Close(File);
  end Read_Prot_Seq_GCG;

  procedure Read_Prot_Seq_Dna_Strider(Name : in String;
                                      Sequence: in out Sequence_Data'Class;
				      Field    : in Field_Mode) is
    Header_Length : constant Integer := 112;
    Header    : Tabchar(1..Header_Length);
    Seq       : String_Access renames Sequence.Composition.Value;
    File      : Char_Io.File_Type;
    Lseq      : Natural := 0;        -- longueur de la sequence a lire
    Aseq      : Natural := 0;        -- longueur de la sequence lue
    Value     : Character;
    Ch2       : Char_2;
  begin
    -- determination du nombre de nucleotide
    Open(File, In_File, Name);
    -- lire les 112 premiers octets dont on ignore la valeur
    for I in Header'range loop
      Read(File,Header(I));
    end loop;
    Ch2 := Header(31..32);
    Lseq := Integer(To_Short(Ch2));
    -- lecture effective de la sequence
    case Field is
      when Internal =>
	-- Lecture dans Amino_Acids comme un tableau d'acides amins
	declare
	  Seq : Amino_Acids_Sequence renames Sequence.Amino_Acids;
	begin
	  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new Amino_Acids_Array(1..Lseq);
	  for I in Seq'Range loop
	    Read(File,Value);
	    if Is_An_Amino_Acid(Value) then
	      Aseq := Aseq + 1;
	      Seq(Aseq) := To_Amino_Acid(Value);
	    else
	      Put("Lecture d'un caractere anormal. code: ");
	      Put(Character'Pos(Value),4);
	      if Character'Pos(Value) > 32 then
		Put(" Valeur: ");
		Put(Value);
	      end if;
	      New_Line;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
      when External =>
	-- Lecture dans composition comme une chaine de caracteres
	declare
	  Seq : String_Access renames Sequence.Composition.Value;
	begin  -- initialisation de la variable sequence
	  Sequence.Length := Lseq;
	  Seq := new String(1..Lseq);
	  for I in Seq'Range loop
	    Read(File,Value);
	    if Is_An_Amino_Acid(Value) then
	      Aseq := Aseq + 1;
	      Seq(Aseq) := To_Amino_Acid(Value);
	    else
	      Put("Lecture d'un caractere anormal. code: ");
	      Put(Character'Pos(Value),4);
	      if Character'Pos(Value) > 32 then
		Put(" Valeur: ");
		Put(Value);
	      end if;
	      New_Line;
	    end if;
	  end loop;
	  Sequence.Start := Seq'First;
	  Sequence.Stop  := Seq'Last;
	end;
    end case;
    --
    Sequence.Name  := To_Varstring(Name);
    Close(File);
  end Read_Prot_Seq_Dna_Strider;

--------------------------------------------------------------------------
-- Procdure de dtermination du model et du format des squences
--------------------------------------------------------------------------
  procedure Check_Sequence(Name   : in String;
			   Model  : out Sequence_model;
			   Format : out Sequence_Format) is
    Line      : String(1..Max_Line_Length);
    Length    : Natural;
    Position  : Integer := 1;
    Mot       : Varstring;
    File      : Ada.Text_Io.File_Type;
  begin
    -- ouverture du fichier
    Open(File, In_File, Name);
    -- lecture
    Model := No;
    Format := Unknown;
    En_Tete: while not End_Of_File(File) loop
      Get_Line(File,Line,Length);
      -- Analyse de la ligne
      -- Est-ce que la ligne commence par "->:"
      if Length > 0 and then Is_In_List(Line(1),"->:") then
	Format := Plain_Ascii;
	while not End_Of_File(File) loop
	  -- On lit la suite du fichier pour voir si le contenu
	  -- permet de supposer s'il s'agit d'une squence nuclique
	  -- ou protique
	  Get_Line(File,Line,Length);
	  if Length > 0 and then not Is_In_List(Line(1),"->:") then
	    for I in 1..Length loop
	      if (not(Is_A_Nucleotide(Line(I))) and then
		  Is_An_Amino_Acid(Line(I))) then
		Model := Protein;
		exit En_Tete;
	      end if;
	    end loop;
	  end if;
	end loop;
	-- On n'a pas trouve d'acides amins, on suppose qu'il s'agit
	-- donc de DNA
	Model := Dna;
	exit En_Tete;
      end if;
      -- Est-ce que la ligne commence par "!!AA" ou "!!NA"
      if Length >= 4 then
	if Line(1..4) = "!!NA" then
	  -- C'est du GCG / nuclique
	  Model := DNA;
	  Format := GCG;
	  exit En_Tete;
	elsif Line(1..4) = "!!AA" then
	  -- C'est du GCG / protique
	  Model := Protein;
	  Format := GCG;
	  exit En_Tete;
	end if;
      end if;
      -- Est-ce que la ligne commence par "LOCUS"
      if Length >= 5 then
	if Line(1..5) = "LOCUS" then
	  -- C'est du GCG / nuclique
	  Model := DNA;
	  Format := Genbank;
	  exit En_Tete;
	end if;
      end if;
      -- Est-ce que la ligne se termine par ..
      if Length > 10 and then Line(Length-1..Length) = ".." then
        -- on a atteind la ligne de definition d'une squence au
        -- format GCG 
	-- le premier mot est le nom de la sequence
        Position := 1;
	for I in 1..7 loop
	  exit when Position > Length;
	  Next_Word(Line(Position..Length),Mot,Position," ");
	end loop;
        Next_Word(Line(Position..Length),Mot,Position," ");
        if Mot.Value.all = "Type:" then
          -- on a bien ce qui est attendu
          Next_Word(Line(Position..Length),Mot,Position," ");
	  if Mot.Value.all = "N" then
	    -- C'est du GCG / nuclique
	    Model := DNA;
	    Format := GCG;
	    exit En_Tete;
	  elsif Mot.Value.all = "N" then
	    -- C'est du GCG / protique
	    Model := Protein;
	    Format := GCG;
	    exit En_Tete;
	  end if;
        end if;
      end if;
    end loop En_Tete;
    Close(File);
  end Check_Sequence;

--------------------------------------------------------------------------
-- Fonction gnrale de lecture des squences 
--------------------------------------------------------------------------
  procedure Read_Sequence(Name     : in String;
                          Sequence : in out Sequence_Data'Class;
                          Model    : in Sequence_model;
                          Format   : in Sequence_Format := Plain_Ascii;
			  Field    : in Field_Mode := External) is
    L_Model  : Sequence_Model := Model;
    L_Format : Sequence_Format := Format;
  begin
    -- Si model = no ou format = unknow alors on essaye de le determiner
    if Model = No or Format = Unknown then
      Check_Sequence(Name,L_Model,L_Format);
      -- Si pas trouv, on reprend au cas o cela aurait t connu
      if L_Model = No then
	L_Model := Model;
      end if;
      if L_Format = Unknown then
	L_Format := Format;
      end if;
      -- Si pas connu, on arrete maintenant
      if L_Model = No or L_Format = Unknown then
	raise INVALID_SEQUENCE_FORMAT;
      end if;
      -- Vrifier la cohrence
      if Model /= No and then L_Model /= Model then
	-- On fait confiance  l'utilisateur
	L_Model := Model;
      end if;
      if Format /= Unknown and L_Format /= Format then
	-- On rale
	raise INVALID_SEQUENCE_FORMAT;
      end if;
    end if;
    -- faire le menage si necessaire
    Clear(Sequence.Name);
    Clear(Sequence.Composition);
    -- initialiser les donnees caracteristiques de la sequence
    Sequence.Model := L_Model;
    Sequence.Format := L_Format;
    -- lire la sequence
    case L_Model is
      when Dna | Rna =>
        case L_Format is
          when Plain_Ascii => Read_Nuc_Seq_Text(Name,Sequence,Field);
          when Dna_Strider => Read_Nuc_Seq_Dna_Strider(Name,Sequence,Field);
          when GCG         => Read_Nuc_Seq_GCG(Name,Sequence,Field);
          when Genbank     => Read_Nuc_Seq_Genbank(Name,Sequence,Field);
          when others      => raise NOT_IMPLEMENTED;
        end case;
      when Protein =>
        case L_Format is
          when Plain_Ascii => Read_Prot_Seq_Text(Name,Sequence,Field);
          when Dna_Strider => Read_Prot_Seq_Dna_Strider(Name,Sequence,Field);
          when GCG         => Read_Prot_Seq_GCG(Name,Sequence,Field);
          when others      => raise NOT_IMPLEMENTED;
        end case;
      when others    => raise NOT_IMPLEMENTED;
    end case;
  end Read_Sequence;

end Sequence_Io;

