--------------------------------------------------------------------------
-- Sequence_Specifications                  Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 12/03/1998 --
-- Pierre VINCENS                   V 01.002                 04/03/1999 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;

package body Sequence_Specifications.basic is

  function Image(Strand : in Strand_Type) return Character is
  begin
    case Strand is
      when Watson => return 'W';
      when Crick  => return 'C'; 
      when Both   => return 'B'; 
    end case;
  end Image;

  procedure Reverse_Sequence_String(Sequence : in out Sequence_Data) is
    Seq   : String_Access renames Sequence.Composition.Value;
    N_Seq : String_Access;
    I     : Integer := Seq'First;
    J,K   : Integer := Seq'Last;
    Tampon : Character;
  begin
    if Seq'First = 1 and Seq'Last = Sequence.Length then
      while I < J loop
	Tampon := Seq(I);
	Seq(I) := Seq(J);
	Seq(J) := Tampon;
	I := I + 1;
	J := J - 1;
      end loop;
    else
      -- L'inversion de squence doit prendre en compte un dcalage de zone
      -- Partie d'un chromosome.
      I := Sequence.Length - Seq'Last + 1;
      J := Sequence.Length - Seq'First + 1;      
      N_Seq := new String(I..J);
      for U in I..J loop
	N_Seq(U) := Seq(K);
	K := K - 1;
      end loop;
      Free(Seq);
      Sequence.Composition.Value := N_Seq;
    end if;
  end Reverse_Sequence_String;

  procedure Reverse_Sequence_Nucleotides(Sequence : in out Sequence_Data) is
    Seq   : Nucleotidic_Sequence renames Sequence.Nucleotides;
    N_Seq : Nucleotidic_Sequence;
    I     : Integer := Seq'First;
    J,K   : Integer := Seq'Last;
    Tampon : Nucleotide_Type ;
  begin
    if Seq'First = 1 and Seq'Last = Sequence.Length then
      while I < J loop
	Tampon := Seq(I);
	Seq(I) := Seq(J);
	Seq(J) := Tampon;
	I := I + 1;
	J := J - 1;
      end loop;
    else
      -- L'inversion de squence doit prendre en compte un dcalage de zone
      -- Partie d'un chromosome.
      I := Sequence.Length - Seq'Last + 1;
      J := Sequence.Length - Seq'First + 1;      
      N_Seq := new Nucleotidic_Array(I..J);
      for U in I..J loop
	N_Seq(U) := Seq(K);
	K := K - 1;
      end loop;
      Free(Seq);
      Sequence.Nucleotides := N_Seq;
    end if;
  end Reverse_Sequence_Nucleotides;

  procedure Reverse_Sequence(Sequence : in out Sequence_Data) is
  begin
    if Sequence.Model /= DNA and Sequence.Model /= RNA then
      raise NOT_NUCLEOTIDIC;
    end if;
    if Sequence.Composition.Value /= null then
      Reverse_Sequence_String(Sequence);
    end if;
    if Sequence.Nucleotides /= null then
      Reverse_Sequence_Nucleotides(Sequence);
    end if;
  end Reverse_Sequence;

  procedure Cmp_Sequence_String(Sequence : in out Sequence_Data) is
    Seq   : String_Access renames Sequence.Composition.Value;
  begin
    for I in Seq'range loop
      case Seq(I) is
        when 'A' => Seq(I) := 'T';
        when 'T' => Seq(I) := 'A';
        when 'G' => Seq(I) := 'C';
        when 'C' => Seq(I) := 'G';
        when 'N' => Seq(I) := 'N';
        when others => null;
      end case;
    end loop;
  end Cmp_Sequence_String;

  procedure Cmp_Sequence_Nucleotides(Sequence : in out Sequence_Data) is
    Seq   : Nucleotidic_Sequence renames Sequence.Nucleotides;
  begin
    for I in Seq'range loop
      case Seq(I) is
        when 'A' => Seq(I) := 'T';
        when 'T' => Seq(I) := 'A';
        when 'G' => Seq(I) := 'C';
        when 'C' => Seq(I) := 'G';
        when 'N' => Seq(I) := 'N';
        when others => null;
      end case;
    end loop;
  end Cmp_Sequence_Nucleotides;

  procedure Cmp_Sequence(Sequence : in out Sequence_Data) is
  begin
    if Sequence.Model /= DNA and Sequence.Model /= RNA then
      raise NOT_NUCLEOTIDIC;
    end if;
    if Sequence.Composition.Value /= null then
      Cmp_Sequence_String(Sequence);
    end if;
    if Sequence.Nucleotides /= null then
      Cmp_Sequence_Nucleotides(Sequence);
    end if;
  end Cmp_Sequence;

  procedure To_Nucleotides(Sequence : in out Sequence_Data) is
    Src : String_Access renames Sequence.Composition.Value;
    Dst : Nucleotidic_Sequence renames Sequence.Nucleotides;
  begin
    if Sequence.Model /= DNA and Sequence.Model /= RNA then
      raise NOT_NUCLEOTIDIC;
    end if;
    if Src = null or else Src'Length = 0 then
      -- la sequence au format string n'existe pas
      return;
    end if;
    -- liberer la place eventuelle occupee par la sequence
    -- au format nucleotidic
    Free(Dst);
    -- allouer la place necessaire
    Dst := new Nucleotidic_Array(Src'Range);
    -- realiser la conversion
    for I in Src'Range loop
      Dst(I) := To_Nucleotide(Src(I));
    end loop;
  exception
    when UNKNOWN_NUCLEOTIDE =>
      -- la sequence n'est pas nucleotidique
      Free(Dst);
      raise NOT_NUCLEOTIDIC;
  end To_Nucleotides;

  procedure To_Amino_Acids(Sequence : in out Sequence_Data) is
    Src : String_Access renames Sequence.Composition.Value;
    Dst : Amino_Acids_Sequence renames Sequence.Amino_Acids;
  begin
    if Sequence.Model /= Protein then
      raise NOT_PROTEINIC;
    end if;
    if Src = null or else Src'Length = 0 then
      -- la sequence au format string n'existe pas
      return;
    end if;
    -- liberer la place eventuelle occupee par la sequence
    -- au format amino_acids
    Free(Dst);
    -- allouer la place necessaire
    Dst := new Amino_Acids_Array(Src'Range);
    -- realiser la conversion
    for I in Src'Range loop
      Dst(I) := To_Amino_Acid(Src(I));
    end loop;
  exception
    when UNKNOWN_AMINO_ACID =>
      -- la sequence n'est pas proteique
      Free(Dst);
      raise NOT_PROTEINIC;
  end To_Amino_Acids;

  procedure Nuc_To_String(Sequence : in out Sequence_Data) is
    Src : Nucleotidic_Sequence renames Sequence.Nucleotides;
    Dst : String_Access renames Sequence.Composition.Value;
  begin
    if Src = null or else Src'Length = 0 then
      -- la sequence au format nucleotidic n'existe pas
      return;
    end if;
    -- liberer la place eventuelle occupee par la sequence
    -- au format string
    Free(Dst);
    -- allouer la place necessaire
    Dst := new String(Src'Range);
    -- realiser la conversion
    for I in Src'Range loop
      Dst(I) := To_Character(Src(I));
    end loop;
  exception
    when UNKNOWN_NUCLEOTIDE =>
      -- la sequence n'est pas nucleotidique
      Free(Dst);
      raise NOT_NUCLEOTIDIC;
  end Nuc_To_String;

  procedure AA_To_String(Sequence : in out Sequence_Data) is
    Src : Amino_Acids_Sequence renames Sequence.Amino_Acids;
    Dst : String_Access renames Sequence.Composition.Value;
  begin
    if Src = null or else Src'Length = 0 then
      -- la sequence au format nucleotidic n'existe pas
      return;
    end if;
    -- liberer la place eventuelle occupee par la sequence
    -- au format string
    Free(Dst);
    -- allouer la place necessaire
    Dst := new String(Src'Range);
    -- realiser la conversion
    for I in Src'Range loop
      Dst(I) := To_Character(Src(I));
    end loop;
  exception
    when UNKNOWN_AMINO_ACID =>
      -- la sequence n'est pas nucleotidique
      Free(Dst);
      raise NOT_PROTEINIC;
  end AA_To_String;

  procedure To_String(Sequence : in out Sequence_Data) is
  begin
    case Sequence.Model is
      when DNA | RNA => Nuc_To_String(Sequence);
      when Protein   => AA_To_String(Sequence);
      when others    => raise NOT_IMPLEMENTED;
    end case;
  end To_String;

  procedure Clear_Nucleotides(Sequence : in out Sequence_Data) is
  begin
    Free(Sequence.Nucleotides);
  end Clear_Nucleotides;

  procedure Clear_Amino_Acids(Sequence : in out Sequence_Data) is
  begin
    Free(Sequence.Amino_Acids);
  end Clear_Amino_Acids;

  procedure Clear_String(Sequence : in out Sequence_Data) is
  begin
    Clear(Sequence.Composition);
  end Clear_String;

  procedure Clear(Sequence : in out Sequence_Data) is
  begin
    Clear(Sequence.Name);
    Free(Sequence.Nucleotides);
    Free(Sequence.Amino_Acids);
    Clear(Sequence.Composition);
    Sequence.Model := NO;
    Sequence.Length := 0;
    Sequence.Start := -1;
    Sequence.Stop := -1;
  end Clear;

end Sequence_Specifications.Basic;

