--------------------------------------------------------------------------
-- Sequence_Specifications-basic            Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 12/03/1998 --
-- Pierre VINCENS                   V 01.001                 04/03/1999 --
--------------------------------------------------------------------------
--
-- Paquetage de dclaration des types et fonctions associes a l'analyse
-- de squences
--
--------------------------------------------------------------------------
-- V 01.001                                                             
-- Ce paquetage contient les parties ne repondant pas au critre pure  
-- qui tait initialement dans le paquetage racine
--------------------------------------------------------------------------
-- V 01.002                                                             
-- Ajout des fonctionnalits de gestion des protines
--------------------------------------------------------------------------
with Type_Specifications; use Type_Specifications;
with Strlib; use Strlib;
with Ada.Unchecked_Deallocation;

package Sequence_Specifications.Basic is

  --pragma Preelaborate;

  type Nucleotidic_Sequence is access Nucleotidic_Array;
  procedure Free is new Ada.Unchecked_Deallocation(Nucleotidic_Array,
                                                   Nucleotidic_Sequence);

  type Amino_Acids_Sequence is access Amino_Acids_Array;
  procedure Free is new Ada.Unchecked_Deallocation(Amino_Acids_Array,
                                                   Amino_Acids_Sequence);

  type Sequence_Data is tagged record
    Name          : Varstring;
    Identifier    : Varstring;
    Model         : Sequence_Model := NO;
    Format        : Sequence_Format := Unknown;
    Length        : Natural := 0;
    Composition   : Varstring;
    Nucleotides   : Nucleotidic_Sequence;
    Amino_Acids   : Amino_Acids_Sequence;
    Start,Stop    : Integer := -1;
  end record;
  -- La sequence peut etre represente dans le champ "Composition"
  -- comme une chaine de caracteres, et dans le champ "Nucleotides"
  -- sur la base d'un format nucleotidic_sequence.

  function Image(Strand : in Strand_Type) return Character;
  -- La fonction image retourne les caractres W, C ou B en fonction 
  -- de la valeur de Strand

  procedure Reverse_Sequence(Sequence : in out Sequence_Data);
  -- retourne la sequence inverse
  -- n'a un sens que pour les squences nucliques

  procedure Cmp_Sequence(Sequence : in out Sequence_Data);
  -- retourne la sequence complementaire
  -- n'a un sens que pour les squences nucliques

  procedure To_Nucleotides(Sequence : in out Sequence_Data);
  -- determine le champ "nucleotides" a partir du champ "composition"

  procedure To_Amino_Acids(Sequence : in out Sequence_Data);
  -- determine le champ "amino_acids" a partir du champ "composition"

  procedure To_String(Sequence : in out Sequence_Data);
  -- determine le champ "composition" a partir du champ "nucleotides"

  procedure Clear_Nucleotides(Sequence : in out Sequence_Data);
  -- libere la place associe au champ "nucleotides"

  procedure Clear_Amino_Acids(Sequence : in out Sequence_Data);
  -- libere la place associe au champ "amino_acids"

  procedure Clear_String(Sequence : in out Sequence_Data);
  -- libere la place associe au champ "composition"

  procedure Clear(Sequence : in out Sequence_Data);
  -- libere la place associe  une squence

  type Sequence_Array is array(Integer range <>) of aliased Sequence_Data;
  type Sequence_Array_Access is access Sequence_Array;
  procedure Free is new Ada.Unchecked_Deallocation(Sequence_Array,
						   Sequence_Array_Access);
  -- Une structure tableau pour mmoriser une collection de squences

end Sequence_Specifications.Basic;
