--------------------------------------------------------------------------
-- Sequence_Specifications                  Copyrigth ENS ULM 1986-1998 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 12/03/1998 --
-- Pierre VINCENS                   V 01.002                 04/03/1999 --
--------------------------------------------------------------------------
package body Sequence_Specifications is

  function Is_A_Space(Value : in Character) return Boolean is
  begin
    return Is_In_List(Value," ");
  end Is_A_Space;

  function Is_A_Nucleotide(Value : in Character) return Boolean is
  begin
    return Is_In_List(Value,"aAtTgGcCuUnN");
  end Is_A_Nucleotide;

  function Is_An_Amino_acid(Value: in Character) return Boolean is
  begin
    return Is_In_List(Value,
                      "aAbBcCdDeEfFgGhHiIkKlLmMnNpPqQrRsStTvVwWxXyYzZ*");
  end Is_An_Amino_Acid;

  function To_Nucleotide(Value : Character) return Character is
  begin
    case Value is
      when 'a' | 'A' => return 'A';
      when 't' | 'T' => return 'T';
      when 'g' | 'G' => return 'G';
      when 'c' | 'C' => return 'C';
      when 'u' | 'U' => return 'U';
      when 's' | 'S' => return 'S';
      when 'w' | 'W' => return 'W';
      when 'r' | 'R' => return 'R';
      when 'y' | 'Y' => return 'Y';
      when 'k' | 'K' => return 'K';
      when 'm' | 'M' => return 'M';
      when 'b' | 'B' => return 'B';
      when 'v' | 'V' => return 'V';
      when 'h' | 'H' => return 'H';
      when 'd' | 'D' => return 'D';
      when 'n' | 'N' => return 'N';
      when others    => raise UNKNOWN_NUCLEOTIDE;
    end case;
  end To_Nucleotide;

  function To_Amino_acid(Value: in Character) return character is
  begin
    case Value is
      when 'A' | 'a' => return 'A';
      when 'B' | 'b' => return 'B';
      when 'C' | 'c' => return 'C';
      when 'D' | 'd' => return 'D';
      when 'E' | 'e' => return 'E';
      when 'F' | 'f' => return 'F';
      when 'G' | 'g' => return 'G';
      when 'H' | 'h' => return 'H';
      when 'I' | 'i' => return 'I';
      when 'K' | 'k' => return 'K';
      when 'L' | 'l' => return 'L';
      when 'M' | 'm' => return 'M';
      when 'N' | 'n' => return 'N';
      when 'P' | 'p' => return 'P';
      when 'Q' | 'q' => return 'Q';
      when 'R' | 'r' => return 'R';
      when 'S' | 's' => return 'S';
      when 'T' | 't' => return 'T';
      when 'V' | 'v' => return 'V';
      when 'W' | 'w' => return 'W';
      when 'X' | 'x' => return 'X';
      when 'Y' | 'y' => return 'Y';
      when 'Z' | 'z' => return 'Z';
      when '*'       => return '*';
      when others    => raise UNKNOWN_AMINO_ACID;
    end case;
  end To_Amino_Acid;

  function Is_In_list(Value : in Character;
                      List  : in String) return Boolean is
  begin
    for I in List'range loop
      if Value = List(I) then
        return True;
      end if;
    end loop;
    return False;
  end Is_In_List;

  function To_Nucleotide(Value : Character) return Nucleotide_Type is
  begin
    case Value is
      when 'a' | 'A' => return 'A';
      when 't' | 'T' => return 'T';
      when 'g' | 'G' => return 'G';
      when 'c' | 'C' => return 'C';
      when 'u' | 'U' => return 'T';
      when 's' | 'S' => return 'S';
      when 'w' | 'W' => return 'W';
      when 'r' | 'R' => return 'R';
      when 'y' | 'Y' => return 'Y';
      when 'k' | 'K' => return 'K';
      when 'm' | 'M' => return 'M';
      when 'b' | 'B' => return 'B';
      when 'v' | 'V' => return 'V';
      when 'h' | 'H' => return 'H';
      when 'd' | 'D' => return 'D';
      when 'n' | 'N' => return 'N';
      when others    => raise UNKNOWN_NUCLEOTIDE;
    end case;
  end To_Nucleotide;

  function To_Character(Value : Nucleotide_Type) return Character is
  begin
    case Value is
      when 'A' => return 'A';
      when 'T' => return 'T';
      when 'G' => return 'G';
      when 'C' => return 'C';
      when 'S' => return 'S';
      when 'W' => return 'W';
      when 'R' => return 'R';
      when 'Y' => return 'Y';
      when 'K' => return 'K';
      when 'M' => return 'M';
      when 'B' => return 'B';
      when 'V' => return 'V';
      when 'H' => return 'H';
      when 'D' => return 'D';
      when 'N' => return 'N';
      when others    => raise UNKNOWN_NUCLEOTIDE;
    end case;
  end To_Character;

  function To_Amino_Acid(Value : Character) return Amino_Acid_Type is
  begin
    case Value is
      when 'A' | 'a' => return 'A';
      when 'B' | 'b' => return 'B';
      when 'C' | 'c' => return 'C';
      when 'D' | 'd' => return 'D';
      when 'E' | 'e' => return 'E';
      when 'F' | 'f' => return 'F';
      when 'G' | 'g' => return 'G';
      when 'H' | 'h' => return 'H';
      when 'I' | 'i' => return 'I';
      when 'K' | 'k' => return 'K';
      when 'L' | 'l' => return 'L';
      when 'M' | 'm' => return 'M';
      when 'N' | 'n' => return 'N';
      when 'P' | 'p' => return 'P';
      when 'Q' | 'q' => return 'Q';
      when 'R' | 'r' => return 'R';
      when 'S' | 's' => return 'S';
      when 'T' | 't' => return 'T';
      when 'V' | 'v' => return 'V';
      when 'W' | 'w' => return 'W';
      when 'X' | 'x' => return 'X';
      when 'Y' | 'y' => return 'Y';
      when 'Z' | 'z' => return 'Z';
      when '*'       => return '*';
      when others    => raise UNKNOWN_AMINO_ACID;
    end case;
  end To_Amino_Acid;

  function To_Character(Value : Amino_Acid_Type) return Character is
   begin
    case Value is
      when 'A' => return 'A';
      when 'B' => return 'B';
      when 'C' => return 'C';
      when 'D' => return 'D';
      when 'E' => return 'E';
      when 'F' => return 'F';
      when 'G' => return 'G';
      when 'H' => return 'H';
      when 'I' => return 'I';
      when 'K' => return 'K';
      when 'L' => return 'L';
      when 'M' => return 'M';
      when 'N' => return 'N';
      when 'P' => return 'P';
      when 'Q' => return 'Q';
      when 'R' => return 'R';
      when 'S' => return 'S';
      when 'T' => return 'T';
      when 'V' => return 'V';
      when 'W' => return 'W';
      when 'X' => return 'X';
      when 'Y' => return 'Y';
      when 'Z' => return 'Z';
      when '*' => return '*';
      when others    => raise UNKNOWN_AMINO_ACID;
    end case;
  end To_Character; 

end Sequence_Specifications;

