--------------------------------------------------------------------------
-- Sequence_Specifications                  Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 12/03/1998 --
-- Pierre VINCENS                   V 01.001                 04/03/1999 --
--------------------------------------------------------------------------
--
-- Paquetage de dclaration des types et fonctions associes a l'analyse
-- de squences
--
--------------------------------------------------------------------------
-- V 01.001                                                             
-- Paquetage decompose en deux parties de manire  pouvoir affecter la
-- proprit Pure au paquetage racine. Les parties ne rpondant pas  
-- ce critre ont t transfre dans le fils basic.
--------------------------------------------------------------------------
-- V 01.002                                                             
-- Ajout des fonctionnalits ncessaires pour reprsenter des proteines
--------------------------------------------------------------------------
with Type_Specifications; use Type_Specifications;

package Sequence_Specifications is

  pragma Pure;

  UNKNOWN_NUCLEOTIDE  : exception;
  UNKNOWN_AMINO_ACID  : exception;
  -- levee si un nucleotide ou un amino-acide n'existe pas.

  EMPTY_SEQUENCE: exception;
  -- levee si l'on tente de faire un calcul sur une sequence vide.

  TOO_SHORT_SEQUENCE: exception;
  -- levee si la sequence est de longueur insuffisante pour faire le calcul.

  NOT_NUCLEOTIDIC : exception;
  -- levee si la sequence nucleotidique attendue ne l'est pas
  
  NOT_PROTEINIC : exception;
  -- levee si la sequence protique attendue ne l'est pas
  
  INVALID_SEQUENCE_FORMAT : exception;
  -- levee si la specification de format est incorrecte
  
  INVALID_SEQUENCE_MODEL : exception;
  -- levee si la specification de modle est incorrecte
  
  INVALID_STRAND : exception;
  -- levee si la specification d'un brin est incorrecte

  NOT_IMPLEMENTED : exception;
  -- levee si fonctionnalite non disponible

  type Nucleotide_Type is ('A','T','G','C','S','W','R','Y','K','M','B',
			   'V','H','D','N');
  for Nucleotide_Type use ('A'=> 0,'T'=> 1,'G'=> 2,'C'=> 3,'S'=> 4,'W'=> 5,
			   'R'=> 6,'Y'=> 7,'K'=> 8,'M'=> 9,'B'=>10,'V'=>11,
			   'H'=>12,'D'=>13,'N'=>14);
  for Nucleotide_Type'Size use 8;
  type Set_Of_Nucleotides is array(Integer range <>) of Nucleotide_Type;

  type Amino_Acid_Type is ('A','B','C','D','E','F','G','H','I',
			   'K','L','M','N','P','Q','R','S','T',
			   'V','W','X','Y','Z','*');
  for Amino_Acid_Type use ('A'=>0,'B'=>1,'C'=>2,'D'=>3,'E'=>4,
			   'F'=>5,'G'=>6,'H'=>7,'I'=>8,'K'=>9,
			   'L'=>10,'M'=>11,'N'=>12,'P'=>13,'Q'=>14,
			   'R'=>15,'S'=>16,'T'=>17,'V'=>18,'W'=>19,
			   'X'=>20,'Y'=>21,'Z'=>22,'*'=>23);
  for Amino_Acid_Type'Size use 8;
  type Set_Of_Amino_Acids is array(Integer range <>) of Amino_Acid_Type;

  type Sequence_Model is (NO, DNA, RNA, Protein);

  type Sequence_Format is (Plain_Ascii,Dna_Strider,GCG,Genbank,Unknown);
  
  type Strand_Type is (Watson, Crick, Both);

  type Nucleotidic_Array is array(Natural range <>) of Nucleotide_Type;
  type Amino_Acids_Array is array(Natural range <>) of Amino_Acid_Type;
  
  function Is_A_Space(Value : in Character) return Boolean;
  function Is_a_Nucleotide(Value : in Character) return Boolean;
  function Is_An_Amino_Acid(Value: in Character) return Boolean;

  function To_Nucleotide(Value : Character) return Character;
  function To_Amino_Acid(Value: in Character) return Character;
  -- Test et conversion d'un nuc ou d'un aa

  function To_Nucleotide(Value : Character) return Nucleotide_Type;
  function To_Character(Value : Nucleotide_Type) return Character;
  -- Conversion des nuclotides en caractres et inversement

  function To_Amino_Acid(Value : Character) return Amino_Acid_Type;
  function To_Character(Value : Amino_Acid_Type) return Character;
  -- Conversion des amino_acides en caractres et inversement

  function Is_In_list(Value : in Character;
                      List  : in String) return Boolean;
  -- verifie si la valeur passee est bien dans la liste donnee

end Sequence_Specifications;
