--------------------------------------------------------------------------
-- SKnife                                   Copyrigth ENS ULM 1996-1997 --
--------------------------------------------------------------------------
-- Pierre VINCENS                   V 01.000                 25/03/1997 --
-- Conversion de squences nucliques.                                  --
-- Gnration de squences alaroires obtenues par mutation d'une       --
-- squence initiale.                                                   --
--------------------------------------------------------------------------
-- Pierre VINCENS                   V 01.001                 05/02/1998 --
-- Dcoupage en fragment de longueur identique avec chevauchement       --
-- possible d'une squence (option interval). Cette option est          --
-- incompatible avec la contatnation de plusieurs fragments. Cette     --
-- opration se fait APRES l'inversion ou la complmentation de la      --
-- squence et APRES l'application des mutations ventuelles.           --
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;
with Ada.Text_Io; use  Ada.Text_Io;
with Ada.Text_Io.Editing; use Ada.Text_Io.Editing;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Exceptions; use Ada.Exceptions;
with Ada.Calendar; use Ada.Calendar;
with Ada.Numerics.Discrete_Random;

with Type_specifications; use Type_specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use  Strlib;
with Strlib.IO; use  Strlib.IO;
with Options_Util; use  Options_Util;
with Linked_List;
with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Io; use  Sequence_Io;
with Options_Util; use  Options_Util;

procedure Sknife is

  use Int_Io;
  use Real_Io;
  use Real_Lib;

  -- dclarations utilises pour la gnration dynamique du nom du fichier
  type Decimal_Position is delta 1.0 digits 12;
  package Decimal_Position_Output is new Decimal_Output
    (Decimal_Position,"$",'X',',','.');
  use Decimal_Position_Output;
  Identificator_Length : constant Integer := 8;
  Pic_For_Name : constant Picture := To_Picture("99999999",False);

  -- dclarations ncessaires  la gestion de la liste des positions
  --  concatner
  type Segment is record
    Start : Integer;
    Stop  : Integer;
  end record;
  package Segment_Linked_List is new Linked_List(Segment);
  package SLL renames Segment_Linked_List;
  use Segment_Linked_List;
  subtype Segment_List is Segment_Linked_List.List_Handler;

  -- dclarations lies aux options du programme
  type Options_Choice_Type is (Extract_Nuc_Ascii,Extract_Nuc_GCG,
                               Extract_Prot_Ascii,Extract_Prot_GCG);
  type Options_Type is record
    Name : Varstring;
    Output : Varstring;
    Start,Stop : Integer := -1;
    length : Integer := 60;
    Choice : Options_Choice_Type := Extract_Nuc_Ascii;
    Format : Sequence_Format := Plain_Ascii;
    Nb_Mutation : Integer := 0;
    Inverse : Boolean := False;
    Complementaire : Boolean := False;
    Initial : Integer := 0;
    Protein : Boolean := False;
    Positions : Segment_List;
    Step_Length : Integer := 0;
    Step_Recover : Integer := 0;
  end record;


  Sequence,Protein : Sequence_Data;
  Options : Options_Type;

  Current : Segment; -- position courante pour un intervalle
  Output_Name : Varstring ; -- nom du fichier pour l'enregistrement

  INVALID_POSITION, INVALID_COUPLE, OUT_OF_SEQUENCE : exception;

  procedure Get_Options(Options : in out Options_Type) is
    Count : Positive := 1;
  begin
    while Count < Argument_Count loop
      if Is_An_Alias(Argument(Count),"-LENGTH",2) then
        -- lire la longueur des lignes voulue
        Count := Count + 1;
        Options.Length := Opt_Int(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-POSITIONS",3) then
        -- lire les positions a extraire
        Read_Positions: declare
          Position : Segment;
          Decouple : Boolean;
        begin
          loop
            Decouple := False;
            Count := Count + 1;
            Position.Start := Opt_Int(Argument(Count));
            Decouple := True;
            Count := Count + 1;
            Position.Stop  := Opt_Int(Argument(Count));
            if Position.Stop < Position.Start then
              raise INVALID_POSITION;
            end if;
            SLL.Append(Options.Positions,Position);
          end loop;
        exception
          when Data_Error =>
            Count := Count - 1;
            if Decouple then
              raise INVALID_COUPLE;
            end if;
        end Read_Positions;
      elsif Is_An_Alias(Argument(Count),"-BEGIN",2) then
        -- lire la position de debut dans la sequence
        Count := Count + 1;
        Options.Start := Opt_Int(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-END",2) then
        -- lire la position de fin dans la sequence
        Count := Count + 1;
        Options.Stop := Opt_Int(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-REVERSE",2) then
        -- indique que l'on inverse la sequence
         Options.inverse := True;
      elsif Is_An_Alias(Argument(Count),"-COMPLEMENTAIRE",2) then
        -- indique que l'on prend la sequence complementaire
         Options.Complementaire := True;
      elsif Is_An_Alias(Argument(Count),"-INTERVAL",2) then
        -- indiquer la longueur du pas et le dcalage pour chaque
        -- intervalle
        Count := Count + 1;
        Options.Step_Length := Opt_Int(Argument(Count));
        Count := Count + 1;
        Options.Step_Recover := Opt_Int(Argument(Count));
        if Options.Step_Recover >= Options.Step_Length then
          -- option invalide, car cela amne  reculer dans la squence
          -- ou pire  rester toujours  la mme position
          raise ARGUMENT_ERROR;
        end if;
      elsif Is_An_Alias(Argument(Count),"-SEED",2) then
        -- indiquer la valeur initiale du generateur aleatoire
        Count := Count + 1;
        Options.Initial := Opt_Int(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-MUTATION",2) then
        -- indiquer le nombre de mutation que l'on souhaite
        Count := Count + 1;
        Options.Nb_Mutation := Opt_Int(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-ASCII",2) then
        -- preciser qul s'agit d'une fonction d'extraction ascii
        Options.choice := Extract_Nuc_Ascii;
      elsif Is_An_Alias(Argument(Count),"-GCG",2) then
        -- preciser qul s'agit d'une fonction d'extraction ascii
        Options.choice := Extract_Nuc_GCG;
      elsif Is_An_Alias(Argument(Count),"-PROTEIN",4) then
        -- preciser qu'il faut traduire en proteine
        Options.Protein := True;
      elsif Is_An_Alias(Argument(Count),"-FORMAT",2) then
        -- definir le format de lecture de la sequence
        Count := Count + 1;
        Options.Format := Opt_Sequence_Format(Argument(Count));
      elsif Is_An_Alias(Argument(Count),"-OUTPUT",2) then
        -- lire le nom du fichier de sortie
        Count := Count + 1;
        Copy(Argument(Count),Options.Output);
      else
        raise ARGUMENT_ERROR;
      end if;
      Count := Count + 1;
    end loop;
    -- il reste a lire le nom du fichier contenant la sequence
    if Count /= Argument_Count then
      -- il devrait laisser un argument a lire
      raise ARGUMENT_ERROR;
    end if;
    Copy(Argument(count),Options.Name);
    if Length(Options.Output) = 0 then
      Copy(Argument(Count),Options.Output);
    end if;
    if Options.Protein then
      case Options.Choice is
        when Extract_Nuc_Ascii => Options.Choice := Extract_Prot_Ascii;
        when Extract_Nuc_GCG   => Options.Choice := Extract_Prot_GCG;
        when others            => null;
      end case;
    end if;
    -- Test de la compatibilit des options
    if Options.Step_Length > 0 and not Is_Empty(Options.Positions) then
      -- on ne peut demander en mme temps une slection de fraction
      -- de squences et une extraction par intervalle
      raise ARGUMENT_ERROR;
    end if;
  end Get_Options;

  procedure Translate(Sequence : in Sequence_Data;
                      Protein  : in out Sequence_Data) is
    type Code_Genetique_Type is array
      (Nucleotide_Type,Nucleotide_Type,Nucleotide_Type) of Character;
    Code_Genetique : Code_Genetique_Type;
    Seq  : String_Access renames Sequence.Composition.Value;
    I    : Integer := Sequence.Start;
    J    : Integer;
    Length : Integer;
  begin
    Code_Genetique('A','A','A') := 'K';
    Code_Genetique('A','A','T') := 'N';
    Code_Genetique('A','A','G') := 'K';
    Code_Genetique('A','A','C') := 'N';
    Code_Genetique('A','T','A') := 'I';
    Code_Genetique('A','T','T') := 'I';
    Code_Genetique('A','T','G') := 'M';
    Code_Genetique('A','T','C') := 'I';
    Code_Genetique('A','G','A') := 'R';
    Code_Genetique('A','G','T') := 'S';
    Code_Genetique('A','G','G') := 'R';
    Code_Genetique('A','G','C') := 'S';
    Code_Genetique('A','C','A') := 'T';
    Code_Genetique('A','C','T') := 'T';
    Code_Genetique('A','C','G') := 'T';
    Code_Genetique('A','C','C') := 'T';
    --
    Code_Genetique('T','A','A') := '*';
    Code_Genetique('T','A','T') := 'Y';
    Code_Genetique('T','A','G') := '*';
    Code_Genetique('T','A','C') := 'Y';
    Code_Genetique('T','T','A') := 'L';
    Code_Genetique('T','T','T') := 'F';
    Code_Genetique('T','T','G') := 'L';
    Code_Genetique('T','T','C') := 'F';
    Code_Genetique('T','G','A') := '*';
    Code_Genetique('T','G','T') := 'C';
    Code_Genetique('T','G','G') := 'W';
    Code_Genetique('T','G','C') := 'C';
    Code_Genetique('T','C','A') := 'S';
    Code_Genetique('T','C','T') := 'S';
    Code_Genetique('T','C','G') := 'S';
    Code_Genetique('T','C','C') := 'S';
    --
    Code_Genetique('G','A','A') := 'E';
    Code_Genetique('G','A','T') := 'D';
    Code_Genetique('G','A','G') := 'E';
    Code_Genetique('G','A','C') := 'D';
    Code_Genetique('G','T','A') := 'V';
    Code_Genetique('G','T','T') := 'V';
    Code_Genetique('G','T','G') := 'V';
    Code_Genetique('G','T','C') := 'V';
    Code_Genetique('G','G','A') := 'G';
    Code_Genetique('G','G','T') := 'G';
    Code_Genetique('G','G','G') := 'G';
    Code_Genetique('G','G','C') := 'G';
    Code_Genetique('G','C','A') := 'A';
    Code_Genetique('G','C','T') := 'A';
    Code_Genetique('G','C','G') := 'A';
    Code_Genetique('G','C','C') := 'A';
    --
    Code_Genetique('C','A','A') := 'Q';
    Code_Genetique('C','A','T') := 'H';
    Code_Genetique('C','A','G') := 'Q';
    Code_Genetique('C','A','C') := 'H';
    Code_Genetique('C','T','A') := 'L';
    Code_Genetique('C','T','T') := 'L';
    Code_Genetique('C','T','G') := 'L';
    Code_Genetique('C','T','C') := 'L';
    Code_Genetique('C','G','A') := 'R';
    Code_Genetique('C','G','T') := 'R';
    Code_Genetique('C','G','G') := 'R';
    Code_Genetique('C','G','C') := 'R';
    Code_Genetique('C','C','A') := 'P';
    Code_Genetique('C','C','T') := 'P';
    Code_Genetique('C','C','G') := 'P';
    Code_Genetique('C','C','C') := 'P';
    -- initialisation de la chaine pour stocker la sequence proteique
    Length := (Sequence.Stop-Sequence.Start+1)/ 3;
    Free(Protein.Composition.Value);
    Protein.Composition.Value := new String(1..Length);
    J := Protein.Composition.Value'First;
    -- boucle de conversion
    while I+2 <= Sequence.Stop loop
      Protein.Composition.Value(J) := Code_Genetique(To_Nucleotide(Seq(I)),
                                                     To_Nucleotide(Seq(I+1)),
                                                     To_Nucleotide(Seq(I+2)));
      I := I + 3;
      J := J + 1;
    end loop;
    -- noter les infos associees
    Protein.Model := Sequence_Specifications.Protein;
    Protein.Start := Protein.Composition.Value'First;
    Protein.Stop  := Protein.Composition.Value'Last;
  end Translate;

  procedure Put_Date(File : Ada.Text_Io.File_Type) is
    use Ada.Calendar;
    Date : constant Time := Clock;
    Ecoule : constant Integer := Integer(Seconds(Date));
    Heure : constant Integer := Ecoule / 3600;
    Minute : constant Integer := (Ecoule - Heure * 3600) / 60;
  begin
    Put(File,' ');
    case Month(Date) is
      when  1 => Put(File,"January");
      when  2 => Put(File,"February");
      when  3 => Put(File,"March");
      when  4 => Put(File,"April");
      when  5 => Put(File,"May");
      when  6 => Put(File,"June");
      when  7 => Put(File,"July");
      when  8 => Put(File,"August");
      when  9 => Put(File,"September");
      when 10 => Put(File,"October");
      when 11 => Put(File,"November");
      when 12 => Put(File,"December");
    end case;
    Put(File,Day(Date),3);
    Put(File,',');
    Put(File,Year(Date),5);
    if Heure < 10 then
      Put(File,'0');
      Put(File,Heure,2);
    else
      Put(File,Heure,3);
    end if;
    Put(File,':');
    if Minute < 10 then
      Put(File,'0');
      Put(File,Minute,1);
    else
      Put(File,Minute,2);
    end if;
  end Put_Date;

  procedure Write_Fragment_Ascii(Sequence : in Sequence_Data;
                                 Name     : in String;
                                 Length   : in Integer := 60)  is
    Seq : String_Access renames Sequence.Composition.Value;
    Fic : Ada.Text_Io.File_Type;
    From, To : Integer;
  begin
    if Sequence.Model = Sequence_Specifications.Protein then
      Create(Fic,Out_File,Name & ".frgpep");
    else
      Create(Fic,Out_File,Name & ".frgseq");
    end if;
    -- ecrire l'entete
    Put_Line(Fic,"> " & Name);
    -- compter les frequences
    From := Sequence.Start;
    while From < Sequence.Stop loop
      To := From + Length - 1;
      if To > Sequence.Stop then
        To := Sequence.Stop;
      end if;
      Put_Line(Fic,Seq(From..To));
      From := To + 1;
    end loop;
    Close(Fic);
  end Write_Fragment_Ascii;

  procedure Write_Fragment_GCG(Sequence : in Sequence_Data;
                               Name     : in String)  is
    Seq : String_Access renames Sequence.Composition.Value;
    Fic : Ada.Text_Io.File_Type;
    From, To, Inf, Sup : Integer;
    Length   : constant Integer := 50;
  begin
    Create(Fic,Out_File,Name & ".gcg");
    -- ecrire l'entete
    Put(Fic,Name);
    Put(Fic," Length:");
    Put(Fic,Sequence.Stop-Sequence.Start+1,5);
    Put_Date(Fic);
    if Sequence.Model = Sequence_Specifications.Protein then
      Put(Fic,"  Type: P  Check: ..");
    else
      Put(Fic,"  Type: N  Check: ..");
    end if;
    New_Line(Fic,2);
    -- crire la sequence
    From := Sequence.Start;
    while From <= Sequence.Stop loop
      To := From + Length - 1;
      if To > Sequence.Stop then
        To := Sequence.Stop;
      end if;
      -- ecrire le debut de la ligne
      Put(Fic,From-Sequence.Start+1,8);
      Put(Fic,' ');
      -- ecrire les 5 blocs de 10 caracteres
      Inf := From;
      while Inf <= To loop
        Sup := Inf + 9;
        if Sup > To then
          Sup := To;
        end if;
        Put(Fic,' ');
        Put(Fic,Seq(Inf..Sup));
        Inf := Sup + 1;
      end loop;
      New_Line(Fic,2);
      -- passer a la ligne suivante
      From := To + 1;
    end loop;
    Close(Fic);
  end Write_Fragment_GCG;

  procedure Mute(Sequence : in out Sequence_Data;
                 Number   : in Integer;
                 Initial  : in Integer)  is
    Seq : String_Access renames Sequence.Composition.Value;
    subtype Intervalle is Integer range Sequence.Start..Sequence.Stop;
    package Intervalle_Random is new Ada.Numerics.Discrete_Random(Intervalle);
    subtype Thrue_Nucleotide_Type is Nucleotide_Type range 'A'..'C';
    package Nucleotide_Random is new
      Ada.Numerics.Discrete_Random(Thrue_Nucleotide_Type);
    use Intervalle_Random,Nucleotide_Random;
    I_Gen : Intervalle_Random.Generator;
    N_Gen : Nucleotide_Random.Generator;
    Pos : Intervalle;
    Base : Nucleotide_Type;
  begin
    if Initial = 0 then
      -- initialisation aleatoire
      Reset(I_Gen);
      Reset(N_Gen);
    else
      -- initialisation avec une graine fournie
      Reset(I_Gen,Initial);
      Reset(N_Gen,Initial+1);
    end if;
    for I in 1..Number loop
      Pos := Random(I_Gen);
      loop
        Base := Random(N_Gen);
        exit when Base /= To_Nucleotide(Seq(Pos));
      end loop;
      Put_Line("Mutation a la position" & Intervalle'Image(Pos) &
               " par " & Nucleotide_Type'Image(Base) & " a la place de '" &
               Seq(Pos) & "'");
      Seq(Pos) := To_Character(Base);
    end loop;
  end Mute;

  procedure Select_Zone(Sequence  : in out Sequence_Data;
                        Positions : in Segment_List) is
    -- Cette procedure retourne la sequence en ne conservant que les
    -- zones demandees. Il n'y a pas de controle sur les positions fournies
    O_Seq : String_Access renames Sequence.Composition.Value;
    N_Seq : String_Access;
    Entries : SLL.List_Entries;
    Length : Integer := 0;
    Current : Integer;
  begin
    if Is_Empty(Positions)  then
      -- il n'y a pas de selection a faire
      return;
    end if;
    -- recuperer les positions
    SLL.Get_Entries(Positions,Entries);
    -- compter la longueur
    for I in Entries'Range loop
      Length := Length + SLL.Consult(Entries(I)).Stop -
        SLL.Consult(Entries(I)).Start + 1;
    end loop;
    -- initialiser la chaine
    N_Seq := new String(1..Length);
    Current := N_Seq'First;
    -- extraire les zones
    for I in Entries'Range loop
      if (SLL.Consult(Entries(I)).Start < Sequence.Start or
          SLL.Consult(Entries(I)).Stop  > Sequence.Stop) then
        raise OUT_OF_SEQUENCE;
      end if;
      Length := SLL.Consult(Entries(I)).Stop -
        SLL.Consult(Entries(I)).Start;
      N_Seq(Current..Current+Length) :=
        O_Seq(SLL.Consult(Entries(I)).Start..SLL.Consult(Entries(I)).Stop);
      Current := Current+Length+1;
    end loop;
    -- liberer la place associee a l'ancienne sequence
    Free(O_Seq);
    -- et positionner la nouvelle sequence
    Sequence.Composition.Value := N_Seq;
    Sequence.Start := N_Seq'First;
    Sequence.Stop  := N_Seq'Last;
  end Select_Zone;

begin
  Put_Line("Command: " & Command_Name);
  Get_Options(Options);
  Put_Line("Reading: " & To_String(Options.Name));
  Read_Sequence(To_String(Options.Name),Sequence,DNA,Options.Format);
  Put_Line("La longueur de la sequence est" &
           Integer'Image(Sequence.Composition.Value'Length));
  -- prendre le complementaire ou l'inverse
  if Options.Inverse then
    Reverse_Sequence(Sequence);
  end if;
  if Options.Complementaire then
    Cmp_Sequence(Sequence);
  end if;
  -- selection de la zone indiquee
  Select_Zone(Sequence,Options.Positions);
  -- Prendre en compte si necessaire les positions de debut et de fin
  -- de sequence et en verifier la validite
  -- ATTENTION: begin et end referent aux zones selectionnees
  if Options.Start >= 0 and Options.Start >= Sequence.Start then
    Sequence.Start := Options.Start;
  else
    Options.Start := Sequence.Start;
  end if;
  if Options.Stop >= 0 and Options.Stop <= Sequence.Stop then
    Sequence.Stop := Options.Stop;
  else
    Options.Stop := Sequence.Stop;
  end if;
  -- realiser une option en fonction de l'option choisie
  -- muter la sequence si necessaire
  if Options.Nb_Mutation > 0 then
    Mute(Sequence,Options.Nb_Mutation,Options.Initial);
  end if;
  -- realiser une action en fonction de l'option choisie
  Current.Start := Options.Start;
  Build_Files: loop
    if Options.Step_Length > 0 then
      -- Cas d'une extraction par intervalle successif
      Sequence.Start := Current.Start;
      Sequence.Stop := Current.Start + Options.Step_Length - 1;
      if Sequence.Stop > Options.Stop then
        -- on atteind la borne suprieure
        Sequence.Stop := Options.Stop;
        -- c'est donc le dernier tour que l'on note en positionnant
        -- options.step_length  zro pour viter de boucler
        -- indfiniment sur le dernier fichier
        Options.Step_Length := 0;
      end if;
      Build_Name: declare
        Identificator : String(1..Identificator_Length);
      begin
        Put(Identificator,Decimal_Position(Current.Start),Pic_For_Name);
        Output_Name := Options.Output & "_" & Identificator;
        Ada.Text_Io.Put_line("Build " & Root_Name(Output_Name) &
                             " from" & Integer'Image(Sequence.Start) &
                             " to" & Integer'Image(Sequence.Stop));
    end Build_name;
    else
      Copy(Options.Output,Output_Name);
    end if;
    case Options.Choice is
      when Extract_Nuc_Ascii =>
        -- initialiser la matrice de stockage
        Write_Fragment_Ascii(Sequence,
                             Root_Name(Output_Name),
                             Options.Length);
      when Extract_Nuc_GCG =>
        -- initialiser la matrice de stockage
        Write_Fragment_GCG(Sequence,Root_Name(Output_Name));
      when Extract_Prot_Ascii =>
        Translate(Sequence,Protein);
        Write_Fragment_Ascii(Protein,Root_Name(Output_Name),Options.Length);
      when Extract_Prot_GCG =>
        Translate(Sequence,Protein);
        Write_Fragment_GCG(Protein,Root_Name(Output_Name));
    end case;
    Clear(Output_Name);
    -- Calcul de la nouvelle position initiale
    Current.Start := Sequence.Stop + 1 - Options.Step_Recover;
    -- dtermination du nouveau point de dpart
    exit when (Options.Step_Length <= 0 or
               Current.Start > Options.Stop or
               Current.Start < Options.Start);
    -- sortie de la boucle si l'option par intervalle n'est pas
    -- active ou si l'on a atteind la fin de la squence ou enfin
    -- si la nouvelle position de dpart est hors de la squence
    -- (cas d'un intervalle de recouvrement absurde)
  end loop Build_Files;
exception
  when OUT_OF_SEQUENCE =>
    Put_Line("Dans un intervalle, la position donnee est hors de");
    Put_Line("de la sequence");
  when INVALID_POSITION =>
    Put_Line("Dans un intervalle, la position de fin doit etre");
    Put_Line("plus grande ou egale a la position de debut.");
  when INVALID_COUPLE =>
    Put_Line("Seule la borne de debut a ete fournie pour l'intervalle");
  when ARGUMENT_ERROR =>
    Put_Line("sknife: argument error");
    Put_Line("      syntax is: sknife [-Ascii | -Gcg] ");
    Put_Line("                        [-Length <length_line>]");
    Put_Line("                        [-Format <input_format>]");
    Put_Line("                        [-Output <name>]");
    Put_Line("                        [-Mutation <nb_mutation>]");
    Put_Line("                        [-Interval <length> <step>]");
    Put_Line("                        [-Complementaire]");
    Put_Line("                        [-Reverse]");
    Put_Line("                        [-PROtein]");
    Put_Line("                        [-POSition [<start><stop>]* ]");
    Put_Line("                        [-Seed <seed_value>");
    Put_Line("                        [-Begin <start_position>]");
    Put_Line("                        [-End   <stop_position>] sequence_file");
    New_Line;
  when The_Error : others =>
    Put_Line(Ada.Exceptions.Exception_Name(The_Error));
    Put_Line(Ada.Exceptions.Exception_Message(The_Error));
    Put_Line(Ada.Exceptions.Exception_Information(The_Error));
end Sknife;
