--------------------------------------------------------------------------
-- Static_Fifo                              Copyrigth ENS ULM 1998-1999 --
-- Pierre VINCENS                   V 01.000                 18/04/1998 --
-- Pierre VINCENS                   V 01.001                 28/12/1998 --
-- Pierre VINCENS                   V 01.002                 04/03/1999 --
--------------------------------------------------------------------------
with Unchecked_Deallocation;

package body Static_Fifo is

  procedure Free is new Unchecked_Deallocation(Stack_Array,Stack_Access);

  procedure Push(Stack : in out Stack_Handler; Element : in Element_Type) is
  begin 
    if Stack.Data = null then
      Stack.Data := new Stack_Array(0..Stack.Size-1);  
    end if;
    if Stack.Count = Stack.Size then
      raise STACK_FULL;
    end if;
    Stack.Data(Stack.Bottom) := Element;
    Stack.Bottom := (Stack.Bottom + 1) mod Stack.Size;
    Stack.Count := Stack.Count + 1;
  end Push;

  procedure Pull(Stack : in out Stack_Handler; Element : out Element_Type) is
  begin
    if Stack.Count = 0 then
      raise STACK_EMPTY;
    end if;
    Element := Stack.Data(Stack.Top);
    Stack.Top := (Stack.Top + 1) mod Stack.Size;
    Stack.Count := Stack.Count - 1;
  end Pull;

  procedure Top(Stack : in Stack_Handler; Element : out Element_Type) is
  begin
    if Stack.Count = 0 then
      raise STACK_EMPTY;
    end if;
    Element := Stack.Data(Stack.Top);
  end Top;
  
  function Top(Stack : in Stack_Handler) return Element_Type is
  begin
    if Stack.Count = 0 then
      raise STACK_EMPTY;
    end if;
    return Stack.Data(Stack.Top);
  end Top;

  function Is_In(Stack : in Stack_Handler;
		 Value : in Element_Type) return Boolean is
    I : Natural := Stack.Top;
  begin
    for J in 1..Stack.Count loop
      if Feature(Stack.Data(I),Value) then
	return True;
      end if;
      I := (I + 1) mod Stack.Size;
    end loop;
    return False;
  end Is_In;
  
  procedure Resize(Stack : in out Stack_Handler; Size : in Positive) is
  begin
    if Size = Stack.Size then return; end if;
    -- La taille demandee est la meme, on ne fait rien
    if Stack.Data = null then Stack.Size := Size; return; end if;
    -- Il n'y a pas de tableau alloue, on prend juste en compte la 
    -- nouvelle taille
    if Stack.Count > Size then raise STACK_TOO_LARGE; end if;
    -- Le nombre d'elements de la pile est plus grand que la taille demandee
    declare
      I : Natural := Stack.Top;
      Data : Stack_Access := new Stack_Array(0..Size-1);
    begin
      for J in 0..Stack.Count-1 loop
        Data(J) := Stack.Data(I);
        I := (I + 1) mod Stack.Size;
      end loop;
      Free(Stack.Data);
      Stack.Data := Data;
      Stack.Bottom := Stack.Count mod Size;
      Stack.Top := 0;
      Stack.Size := Size;
    end;
  end Resize;

  procedure Reset(Stack : in out Stack_Handler) is
  begin
    Stack.Count  := 0;
    Stack.Top    := 0;
    Stack.Bottom := 0;
    Free(Stack.Data);
  end Reset;

  procedure Clean_And_Reset(Stack : in out Stack_Handler) is
    
  begin
    if Stack.Data /= null then
      declare
        Current : Natural := Stack.Top;
      begin
        for I in 1..Stack.Count loop
          Clean(Stack.Data(Current));
          Current := (Current + 1) mod Stack.Size;
        end loop;
      end;
    end if;
    Stack.Count  := 0;
    Stack.Top    := 0;
    Stack.Bottom := 0;
    Free(Stack.Data);
  end Clean_And_Reset;

  function Is_Empty(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Count = 0;
  end Is_Empty;

  function Is_Not_Empty(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Count /= 0;
  end Is_Not_Empty;

  function Is_Full(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Count = Stack.Size;
  end Is_Full;
  
  function Is_Not_Full(Stack : in Stack_Handler) return Boolean is
  begin
    return Stack.Count /= Stack.Size;
  end Is_Not_Full;
  
  function Count(Stack : in Stack_Handler) return Natural is
  begin
    return Stack.Count;
  end Count;

end Static_Fifo;
  
