--------------------------------------------------------------------------
-- Strlib.IO                                Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 06/11/1997 --
-- Pierre VINCENS                   V 01.001                 19/03/1999 --
--------------------------------------------------------------------------

package body Strlib.IO is

--===========================================================================
-- Quelques fonctions utilitaires
--===========================================================================
  function End_Of_String(Str : in String) return Natural is
    -- Cette fonction retourne la derniere position de la
    -- chaine qui n'est pas le caractere. On conserve toujours
    -- le premier caractere de la chaine
    Count : Natural := Str'Last;
  begin
    while Count > Str'First and then Str(Count) = ' ' loop
      Count := Count - 1;
    end loop;
    return Count;
  end End_Of_String;

--===========================================================================
-- Fonctionnalites d'entree-sortie (utilise text_io)
--===========================================================================
  procedure Put(Varstr : in Varstring) is
  begin
    if Varstr.Value /= null then
      Put(Varstr.Value.all);
    end if;
  end Put;

  procedure Put(File : in File_Type; Varstr : in Varstring) is
  begin
    if Varstr.Value /= null then
      Put(File,Varstr.Value.all);
    end if;
  end Put;

  procedure Put_Line(Varstr : in Varstring) is
  begin
    if Varstr.Value /= null then
      Put_Line(Varstr.Value.all);
    end if;
  end Put_Line;

  procedure Put_Line(File : in File_Type; Varstr : in Varstring) is
  begin
    if Varstr.Value /= null then
      Put_Line(File,Varstr.Value.all);
    end if;
  end Put_Line;

  procedure Put_The_N_Last_Characters(Str    : in String;
                                      N      : in Integer) is
  begin
    if Str'Length < N then
      -- la taille de la chaine est plus petite que le nombre de
      -- caracteres qu'il est demande d'imprimer. On les prend tous
      -- et on complete avec des blancs. L'alignement se fait a
      -- gauche
      Put(Str);
      for I in Str'Length+1..N loop
        Put(' ');
      end loop;
    else
      -- la taille de la chaine est plus longue. On ne conserve que
      -- les N derniers caracteres
      Put(Str(Str'Last-N+1..Str'Last));
    end if;
  end Put_The_N_Last_Characters;

  procedure Put_The_N_Last_Characters(Varstr : in Varstring;
                                      N      : in Integer) is
  begin
    if Varstr.Value /= null then
      Put_The_N_Last_Characters(varstr.Value.all,N);
    end if;
  end Put_The_N_Last_Characters;

  procedure Put_The_N_Last_Characters(File   : in File_Type;
                                      Str    : in String;
                                      N      : in Integer) is
  begin
    if Str'Length < N then
      -- la taille de la chaine est plus petite que le nombre de
      -- caracteres qu'il est demande d'imprimer. On les prend tous
      -- et on complete avec des blancs. L'alignement se fait a
      -- gauche
      Put(File,Str);
      for I in Str'Length+1..N loop
        Put(File,' ');
      end loop;
    else
      -- la taille de la chaine est plus longue. On ne conserve que
      -- les N derniers caracteres
      Put(File,Str(Str'Last-N+1..Str'Last));
    end if;
  end Put_The_N_Last_Characters;

  procedure Put_The_N_Last_Characters(File   : in File_Type;
                                      Varstr : in Varstring;
                                      N      : in Integer) is
  begin
    if Varstr.Value /= null then
      Put_The_N_Last_Characters(File,Varstr.Value.all,N);
    end if;
  end Put_The_N_Last_Characters;

  procedure Get(Varstr : in out Varstring;
                Length : in Positive;
                Mode   : in End_Mode := Normal) is
    Str : String(1..Length);
    Last : Natural;
  begin
    if Varstr.Value /= null then Free(Varstr.Value); end if;
    Get(Str);
    if Mode = Removing then
      -- eliminer les blancs de fin de chaine
      Last := End_Of_String(Str);
    else
      Last := Str'Last;
    end if;
    Varstr.Value := new String(1..Last);
    Varstr.Value.all := Str(1..Last);
  end Get;

  procedure Get(File   : in File_Type;
                Varstr : in out Varstring;
                Length : in Positive;
                Mode   : in End_Mode := Normal) is
    Str : String(1..Length);
    Last : Natural;
  begin
    if Varstr.Value /= null then Free(Varstr.Value); end if;
    Get(File,Str);
    if Mode = Removing then
      -- eliminer les blancs de fin de chaine
      Last := End_Of_String(Str);
    else
      Last := Str'Last;
    end if;
    Varstr.Value := new String(1..Last);
    Varstr.Value.all := Str(1..Last);
  end Get;

  procedure Get_Word(File     : in File_Type;
                     Varstr   : in out Varstring;
		     Length   : in Positive;
		     Pattern  : in String := " :.,?!") is
    Str : String(1..Length);
    Char : Character;
    Last : Natural := 0;
    End_Of_Line : Boolean;
    Found : Boolean;
  begin
    if Varstr.Value /= null then Free(Varstr.Value); end if;
    Word_Loop: while Last < Length loop
      -- on peut encore lire des caracteres
      Look_Ahead(File,Char,End_Of_Line);
      exit Word_Loop when End_Of_Line; -- on a atteint la fin de la ligne
      -- on lit effectivement le caractere
      Get(File,Char);
      Found := False;
      for I in Pattern'Range loop
	if Char = Pattern(I) then
	  if Last > 0 then
	    -- c'est la fin du mot
	    exit Word_Loop;
	  end if;
	  -- on n'a pas encore enregistre de caracteres
	  Found := True;
	  exit;
	end if;
      end loop;
      if not Found then
	-- il y a un caractere a enregistrer
	Last := Last + 1;
	Str(Last) := Char;
      end if;
    end loop Word_loop;
    if Last > 0 then
      Varstr.Value := new String(1..Last);
      Varstr.Value.all := Str(1..Last);
    end if;
  end Get_Word;

  procedure Get_Word(Varstr   : in out Varstring;
		     Length   : in Positive;
		     Pattern  : in String := " :.,?!") is
  begin
    Get_Word(Standard_Input,Varstr,Length,Pattern);
  end Get_Word;

  procedure Get_Line(Varstr     : in out Varstring;
                     Max_Length : in Positive := 256;
                     Mode       : in End_Mode := Normal) is
    Str : String(1..Max_Length);
    Last : Natural;
  begin
    if Varstr.Value /= null then Free(Varstr.Value); end if;
    Get_Line(Str,Last);
    if Mode = Removing then
      -- eliminer les blancs de fin de chaine
      Last := End_Of_String(Str(1..Last));
    end if;
    Varstr.Value := new String(1..Last);
    Varstr.Value.all := Str(1..Last);
  end Get_Line;

  procedure Get_Line(File       : in File_Type;
                     Varstr     : in out Varstring;
                     Max_Length : in Positive := 256;
                     Mode       : in End_Mode := Normal) is
    Str : String(1..Max_Length);
    Last : Natural;
  begin
    if Varstr.Value /= null then Free(Varstr.Value); end if;
    Get_Line(File,Str,Last);
    if Mode = Removing then
      -- eliminer les blancs de fin de chaine
      Last := End_Of_String(Str(1..Last));
    end if;
    Varstr.Value := new String(1..Last);
    Varstr.Value.all := Str(1..Last);
  end Get_Line;

--===========================================================================
-- La lecture en ignorant les caracteres souhaites
--===========================================================================
  procedure Selective_Get(Str    : in out String;
                          Ignore : in String := " ") is
    Last  : Natural := Str'First;
  begin
    while Last <= Str'Last loop
      Get(Str(Last));
      if not Is_In(Str(Last),Ignore) then
        Last := Last + 1;
      end if;
    end loop;
  end Selective_Get;

  procedure Selective_Get(File   : in File_Type;
                          Str    : in out String;
                          Ignore : in String := " ") is
    Last  : Natural := Str'First;
  begin
    while Last <= Str'Last loop
      Get(File,Str(Last));
      if not Is_In(Str(Last),Ignore) then
        Last := Last + 1;
      end if;
    end loop;
  end Selective_Get;

end Strlib.IO;
