--------------------------------------------------------------------------
-- Yaap                                     Copyrigth ENS ULM 1986-1999 --
-- Pierre VINCENS                   V 01.000                 21/01/1998 --
-- Pierre VINCENS                   V 01.001                 04/03/1999 --
-- Pierre VINCENS                   V 01.002                 02/03/2004 --
--------------------------------------------------------------------------
-- Yet Another Alignment Program
-- Alignement de deux squence nucliques et protiques
--------------------------------------------------------------------------
-- V 01.001
-- Support de l'alignement de deux squences protiques
--------------------------------------------------------------------------
-- V 01.002
-- Support de l'option Similarity_Threshold (calcul du % de similitude)
--------------------------------------------------------------------------
with Ada.Unchecked_Deallocation;
with Ada.Text_Io; use  Ada.Text_Io;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Exceptions; use Ada.Exceptions;

with Type_specifications; use Type_specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
use  Type_Specifications.IO.Int_Io;
use  Type_Specifications.IO.Real_Io;
use  Type_Specifications.Real_Lib;

with Strlib; use  Strlib;
with Strlib.io; use  Strlib.io;
with Options_Util; use  Options_Util;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Specifications.Basic; use Sequence_Specifications.Basic;
with Sequence_Io; use  Sequence_Io;

with Sequence_Alignment; use Sequence_Alignment;
with Sequence_Alignment.Basic; use Sequence_Alignment.Basic;
with Sequence_Alignment.Print; use Sequence_Alignment.Print;
with Sequence_Alignment.Global; use Sequence_Alignment.Global;
with Sequence_Alignment.Local; use Sequence_Alignment.Local;

with Yaap_Exceptions; use Yaap_Exceptions;
with Yaap_Options; use Yaap_Options;
with Yaap_Read; use Yaap_Read;
with Yaap_Version; use Yaap_Version;

procedure Yaap is

  type Sequence_Access is access all Sequence_Data;
  T_Seq1,T_Seq2 : Sequence_Access;

  Sequences, RC_Sequences : Sequence_Array_Access;
  Type_Of_Sequences : Sequence_Model;
  Start1,Stop1,Start2,Stop2 : Integer;

  Alignments    : Alignments_Access;
  Options       : Options_Type;
  Count         : Integer := 0;

  Pos_File      : Ada.Text_Io.File_Type;
  Align_File    : Ada.Text_Io.File_Type;
  Summary_File  : Ada.Text_Io.File_Type;

  procedure Put_Not_Align(File            : in Ada.Text_Io.File_Type;
			  Seq1,   Seq2    : in Sequence_Access;
			  Begin1, End1    : in Integer;
			  Begin2, End2    : in Integer;
			  Strand1,Strand2 : in Strand_Type) is
    Id_Length : constant Integer := Summary_Id_Length;
  begin
    Put_The_N_Last_Characters(File,To_String(Seq1.Identifier),Id_Length);
    Put(File," " & Image(Strand1));
    case Strand1 is
      when Watson =>
	Put(File,Begin1,12);
	Put(File,End1,12);
      when Crick =>
	Put(File,End1,12);
	Put(File,Begin1,12);
      when Both =>
	raise Invalid_Strand;
    end case;
    Put(File,' ');
    Put_The_N_Last_Characters(File,To_String(Seq2.Identifier),Id_Length);
    Put(File," " & Image(Strand2));
    case Strand2 is
      when Watson =>
	Put(File,Begin2,12);
	Put(File,End2,12);
      when Crick =>
	Put(File,End2,12);
	Put(File,Begin2,12);
      when Both =>
	raise Invalid_Strand;
    end case;
    Put_Line(File," NOT ALIGNED");
  end Put_Not_Align;
  
  procedure Swap(A,B : in out Integer);
  pragma Inline(Swap);
  -- Echange de deux variables entieres
  procedure Swap(A,B : in out Integer) is
    Buffer : Integer;
  begin
    Buffer := A;
    A := B;
    B := Buffer;
  end Swap;

begin
  -- Traitement de la ligne de commande
  Get_Options(Options);
  
  -- Lecture des squences
  if Length(Options.Listfile) = 0 then
    -- Ordonner les options begin et end en fonction du brin 
    -- pour les deux sequences
    case Options.Strand1 is
      when Watson => -- begin doit etre plus petit que stop
	if Options.Begin1 > Options.End1 then
	  Swap(Options.Begin1,Options.End1);
	end if;
      when Crick => -- begin doit etre plus grand que stop
	if Options.Begin1 < Options.End1 then
	  Swap(Options.Begin1,Options.End1);
	end if;
      when others =>
	raise Invalid_Strand;
    end case;
    case Options.Strand2 is
      when Watson => -- begin doit etre plus petit que stop
	if Options.Begin2 > Options.End2 then
	  Swap(Options.Begin2,Options.End2);
	end if;
      when Crick => -- begin doit etre plus grand que stop
	if Options.Begin2 < Options.End2 then
	  Swap(Options.Begin2,Options.End2);
	end if;
      when others =>
	raise Invalid_Strand;
    end case;
	  
    -- On lit les squences donnes dans la ligne de commande
    Sequences := new Sequence_Array(1..2);
    RC_Sequences := new Sequence_Array(1..2);
    -- Lecture de la sequence 1
    -- Si on est sur le brin Crick, il faut echanger les positions
    -- begin et end pour lire la sequence en reference Watson puis 
    -- les remettre dans l'ordre initial apres.
    if Options.Strand1 = Crick then
      Swap(Options.Begin1,Options.End1);
    end if;
    Read_One_Sequence(To_String(Options.Name1),Options.Identifier1,
		      Options.Model1,Options.Format1,
		      Options.Begin1,Options.End1,Options.Strand1,
		      Sequences(1),RC_Sequences(1),
		      Options.Inverse1,Options.Complementaire1,
		      Options.Interactif);
    if Options.Strand1 = Crick then
      Swap(Options.Begin1,Options.End1);
    end if;
    
    -- Lecture de la sequence 2
    -- Si on est sur le brin Crick, il faut echanger les positions
    -- begin et end pour lire la sequence en reference Watson puis
    -- les remettre dans l'ordre initial apres.
    if Options.Strand2 = Crick then
      Swap(Options.Begin2,Options.End2);
    end if;
    Read_One_Sequence(To_String(Options.Name2),Options.Identifier2,
		      Options.Model2,Options.Format2,
		      Options.Begin2,Options.End2,Options.Strand2,
		      Sequences(2),RC_Sequences(2),
		      Options.Inverse2,Options.Complementaire2,
		      Options.Interactif);
    if Options.Strand2 = Crick then
      Swap(Options.Begin2,Options.End2);
    end if;
  else
    -- On lit les squences donnes dans le fichier
    Read_List_Of_Sequences(To_String(Options.Listfile),
			   Sequences,RC_Sequences,
			   Options.Interactif);
  end if;
  -- Dterminer le type de squences  aligner et vrifier
  -- qu'elles sont toutes du meme type
  Type_Of_Sequences := Sequences(1).Model;
  for I in 2..Sequences'Last loop
    case Type_Of_Sequences is
      when DNA | RNA =>
	if Sequences(I).Model /= DNA and Sequences(I).Model /= RNA then
	  raise Inconsistent_Model;
	end if;
      when Protein =>
	if Sequences(I).Model /= Protein then
	  raise Inconsistent_Model;
	end if;
      when others =>
	raise Inconsistent_Model;
    end case;
  end loop;
  -- Ouvrir les fichier o seront mmoriss les rsulats:
  --     Pos_file pour les donnes  traiter
  --     Summary_File pour le fichier donnant les couples aligns
  --     Align_file pour le fichier donnant les alignements
  if Options.Poslist then
    -- On traite un fichier
    begin
      Open(Pos_File,In_File,To_String(Options.Posfile));
    exception
      when Name_Error =>
	Raise_Exception(Name_Error'Identity,
			"Position " & To_String(Options.Posfile));
    end;
    if Options.Print_Result then
      begin
	Create(Align_File,Out_File,
	       Root_Name(Options.Posfile) & ".aln");
      exception
	when Name_Error =>
	  begin
	    Create(Align_File,Out_File,
		   "/tmp/" & Root_Name(Options.Posfile) & ".aln");
	    Put_Line("Warning, alignments will be written in " &
		     "/tmp/" & Root_Name(Options.Posfile) & ".aln");
	  exception
	    when Name_Error =>
	      Raise_Exception(Name_Error'Identity,"Alignment " & 
			      To_String(Options.Posfile) & ".aln");
	  end;
      end;
    end if;
    if Options.Print_Summary then
      begin
	Create(Summary_File,Out_File,Root_Name(Options.Posfile) & ".sim");
      exception
	when Name_Error =>
	  begin
	    Create(Summary_File,Out_File,
		   "/tmp/" & Root_Name(Options.Posfile) & ".sim");
	    Put_Line("Warning, summary will be written in " &
		     "/tmp/" & Root_Name(Options.Posfile) & ".sim");
	  exception
	    when Name_Error =>
	      Raise_Exception(Name_Error'Identity,"Summary " & 
			      "/tmp/" & To_String(Options.Posfile) & ".sim");
	  end;
      end;
    end if;
  else
    declare
      Result_Name : constant String :=
	Root_Name(Options.Name1) & "_" & Root_Name(Options.Name2);
    begin
      if Options.Print_Result then
	begin
	  Create(Align_File,Out_File,Result_Name & ".aln");
	exception
	  when Name_Error =>
	    begin
	      Create(Align_File,Out_File,"/tmp/" & Result_Name & ".aln");
	      Put_Line("Warning, alignments will be written in " &
		       "/tmp/" & Result_Name & ".aln");
	    exception
	      when Name_Error =>
		Raise_Exception(Name_Error'Identity,"Alignment " & 
				To_String(Options.Posfile) & ".aln");
	    end;
	end;
      end if;
      if Options.Print_Summary then
	begin
	  Create(Summary_File,Out_File,Result_Name & ".sim");
	exception
	  when Name_Error =>
	    begin
	      Create(Summary_File,Out_File,Result_Name & ".sim");
	      Put_Line("Warning, summary will be written in " &
		       "/tmp/" & Result_Name & ".aln");
	    exception
	      when Name_Error =>
		Raise_Exception(Name_Error'Identity,"Summary " & 
				To_String(Options.Posfile) & ".sim");
	    end;
	end;
      end if;
    end;
  end if;

  -- Raliser l'alignement pour chaque couple dans le fichier
  -- ou pour la position passe en argument.
  Lecture_Position: loop
    -- ATTENTION pour proteine strand doit etre obligatoirement Watson
    if Options.Poslist then
      -- Il faut lire le brin et les positions  analyser dans
      -- le fichier
      Count := Count + 1;
      Analyse_Ligne: loop
        if End_Of_File(Pos_File) then
          Close(Pos_File);
          exit Lecture_Position;
        end if;
        declare
          Line  : String(1..1024);
          Length : Integer;
          Position : Integer := Line'First;
          Idname,Mot : Varstring;
        begin
          Get_Line(Pos_File,Line,Length);
          -- lire le nom de la premiere sequence (premier mot)
          Next_Word(Line(Position..Length),Idname,Position);
	  -- lire le brin (deuxieme mot)
          Next_Word(Line(Position..Length),Mot,Position);
	  -- En fonction du dernier caractre
	  case Type_Of_Sequences is
	    when DNA | RNA =>
	      case Mot.Value(Mot.Value'First) is
		when 'W' => 
		  Options.Strand1 := Watson; 
		when 'C' => 
		  Options.Strand1 := Crick; 
		when others => 
		  Options.Strand1 := Watson; 
		  -- On considere que l'on est alors sur le brin Watson
	      end case;
	    when Protein =>
	      Options.Strand1 := Watson; 
	    when others =>
	      raise Inconsistent_Model;
	  end case;
	  -- lecture de la position de debut et de fin
	  Get(Line(Position..Length),Options.Begin1,Position);
	  Get(Line(Position+1..Length),Options.End1,Position);
	  -- controler que l'ordre begin/end est bien ce qu'il faut en
	  -- fonction du brin
	  case Options.Strand1 is
	    when Watson => -- begin doit etre plus petit que stop
	      if Options.Begin1 > Options.End1 then
	        Swap(Options.Begin1,Options.End1);
	      end if;
	    when Crick => -- begin doit etre plus grand que stop
	      if Options.Begin1 < Options.End1 then
	        Swap(Options.Begin1,Options.End1);
	      end if;
	    when others =>
	      raise Invalid_Strand;
	  end case;
	  -- dterminer de quelles squences il s'agit
	  T_Seq1 := null;
	  for I in Sequences'Range loop
	    if ((Options.Begin1 >= Sequences(I).Start and 
		 Options.Begin1 <= Sequences(I).Stop) or 
		(Options.End1 >= Sequences(I).Start and 
		 Options.End1 <= Sequences(I).Stop)  or
		(Sequences(I).Start >= Options.Begin1 and
		 Sequences(I).Start <= Options.End1) or
		(Sequences(I).Stop >= Options.Begin1 and
		 Sequences(I).Stop <= Options.End1)) then
	      -- On est dans le bon intervalle
	      -- La comparaison se faisant toujours par rapport au brin
	      -- Watson
	      if Equal(Idname,Sequences(I).Identifier) then
		case Options.Strand1 is
		  when Watson => 
		    if Sequences(I).Model /= No then
		      T_Seq1 := Sequences(I)'access;
		    end if;
		  when Crick  => 
		    if RC_Sequences(I).Model /= No then
		      T_Seq1 := RC_Sequences(I)'access;
		    end if;
		  when others => 
		    Raise_Exception(Invalid_Strand'Identity,"U YAAP-1W");
		end case;
		exit;
	      end if;
	    end if;
	  end loop;
	  -- lire le nom de la seconde sequence
	  Position := Position + 1;
	  Next_Word(Line(Position..Length),Idname,Position);
	  -- lire le brin (deuxieme mot)
          Next_Word(Line(Position..Length),Mot,Position);
	  case Type_Of_Sequences is
	    when DNA | RNA =>
	      case Mot.Value(Mot.Value'First) is
		when 'W' => 
		  Options.Strand2 := Watson; 
		when 'C' => 
		  Options.Strand2 := Crick; 
		when others => 
		  Options.Strand2 := Watson; 
		  -- On considere que l'on est alors sur le brin Watson
	      end case;
	    when Protein =>
	      Options.Strand2 := Watson; 
	    when others =>
	      raise Inconsistent_Model;   
	  end case;
	  -- lecture de la position de debut et de fin
	  Get(Line(Position..Length),Options.Begin2,Position);
	  Get(Line(Position+1..Length),Options.End2,Position);
	  -- controler que l'ordre begin/end est bien ce qu'il faut en
	  -- fonction du brin
	  case Options.Strand2 is
	    when Watson => -- begin doit etre plus petit que stop
	      if Options.Begin2 > Options.End2 then
	        Swap(Options.Begin2,Options.End2);
	      end if;
	    when Crick => -- begin doit etre plus grand que stop
	      if Options.Begin2 < Options.End2 then
	        Swap(Options.Begin2,Options.End2);
	      end if;
	    when others =>
	      raise Invalid_Strand;
	  end case;
	  -- dterminer de quelles squences il s'agit
	  T_Seq2 := null;
	  for I in Sequences'Range loop
	    if ((Options.Begin2 >= Sequences(I).Start and 
		 Options.Begin2 <= Sequences(I).Stop) or 
		(Options.End2 >= Sequences(I).Start and 
		 Options.End2 <= Sequences(I).Stop) or
		(Sequences(I).Start >= Options.Begin2 and
		 Sequences(I).Start <= Options.End2) or
		(Sequences(I).Stop >= Options.Begin2 and
		 Sequences(I).Stop <= Options.End2)) then
	      -- On est dans le bon intervalle
	      -- La comparaison se faisant toujours par rapport au brin
	      -- Watson
	      if Equal(idname,Sequences(I).Identifier) then
		case Options.Strand2 is
		  when Watson => 
		    if Sequences(I).Model /= No then
		      T_Seq2 := Sequences(I)'access;
		    end if;
		  when Crick  => 
		    if RC_Sequences(I).Model /= No then
		      T_Seq2 := RC_Sequences(I)'access;
		    end if;
		  when others =>
		    Raise_Exception(Invalid_Strand'Identity,"U YAAP-2W");
		end case;
		exit;
	      end if;
	    end if;
	  end loop;
	  if T_Seq1 /= null and T_Seq2 /= null then
	    -- Il faut traiter ce couple
	    exit Analyse_Ligne;
	  else
	    -- Ce couple n'est pas dans l'intervalle specifie, on l'ignore
	    -- et on lit la ligne suivante
	    null;
          end if;
	  -- Liberer la place occupe par mot et idname
	  Clear(Mot);
	  Clear(Idname);
        end;
      end loop Analyse_Ligne;
    else
      -- On considre les squences passes en argument de la ligne
      -- de commande
      case Options.Strand1 is
	when Watson => T_Seq1 := Sequences(1)'Access;
	when Crick  => T_Seq1 := RC_Sequences(1)'Access;
	when others => Raise_Exception(Invalid_Strand'Identity,"U YAAP-L1");
      end case;
      case Options.Strand2 is
	when Watson => T_Seq2 := Sequences(2)'Access;
	when Crick  => T_Seq2 := RC_Sequences(2)'Access;
	when others => Raise_Exception(Invalid_Strand'Identity,"U YAAP-L2");
      end case;
    end if;
    -- Calculer les positions  prendre en compte en fonction du brin
    -- en tenant compte du fait que l'on ait sur un brin Crick
    -- et en prenant en compte l'ajustement ventuel
    if Options.Begin1 = Complete then
      Start1 := T_Seq1.Start - Options.Adjust;
    else
      if Options.Strand1 = Crick then
	Start1 := T_Seq1.Length-Options.Begin1-Options.Adjust+1; 
      else
    	Start1 := Options.Begin1 - Options.Adjust;
      end if;
    end if;
    if Options.End1 = Complete then
      Stop1 := T_Seq1.Stop + Options.Adjust;
    else
      if Options.Strand1 = Crick then
	Stop1 := T_Seq1.Length-Options.End1+Options.Adjust+1;
      else
    	Stop1 := Options.End1 + Options.Adjust;
      end if;
    end if;
    if Options.Begin2 = Complete then
      Start2 := T_Seq2.Start - Options.Adjust;
    else
      if Options.Strand2 = Crick then
	Start2 := T_Seq2.Length-Options.Begin2-Options.Adjust+1;
      else
    	Start2 := Options.Begin2 - Options.Adjust;
      end if;
    end if;
    if Options.End2 = Complete then
      Stop2 := T_Seq2.Stop + Options.Adjust;
    else
      if Options.Strand2 = Crick then
	Stop2 := T_Seq2.Length-Options.End2+Options.Adjust+1;
      else
    	Stop2 := Options.End2 + Options.Adjust;
      end if;
    end if;
    -- Valider la zone en la limitant  la portion de la squence
    -- charge
    if Start1 < T_Seq1.Start then
      Start1 := T_Seq1.Start;
    end if;
    if Stop1 > T_Seq1.Stop then
      Stop1 := T_Seq1.Stop;
    end if;
    if Start2 < T_Seq2.Start then
      Start2 := T_Seq2.Start;
    end if;
    if Stop2 > T_Seq2.Stop then
      Stop2 := T_Seq2.Stop;
    end if;

    -- Raliser l'alignement si possible
    if Stop1 < Start1  or Stop2 < Start2 then
      -- Il n'y a rien  aligner, on ignore la demande de traitement
      null;
    elsif (Stop1 - Start1 + 1 > Options.Max_Length or
        Stop2 - Start2 + 1 > Options.Max_Length) then
      -- La sequence est trop longue
      -- On donne les coordonnes d'origine non transformes
      -- le nom est donc en T_seq (celui du pointeur donnant la squence)
      -- et les positions en Seq car telle qu'entrer dans le programme.
      if Options.Summary then
	Put_Not_Align(Ada.Text_Io.Standard_Output,
		      T_Seq1,T_Seq2,
		      Options.Begin1,Options.End1,
		      Options.Begin2,Options.End2,
		      Options.Strand1,Options.Strand2);
      end if;
      if Options.Print_Summary then
	Put_Not_Align(Summary_File,
		      T_Seq1,T_Seq2,
		      Options.Begin1,Options.End1,
		      Options.Begin2,Options.End2,
		      Options.Strand1,Options.Strand2);
      end if;
    else
      -- realiser une option en fonction de l'option choisie
      declare
        Cn : String := Integer'Image(Count);
        Short_Name : aliased constant String :=
          Root_Name(Options.Name1);
        Long_Name : aliased constant String :=
          Short_Name & '_' & Cn(Cn'First+1..Cn'Last);
        type String_Accessible is access constant String;
        Result_Name : String_Accessible := Short_Name'Access;
      begin
        if Options.Poslist then
          Result_Name := Long_Name'Access;
        end if;
        case Options.Method is
          when Global_Alignment =>
            Perform_Global_Alignment(T_Seq1.all,T_Seq2.all,Alignments, 
				     Start1,Stop1,
				     Start2,Stop2,
				     To_String(Options.Matrix),
				     Options.Match_Weight,
				     Options.Mismatch_Weight,
				     Options.Gap_Weight,
				     Options.Gap_Length_Weight,
				     Options.Progression,
				     Options.Building,
				     Options.Optimize,
				     Options.With_End_Weight,
				     Options.Similarity_Threshold,
				     Options.Randomization,
				     Options.Preserve);
  
          when Local_Alignment =>
	    Perform_Local_Alignment(T_Seq1.all,T_Seq2.all,Alignments,
                                    Start1,Stop1,
				    Start2,Stop2,
				    To_String(Options.Matrix),
				    Options.Match_Weight,
				    Options.Mismatch_Weight,
				    Options.Gap_Weight,
				    Options.Gap_Length_Weight,
				    Options.Progression,
				    Options.Building,
				    Options.Optimize,
				    Options.Similarity_Threshold,
				    Options.Randomization,
				    Options.Preserve);
          when others =>
            raise NOT_IMPLEMENTED;
        end case;
	-- Modifier les positions de dbut et de fin de l'alignement pour
	-- tenir compte du brin sur lequel s'est fait le calcul
	-- ATTENTION, pour proteines ce doit etre obligatoirement
	-- en Watson
	for I in Alignments'Range loop
	  Alignments(I).Strand1 := Options.Strand1; 
	  case Options.Strand1 is
	    when Watson =>
	      null;
	    when Crick =>
	      Alignments(I).Start1 := 
		T_Seq1.Length - Alignments(I).Start1 + 1;
	      Alignments(I).Stop1  := 
		T_Seq1.Length - Alignments(I).Stop1 + 1;
	    when others =>
	      Raise_Exception(Invalid_Strand'Identity,"U YAAP-A1");
	  end case;
	  Alignments(I).Strand2 := Options.Strand2; 
	  case Options.Strand2 is
	    when Watson =>
	      null;
	    when Crick =>
	      Alignments(I).Start2 := 
		T_Seq2.Length - Alignments(I).Start2 + 1;
	      Alignments(I).Stop2  := 
		T_Seq2.Length - Alignments(I).Stop2 + 1;
	    when others =>
	      Raise_Exception(Invalid_Strand'Identity,"U YAAP-A2");
	  end case;
	end loop;
	-- Imprimer les rsultats d'alignement si demand
	if Options.Print_Result then
          for I in Alignments'Range loop
	    Print_Alignment(Align_File,Alignments(I).all);
	    if I /= Alignments'Last then
	      New_Line(Align_File,2);
	      Put_Line(Align_File,"---------------");
	      New_Line(Align_File,2);
	    end if;
	  end loop;
	end if;
	-- Imprimer le rsum  l'cran
	for I in Alignments'Range loop
	  -- Imprimer  l'cran
	  if Options.Summary then
	    Print_Summary(Ada.Text_Io.Standard_Output,Alignments(I).all);
	    New_Line(Ada.Text_Io.Standard_Output);
	  end if;
	  -- et dans un fichier si demand
	  if Options.Print_Summary and then
	    Alignments(I).Length >= Options.Lmin and then
	    Alignments(I).Identity >= Options.Smin 
	  then 
	    Print_Summary(Summary_File,Alignments(I).all);
	    New_Line(Summary_File);
	  end if;
	end loop;
	-- Librer la place
	Free(Alignments);
      end ;
    end if;
    exit Lecture_Position when not Options.Poslist;
  end loop Lecture_Position;
  -- Fermer le fichier de rsultat si ncessaire
  if Options.Print_Result then
    Close(Align_File);
  end if;
  if Options.Print_Summary then
    Close(Summary_File);
  end if;
exception
  when ARGUMENT_ERROR =>
    Print_Options;
  when The_Error : others =>
    Check_Exception(The_Error);
end Yaap;

