--------------------------------------------------------------------------
-- Yapp_Exceptions                          Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 01.000                 17/03/1999 --
--------------------------------------------------------------------------
with Ada.Text_Io; use Ada.Text_Io;

with Type_specifications; use Type_Specifications;
with Type_Specifications.IO; use Type_Specifications.IO;
with Strlib; use Strlib;
with Strlib.IO; use Strlib.IO;

with Sequence_Specifications; use Sequence_Specifications;
with Sequence_Io; use Sequence_Io;

with Sequence_Alignment; use Sequence_Alignment; 
with Sequence_Alignment.Utils; use Sequence_Alignment.Utils; 

with Yaap_Options; use Yaap_Options;

package body Yaap_Exceptions is

  procedure Check_Exception(The_Error : in Exception_Occurrence) is
    Word1,Word2 : Varstring;
    Current,Last : Natural;

    procedure Read_First_Word is
    begin
      Current := Exception_Message(The_Error)'First;
      Next_Word(Exception_Message(The_Error),Word1,Current," ");
    end Read_First_Word;

    procedure Read_Second_Word is
      -- ATTENTION: Cette procedure ne doit etre appel qu'aprs 
      -- read_first_word
    begin
      Last := Exception_Message(The_Error)'Last;
      Next_Word(Exception_Message(The_Error)(Current..Last),
		Word2,Current," ");
    end Read_Second_Word;

    procedure Read_Two_Words is
    begin
      Read_First_Word;
      Read_Second_Word;
    end Read_Two_Words;

  begin
    -- Sinon on traite en fonction de l'erreur
    Put_Line("Program interrupted!");
    if Exception_Identity(The_Error) = Inconsistent_Model'Identity then
      Put_Line("You have requested an alignment between DNA and");
      Put_Line("proteins. It is unsupported");
      New_Line;
    elsif Exception_Identity(The_Error) = Inconsistent_Range'Identity then
      Put_Line("You have requested an incorrect range for one ");
      Put_Line("sequence. Please check.");
      New_Line;
    elsif Exception_Identity(The_Error) = Not_Implemented'Identity then
      Put_Line("The required method to align sequences is currently not");
      Put_Line("available.");
      New_Line;
    elsif Exception_Identity(The_Error) = Matrix_Name_Error'Identity then
      -- Erreur d'ouverture d'un fichier
      Read_First_Word;
      Put_Line("The matrix file called " &
	       Word1 & " has not been found. Please check.");
    elsif Exception_Identity(The_Error) = Name_Error'Identity then
      -- Erreur d'ouverture d'un fichier
      Read_Two_Words;
      Put_Line("The " & Word1 & " file called " &
	       Word2 & " has not been found or cannot be created. Please check.");
    elsif Exception_Identity(The_Error) = Sequence_Io_Error'Identity then
      -- Erreur dans le contenu d'un fichier
      Read_First_Word;
      Put_Line("The file called " & Word1 & 
	       " can not be read correctly. Please check.");
    elsif Exception_Identity(The_Error) = Invalid_Data'Identity then
      -- Erreur dans le nom du format donnee
      Read_First_Word;
      if To_String(Word1) = "O_Building" then
	Put_Line("The building option is invalid. Please check.");
      elsif To_String(Word1) = "O_Optimize" then
	Put_Line("The optimize option is invalid. Please check.");
      elsif To_String(Word1) = "O_Backward" then
	Put_Line("The backward option is invalid. Please check.");
      elsif To_String(Word1) = "O_Match_Weight" then
	Put_Line("The match weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Mismatch_Weight" then
	Put_Line("The mismatch weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Gap_Weight" then
	Put_Line("The gap weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Gap_Length_Weight" then
	Put_Line("The gap length weight option is invalid. Please check.");
      elsif To_String(Word1) = "O_Similarity_Threshold" then
	Put_Line("The similarity threshold option is invalid. Please check.");
      elsif To_String(Word1) = "O_Randomization" then
	Put_Line("The randomization option is invalid. Please check.");
      elsif To_String(Word1) = "O_Preserve" then
	Put_Line("The preserve option is invalid. Please check.");
      elsif To_String(Word1) = "O_Length_Max" then
	Put_Line("The length max option is invalid. Please check.");
      elsif To_String(Word1) = "O_Lmin" then
	Put_Line("The length minimal option (Lmin) is invalid. Please check.");
      elsif To_String(Word1) = "O_Smin" then
	Put_Line("The minimal percent of identity option is invalid.");
	Put_Line("Valid range is 0.0..100.0. Please check.");
      elsif To_String(Word1) = "O_Adjust" then
	Put_Line("The adjust option is invalid. Please check.");
      elsif To_String(Word1) = "O_Name" then
	Read_Second_Word;
	Put_Line("You have not specified the name of the sequence " & Word2 & 
		 " to align. Please check.");
      elsif To_String(Word1) = "L_Name" then
	Read_Second_Word;
	Put_Line("You have specified a name for the sequence " & Word2 & 
		 " and then given");
	Put_Line("a file containing a list of sequences. Please check.");
      elsif To_String(Word1) = "L_Poslist" then
	Put_Line("You have specified a list of sequence and not given a file");
	Put_Line("describing the list of couples to align. Please check.");
      elsif To_String(Word1) = "F_Start" then
	Read_Second_Word;
	Put_Line("The start position in line " & Word2 & 
		 " is invalid. Please check.");
      elsif To_String(Word1) = "F_Stop" then
	Read_Second_Word;
	Put_Line("The stop position in line " & Word2 & 
		 " is invalid. Please check.");
      elsif To_String(Word1) = "F_Direction" then
	Read_Second_Word;
	Put_Line("The specifications of direction in line " & Word2 & 
		 " are invalid. Please check.");
      elsif To_String(Word1) = "O_Start" then
	Read_Second_Word;
	Put_Line("The start position for sequence " & Word2 &
		 " is invalid. Please check.");
      elsif To_String(Word1) = "O_Stop" then
	Read_Second_Word;
	Put_Line("The stop position for sequence " & Word2 &
		 " is invalid. Please check.");
     elsif To_String(Word1) = "P_Start" then
	Read_Second_Word;
	Put_Line("The start position for sequence " & Word2 &
		 " is out of range. Please check.");
      elsif To_String(Word1) = "P_Stop" then
	Read_Second_Word;
	Put_Line("The stop position for sequence " & Word2 &
		 " is out of range. Please check.");
       else
	Put_Line("Raised " & Exception_Information(The_Error));      
      end if;
    elsif Exception_Identity(The_Error) = Invalid_Sequence_Format'Identity then
      -- Erreur dans le nom du format donnee
      Read_First_Word;
      if To_String(Word1) = "L" then
	Read_Second_Word;
	Put_Line("The sequence format specified in line " & Word2 & 
		 " is invalid. Please check.");
      else
	Put_Line("The sequence format specified with option " & Word1 &
		 " is invalid. Please check.");
      end if;
    elsif Exception_Identity(The_Error) = Invalid_Sequence_Model'Identity then
      -- Erreur dans le nom du format donnee
      Read_First_Word;
      if To_String(Word1) = "L" then
	Read_Second_Word;
	Put_Line("The sequence type specified in line " & Word2 & 
		 " is invalid. Please check.");
      else
	Put_Line("The sequence type specified with option " & Word1 &
		 " is invalid. Please check.");
      end if;
    elsif Exception_Identity(The_Error) = Invalid_Strand'Identity then
      -- Erreur dans le nom du brin donnee
      Read_First_Word;
      if To_String(Word1) = "L" then
	Read_Second_Word;
	Put_Line("The strand specified in line " & Word2 & 
		 " is invalid. Please check.");
      elsif To_String(Word1) = "U" then
	Read_Second_Word;
	Put_Line("Internal error in " & Word2 & " (Strand unexpected)");
      else
	Put_Line("The strand specified for sequence " & Word1 &
		 " is invalid. Please check.");
      end if;
    else
      -- Erreur indeterminee
      Put_Line("Raised " & Exception_Information(The_Error));
    end if;
  end Check_Exception;

end Yaap_Exceptions;
