--------------------------------------------------------------------------
-- Yaap_Read                                Copyrigth ENS ULM 1996-1999 --
-- Pierre VINCENS                   V 00.000                 17/03/1999 --
--------------------------------------------------------------------------
with Ada.Text_Io; use  Ada.Text_Io;
with Ada.Exceptions; use Ada.Exceptions;

with Type_Specifications.IO; use Type_Specifications.IO;
with Options_Util; use Options_Util;
with Strlib.IO; use Strlib.IO;

with Sequence_Io; use Sequence_Io;
with Sequence_Alignment; use Sequence_Alignment;

with Yaap_Options; use Yaap_Options;

package body Yaap_Read is

--------------------------------------------------------------------------
-- Acces aux bibliothques                                              --
--------------------------------------------------------------------------
  use Type_Specifications.IO.Int_Io;
   
--------------------------------------------------------------------------
-- Lecture d'une squence                                               --
--------------------------------------------------------------------------
  procedure Read_One_Sequence(Name          : in String;
                              Identifier    : in Varstring;
                              Model         : in Sequence_Model;
                              Format        : in Sequence_Format;
                              Start, Stop   : in Integer;
			      Strand        : in Strand_Type;
                              Sequence      : in out Sequence_Data;
                              RC_Sequence   : in out Sequence_Data;
                              Opposite      : in Boolean := False;
                              Complementary : in Boolean := False;
			      Verbose       : in Boolean := False) is
  begin
    -- Lecture de la squence
    Read_Sequence(Name,Sequence,Model,Format,Internal);
    
    -- Afficher le rsultat de la lecture
    if Verbose then
      case Sequence.Model is
	when DNA | RNA =>
	  Put_Line("Sequence " & Name & 
		   " has been read in format " & 
		   Sequence_Format'Image(Sequence.Format));
	  Put_Line("Length is" & Integer'Image(Sequence.Nucleotides'Length) &
		   " nucleotides.");
	when Protein =>
	  Put_Line("Sequence " & Name & 
		   " has been read in format " & 
		   Sequence_Format'Image(Sequence.Format));
	  Put_Line("Length is" &
		   Integer'Image(Sequence.Amino_Acids'Length) &
		   " amino acids.");
	when others => 
	  raise INVALID_SEQUENCE_MODEL;
      end case;
    end if;
    
    -- Noter la position de dbut et de fin a prendre en compte
    -- Si "complete", cela veut dire que l'on prend la position
    -- de dbut /fin de squence pour dbuter/terminer l'alignement
    -- Cela se fait en considrant les positions donnes par rapport
    -- au brin Watson
    if Start = Complete then 
      -- On garde comme dbut la position initiale de la squence
      null;
    elsif Start < Sequence.Start or Start > Sequence.Stop then
      -- On est en dehors de l'espace de coordonnes de la squence
      Raise_Exception(Invalid_Data'Identity,"P_Start " & Name);
    else
      -- La position demande est dans l'intervalle acceptable
      -- On enregistre la nouvelle valeur
      Sequence.Start := Start;
    end if;
    if Stop = Complete then 
      -- On garde comme fin la position finale de la squence
      null;
    elsif Stop < Sequence.Start or Stop > Sequence.Stop then
      -- On est en dehors de l'espace de coordonnes de la squence
      Raise_Exception(Invalid_Data'Identity,"P_Stop " & Name);
    else
      -- La position demande est dans l'intervalle acceptable
      -- On enregistre la nouvelle valeur
      Sequence.Stop := Stop;
    end if;
    
    -- Prparation de la squence
    Rc_Sequence.Model := No; -- Pour etre sur que ce champ est a no
                             -- sir rc_sequence n'est pas utilise
    case Sequence.Model is
      when DNA | RNA =>
	Copy(Identifier,Sequence.Identifier);
	if Opposite then
	  Reverse_Sequence(Sequence);
	end if;
	if Complementary then
	  Cmp_Sequence(Sequence);
	end if;
	if Strand = Crick or Strand = Both then
	  -- On calcule maintenant la complementaire inverse et on duplique la
	  -- sequence pour ne pas ecraser Seq1.Name et Seq1.Nucleotides lors 
	  -- de la copie
	  RC_Sequence := Sequence;
	  RC_Sequence.Name.Value := null;       -- pour ne pas ecraser les
	  RC_Sequence.Identifier.Value := null; -- champs correspondants de 
	  RC_Sequence.Nucleotides :=            -- sequence
	    new Nucleotidic_Array(Sequence.Nucleotides'Range);
	  RC_Sequence.Nucleotides.all := Sequence.Nucleotides.all;
	  RC_Sequence.Start := Sequence.Nucleotides'Last - Sequence.Stop + 1;
	  RC_Sequence.Stop := Sequence.Nucleotides'Last - Sequence.Start + 1;
	  Reverse_Sequence(RC_Sequence);
	  Cmp_Sequence(RC_Sequence);
	  Copy(Sequence.Name,RC_Sequence.Name);
	  Copy(Identifier,RC_Sequence.Identifier);
	end if;
      when Protein =>
	Copy(Identifier,Sequence.Identifier);
      when others => 
	raise INVALID_SEQUENCE_MODEL;
    end case;
  exception
    when Name_Error =>
      Raise_Exception(Name_Error'Identity,"sequence " & Name);
    when Sequence_Io_Error =>
      Raise_Exception(Sequence_Io_Error'Identity,Name);
  end Read_One_Sequence;

--------------------------------------------------------------------------
-- Lecture du fichier contenant les noms des squences                  --
--------------------------------------------------------------------------
  procedure Read_List_Of_Sequences(Name        : in String;
				   Sequences   : in out Sequence_Array_Access;
				   R_Sequences : in out Sequence_Array_Access;
				   Verbose     : in Boolean := False) is
    -- Les lignes commencant par un # sont ignorees. Si une option n'est pas
    -- precisee, toutes les options suivantes sont prises par defaut.
    -- Le format de la ligne du fichier est:
    -- Nom_Sequence Format Brin Identificateur Debut Fin
    File            : Ada.Text_Io.File_Type;
    Count           : Integer := 0;
    Line            : String(1..1024);
    Line_Number     : Natural := 0;
    Last, S_Last    : Natural;
    Current         : Natural;
    Word            : Varstring;
    S_Id            : Integer := 0;
    S_Name          : Varstring;
    S_Identifier    : Varstring;
    S_Start,S_Stop  : Integer;
    S_Strand        : Strand_type;
    S_Model         : Sequence_Model;
    S_Format        : Sequence_Format;
    S_Opposite      : Boolean;
    S_Complementary : Boolean;

    procedure Ignore_Spaces is
      -- Incremente current jusqu'au premier caractre non blanc
      -- ou jusqu' la fin de la ligne
    begin
      while Line(Current) = ' ' loop
        Current := Current + 1;
        exit when Current > Last;
      end loop;
    end Ignore_Spaces;

  begin
    -- Vrifier que les structures squences ne contiennent pas de donnes
    if Sequences /= null or R_Sequences /= null then
      raise Invalid_Access;
    end if;
    -- Compter le nombre de sequences a traiter
    begin
      Open(File,In_File,Name);
    exception
      when Name_Error =>
	Raise_Exception(Name_Error'Identity,"list " & Name);
    end;
    while not End_Of_File(File) loop
      -- Lire la ligne
      Get_Line(File,Line,Last);
      -- Verifier la validite de la ligne
      for i in 1..Last loop
        if Line(i) /= ' ' and line(i) /= '#' then
          -- Ce n'est pas une ligne de commentaire
          Count := Count + 1;
          exit;
        end if;
        exit when Line(i) = '#'; -- C'est une ligne de commentaire
      end loop;      
    end loop;
    Close(File);
    -- Reserver l'espace pour memoriser les sequences
    Sequences := new Sequence_Array(1..Count);
    R_Sequences := new Sequence_Array(1..Count);
    -- Lire effectivement les sequences
    Open(File,In_File,Name);
    while not End_Of_File(File) loop
      Line_Number := Line_Number + 1;
      -- Lire la ligne
      Get_Line(File,Line,Last);
      -- Ignorer les blancs initiaux
      Current := Line'First;
      Ignore_Spaces;
      -- Regarder s'il reste quelque chose sur la ligne
      if Current <= Last and Line(Current) /= '#' then
        -- On lit le nom du fichier contenant la squence (Obligatoire)
        Next_Word(Line(Current..Last),S_Name,Current," ");
        -- Definir le type de la sequence
	Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Model := NO;
	  else
	    S_Model := Opt_Sequence_Model(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Model := NO;
        end if;
        -- Definir le format de la sequence
	Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Format := Unknown;
	  else
	    S_Format := Opt_Sequence_Format(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Format := Unknown;
        end if;
        -- Definir le brin a analyser (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
          if Equal(Word,"-") then
	    S_Strand := Both;
	  else
	    S_Strand := Opt_Sequence_Strand(To_String(Word));
	  end if;
        else
          -- Valeur par defaut
          S_Strand := Both;
        end if;
        -- Definir l'identificateur de la sequence (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En le lisant dans le fichier
          Next_Word(Line(Current..Last),S_Identifier,Current," ");
	  if Equal(S_Identifier,"-") then
	    S_Identifier := To_Varstring(Root_Name(S_Name));
	  end if;
        else
          -- Ou en le deduisant du nom de la sequence
          S_Identifier := To_Varstring(Root_Name(S_Name));
        end if;
        -- Definir la position de debut (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En la lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Start := Complete;
	  else
	    begin
	      Get(To_String(Word),S_Start,S_Last);
	    exception
	      when Data_Error => 
		Raise_Exception(Invalid_Data'Identity,
				"F_Start " & Natural'Image(Line_Number));
	    end;
	  end if;
        else
          -- Valeur par defaut
          S_Start := Complete;
        end if;
        -- Et la position de fin (Optionnel)
        Ignore_Spaces;
        if Current <= Last then
          -- En la lisant dans le fichier
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Equal(Word,"-") then
	    S_Stop := Complete;
	  else
	    begin
	      Get(To_String(Word),S_Stop,S_Last);
	    exception
	      when Data_Error => 
		Raise_Exception(Invalid_Data'Identity,
				"F_Stop " & Natural'Image(Line_Number));
	    end;
	  end if;
        else
          -- Valeur par defaut
          S_Stop := Complete;
        end if;
	-- Indiquer si l'on doit prendre le reverse ou le complementaire
        Ignore_Spaces;
	S_Opposite := False;
	S_Complementary := False;
	while Current <= Last loop
	  -- on lit le mot suivant
          Next_Word(Line(Current..Last),Word,Current," ");
	  if Is_An_Alias(To_String(Word),"REVERSE",1) then
	    S_Opposite := True;
	  elsif Is_An_Alias(To_String(Word),"COMPLEMENTARY",1) then
	    S_Complementary := True;
	  elsif Is_An_Alias(To_String(Word),"RC",2) then
	    S_Opposite := True;
	    S_Complementary := True;
	  elsif Equal(Word,"-") then
	    S_Opposite := False;
	    S_Complementary := False;
	  else
	    Raise_Exception(Invalid_Data'Identity,
			    "F_Direction " & Natural'Image(Line_Number));
	  end if;
	  Ignore_Spaces;
        end loop;
	-- Affichage des infos
	if Verbose then
	  Put_Line("Sequence to read:");
	  Put_Line("   Name       : " & To_String(S_Name));
	  Put_Line("   Type       : " & Sequence_Model'Image(S_Model));
	  Put_Line("   Format     : " & Sequence_Format'Image(S_Format));
	  Put_Line("   Strand     : " & Strand_Type'Image(S_Strand));
	  Put_Line("   Identifier : " & To_String(S_Identifier));
	  Put     ("   Range      : "); 
	  if S_Start < 0 then
	    Put("-");
	  else
	    put(S_Start,0);
	  end if;
	  Put(".."); 
	  if S_Stop < 0 then
	    Put("-");
	  else
	    Put(S_Stop,0);
	  end if;
	  New_Line;
	  Put     ("   Direction  : ");
	  if S_Opposite then Put("R "); else Put("N "); end if;
	  if S_Complementary then Put("C "); else Put("N"); end if;
	  New_Line;
	end if;
        -- Il ne reste plus qu'a lire la sequence
        S_Id := S_Id + 1;
        Read_One_Sequence(To_String(S_Name),S_Identifier,
                          S_Model,S_Format,S_Start,S_Stop,S_Strand,
                          Sequences(S_Id),R_Sequences(S_Id),
			  S_Opposite,S_Complementary,
			  Verbose);
      end if;
    end loop;
    Close(File);
  exception
    when Invalid_Sequence_Format =>
      Raise_Exception(Invalid_Sequence_Format'Identity,
		      "L " & Natural'Image(Line_Number));
    when Invalid_Strand =>
      Raise_Exception(Invalid_Strand'Identity,"L " & 
		      Natural'Image(Line_Number));
  end Read_List_Of_Sequences;

end Yaap_Read;
